//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/

#include "stdafx.h"
#include "Qmol.h"

#include "MainFrm.h"
#include "QmolDoc.h"
#include "QmolView.h"
#include "CommandLine.h"
#include "ReadProgressDialog.h"
#include "TextOutput.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CQmolApp

BEGIN_MESSAGE_MAP(CQmolApp, CWinApp)
	//{{AFX_MSG_MAP(CQmolApp)
	ON_COMMAND(ID_APP_ABOUT, OnAppAbout)
	ON_COMMAND(ID_APP_HELP, OnAppHelp)
	//}}AFX_MSG_MAP
	// Standard file based document commands
	ON_COMMAND(ID_FILE_NEW, CWinApp::OnFileNew)
	ON_COMMAND(ID_FILE_OPEN, CWinApp::OnFileOpen)
	ON_COMMAND(ID_FILE_PRINT_SETUP, CWinApp::OnFilePrintSetup)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
//Global variables
/////////////////////////////////////////////////////////////////////////////
int g_bpp = 16;		// By default, use a 16 bit color display for OpenGL
int g_z_buf = 16;	// By default, use a 16 bit Z buffer for OpenGL
int g_purge_mru = FALSE; // By default, save the MRU list

/////////////////////////////////////////////////////////////////////////////
// CQmolApp construction

CQmolApp::CQmolApp()
{
}

CQmolApp::~CQmolApp()
{
	if(g_purge_mru){
		// Remove the MRU file list
		WriteProfileString("Recent File List", "File1", NULL);
		WriteProfileString("Recent File List", "File2", NULL);
		WriteProfileString("Recent File List", "File3", NULL);
		WriteProfileString("Recent File List", "File4", NULL);
	}
}


/////////////////////////////////////////////////////////////////////////////
// The one and only CQmolApp object

CQmolApp theApp;


/////////////////////////////////////////////////////////////////////////////
// CQmolApp initialization

BOOL CQmolApp::InitInstance()
{
	// Set the registry key under which our settings are stored.
	// Note that this registry key is used for all Shalloway Lab
	// applications, including Qmol.
	SetRegistryKey(_T("Shalloway Lab Applications"));

	LoadStdProfileSettings();  // Load standard INI file options (including MRU)

	// Register the application's document templates.  Document templates
	// serve as the connection between documents, frame windows and views.
	CSingleDocTemplate* pDocTemplate;
	pDocTemplate = new CSingleDocTemplate(
		IDR_MAINFRAME,
		RUNTIME_CLASS(CQmolDoc),
		RUNTIME_CLASS(CMainFrame),       // main SDI frame window
		RUNTIME_CLASS(CQmolView));
	AddDocTemplate(pDocTemplate);

	// Initialize the Rich text control in the text output dialog
	AfxInitRichEdit();

	// Parse command line for standard shell commands, DDE, file open
	CQmolCommandLine cmdInfo;
	ParseCommandLine(cmdInfo);

	// Has the user specified screen color depth or Z buffer size on 
	// the command line?
	if(cmdInfo.m_set_bpp){
		g_bpp = cmdInfo.m_bpp;
	}

	if(cmdInfo.m_set_zbuf){
		g_z_buf = cmdInfo.m_zbuf;
	}

	// Dispatch commands specified on the command line
	if (!ProcessShellCommand(cmdInfo))
		return FALSE;
	
	// The one and only window has been initialized, so show it.
	m_pMainWnd->ShowWindow(SW_SHOW);

	// Do we need to process any input data?
	POSITION doc_loc = pDocTemplate->GetFirstDocPosition();
	CQmolDoc *pDoc = (CQmolDoc *)(pDocTemplate->GetNextDoc(doc_loc));

	// ----- Save command line arguments as default options -----
	// Do we need to determine covalent bonds from atomic coordinates?
	pDoc->connect_on_read = cmdInfo.ConnectOnRead();
	
	// Do we need to interpret multiple structures in a coordinate file
	// as a trajectory?
	pDoc->multi_model_traj = cmdInfo.m_multi_traj;
	
	if(pDoc->multi_model_traj){
		pDoc->multi_model_traj_process = cmdInfo.TrajProcess();
	}

	if(cmdInfo.m_importPDB || cmdInfo.m_importAUTO){
		try{

			if((cmdInfo.m_import_coord_filename.Length() > 1) &&
			   (cmdInfo.m_importDCD)){
				throw QmolError("Only one structure can be read when reading "
					"DCD files");
			}
			
			ListIterator<CString> next(cmdInfo.m_import_coord_filename);
			CString *tmpString;

			while(tmpString = next()){
				pDoc->ReadFile(LPCTSTR(*tmpString));
			}
			
			if(cmdInfo.m_importDCD){
				pDoc->activeMol->ReadTrajHeader(LPCTSTR(cmdInfo.m_import_traj_filename), 
					QM_DCD);
					
				// Bring up a dialog to keep track of the reading progress
				CReadProgressDialog read_progress(0, pDoc->activeMol->steps, pDoc->activeMol,
					cmdInfo.m_import_traj_filename);

				// Read the trajectory
				read_progress.DoModal();

				// Did the user hit cancel?
				if(!read_progress.Status())
					throw QMOL_ERROR("Trajectory read has been canceled");

				// Compute the bounds of the structure + trajectory
				pDoc->activeMol->GetMolDim(cmdInfo.TrajProcess());

				// Set the Window Title
				pDoc->SetTitle(LPCTSTR(cmdInfo.m_import_coord_filename[0] + "+" +
					cmdInfo.m_import_traj_filename));
			
			}
			else{
				// Set the Window Title
				pDoc->SetTitle(LPCTSTR(cmdInfo.m_import_coord_filename[0]));
			}

			pDoc->UpdateAllViews(NULL);
			m_pMainWnd->UpdateWindow();

			if(cmdInfo.m_help)
				OnAppHelp();
		}
		catch(QmolError &e){
				
			AfxMessageBox(e.message());

			// Set any values that may have been modified by reading the
			// trajectory header back to their default vaules.
			pDoc->activeMol->steps = 0;
			pDoc->activeMol->size = 0;
				
			while(!pDoc->activeMol->traj.IsEmpty())
				pDoc->activeMol->traj.Pop();				

			// Allow just the specified structure to be displayed
			// (but not the trajectory, of course).
			pDoc->activeMol->GetMolDim();
		}
	}
	else{
		// Set the Window Title -- no molecule loaded
		pDoc->SetTitle("No Molecule loaded");

		if(cmdInfo.m_help)
			OnAppHelp();
	}

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// CAboutDlg dialog used for App About

class CAboutDlg : public CDialog
{
public:
	CAboutDlg();

// Dialog Data
	//{{AFX_DATA(CAboutDlg)
	enum { IDD = IDD_ABOUTBOX };
	CString	m_timestamp;
	CString	m_version;
	CString	m_screen;
	CString	m_info;
	CString	m_gl_render;
	//}}AFX_DATA

	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CAboutDlg)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

// Implementation
protected:
	//{{AFX_MSG(CAboutDlg)
		// No message handlers
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

CAboutDlg::CAboutDlg() : CDialog(CAboutDlg::IDD)
{
	//{{AFX_DATA_INIT(CAboutDlg)
	m_timestamp = _T("");
	m_version = _T("");
	m_screen = _T("");
	m_info = _T("");
	m_gl_render = _T("");
	//}}AFX_DATA_INIT
	
	m_version.Format("Qmol (win32) version %.2f", QMOL_VERSION);
	m_timestamp.Format("Build date: %s", __TIMESTAMP__);
	m_screen.Format("Using a %d bit Z buffer with %d bpp display",
		g_z_buf, g_bpp);
	m_gl_render.Format("%s", glGetString(GL_RENDERER));

	// Give credit where credit is due:
	m_info.Format("J.D. Gans\r\n"
		"Shalloway Lab\r\n"
		"Cornell University\r\n"
		"http://www.mbg.cornell.edu/shalloway/shalloway.html\r\n"
		"Copyright (C)  2000-2002\r\n"
		"\r\n"
		"Stride (http://www.embl-heidelberg.de/stride/stride.html)\r\n"
		"is used for secondary structure determination and has been\r\n"
		"included with the kind permission of\r\n"
		"Dmitrij Frishman, PhD\r\n"
		"Institute for Bioinformatics\r\n"
		"GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH\r\n"
		"Ingolstdter Landstrae 1,\r\n"
		"D-85764 Neuherberg, Germany\r\n"
		"Telephone: +49-89-3187-4201\r\n"
		"Fax: +49-89-31873585\r\n"
		"e-mail: d.frishman@gsf.de\r\n"
		"WWW: http://mips.gsf.de/mips/staff/frishman/\r\n"
		"\r\n"
		"Covalent bond determination code based on VMD\r\n"
		"http://www.ks.uiuc.edu/Research/vmd/\r\n"
		"\r\n"
		"Printing routines are based on code from\r\n"
		"Craig Fahrnbach <craignan@home.com> via\r\n"
		"Uwe Kotyczka <kotyczka@bnm-gmbh.de>.\r\n"
		"\r\n"
		"OpenGL selection code and quaterion rotation matrix\r\n"
		"formulation from molview demo program by Mark Kilgard\r\n"
		"(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript\r\n"
		"rendering based on work of Mark Kilgard.\r\n"
		"\r\n"
		"Data decompression performed by zlib\r\n"
		"Mark Adler and Jean-loup Gailly\r\n"
		"http://www.gzip.org/zlib/\r\n"
		"\r\n"
		"PNG graphics by libpng\r\n"
		"http://www.libpng.org/pub/png/\r\n"
		"\r\n"
		"Surface Rendering Algorithm (Marching Cubes)\r\n"
		"based on implementation of Cory Bloyd (corysama@yahoo.com)\r\n"
		"http://astronomy.swin.edu.au/pbourke/modelling/polygonise/\r\n");
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CAboutDlg)
	DDX_Text(pDX, IDC_STATIC_DATE, m_timestamp);
	DDX_Text(pDX, IDC_STATIC_VERSION, m_version);
	DDX_Text(pDX, IDC_STATIC_SCREEN, m_screen);
	DDX_Text(pDX, IDC_INFO_EDIT, m_info);
	DDX_Text(pDX, IDC_STATIC_GL, m_gl_render);
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CAboutDlg, CDialog)
	//{{AFX_MSG_MAP(CAboutDlg)
		// No message handlers
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

// App command to run the dialog
void CQmolApp::OnAppAbout()
{
	CAboutDlg aboutDlg;
	aboutDlg.DoModal();
}

void CQmolApp::OnAppHelp()
{
	CTextOutput helpDlg("Command Line Options", 
		"[filename]\t\tOpen coordinate file (type based on file extension)\r\n"
		"-pdb [filename]\tRead PDB coordinate file\r\n"
		"-dcd [filename]\tRead DCD trajectory file\r\n"
		"-align\t\tAlign all trajectory structures by rotation and translation\r\n"
		"-cm\t\tRemove the center of mass from all trajectory structures\r\n"
		"-no_connect\tDon't compute covalent bonds from atomic coordinates\r\n"
		"-multi\t\tInterpret multiple structures in a pdb file as a trajectory\r\n"
		"-bpp [number]\tOverride the default color depth\r\n"
		"-zbuf [number]\tOverride the default Z buffer size\r\n"
		"-?\t\tShow the list of allowed command line options\r\n"
		"-help\t\tShow the list of allowed command line options");
	helpDlg.DoModal();
}

/////////////////////////////////////////////////////////////////////////////
// CQmolApp commands


CDocument* CQmolApp::OpenDocumentFile(LPCTSTR lpszFileName) 
{
	// Get a pointer to the current document
	POSITION doc_loc = GetFirstDocTemplatePosition();
	CDocTemplate* pDocTemplate = GetNextDocTemplate(doc_loc);

	doc_loc = pDocTemplate->GetFirstDocPosition();
	CQmolDoc *pDoc = (CQmolDoc *)(pDocTemplate->GetNextDoc(doc_loc));
	
	if(pDoc == NULL){
		AfxMessageBox("Unable to get valid document. Doh!");
		return NULL;
	}

	// Perform new document work (just for show)
	//pDoc->OnNewDocument();
		
	// Open the File
	try{
		if(pDoc->ReadFile(lpszFileName) == FALSE)
			throw QmolError("Unable to read file",
				__FILE__, __LINE__);

		// Set the window title according to the MRU file
		// we just opened.
		pDoc->SetTitle(lpszFileName);
	}
	catch(QmolError &e){
		// Was there an error opening the file?
		AfxMessageBox(e.message());

		pDoc->SetTitle("No Molecule Loaded");

		// Don't return yet
	}
	catch(...){
		// Yiikes! An unhandled error has occured!
		// Was there an error opening the file?
		AfxMessageBox("Unable to read input file!");

		pDoc->SetTitle("No Molecule Loaded");

		// Don't return yet
	}

	// Refresh the view
	pDoc->UpdateAllViews(NULL);

	// Return a pointer to the existing document
	return pDoc;

	// Don't call the default CWinApp::OpenDocumentFile
	//return CWinApp::OpenDocumentFile(lpszFileName);
}
