//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/
//
//
//	*********** CHANGE LOG ***********
//	Version 1.5
//		- Fixed bug in read progress bar (that caused progress indicator
//		  to cycle several times.
//		- Fixed bug in Kabsch rotation.
//		- Fixed bug in the handling of -pdb import flag.
//		- Fixed bug in light control for the wire frame model (caused
//		  images that were copied to the clipboard to be incorrectly lighted).
//		- Fixed visability bug in the save coordinate dialog.
//		- Added capability to generate uncompressed AVI movies from a 
//		  trajectory.
//		- Fancify the buttons on the Animate dialog.
//	Version 1.6
//		- Added user defined resolution for geometric objects.
//		- Removed the OpenCoordDialog (open a CFileDialog instead).
//		- Fixed bug in pasting an image to clipboard (windows 95/98 only).
//	Version 1.7
//		- Added an option to save (or not save) the CONECT records when
//		  writing a pdb file.
//		- Fixed bug saving a pdb file (MODEL records are now written to each
//		  pdb file).
//		- Fixed bug in solvent display (the "Solvent" menu item should now
//		  work correctly).
// Version 1.8
//		- Added secondary structure assigment via STRIDE (provided by 
//		  Dr. Dmitrij Frishman).
//		- Added color-by-secondary structure
//		- Added trace rendering option to display alpha carbon trace as
//		  a wireframe, stick or spline.
//		- Added variable line widths for wire frame and spline display.
//		- Fixed color-by-element bug in Color dialog.
//		- Fixed adjust atom radius dailog.
//		- Added solid ribbon, flat ribbon, tube (constant raidus, tempature
//		  and occupancy based radius).
//		- Added split-screen stereo.
//	Version 1.9
//		- Fixed bug in ribbon orientation.
//		- Fixed bug in ribbon width.
//		- Added a #define COMPUTE_FPS to QmolView.h to turn on the
//		  computation of number of frames per second rendered to the
//		  screen.
//		- Fixed bug in pdb file reader (multiple models can now reuse
//		  atom numbers).
//	Version 1.91
//		- Added backface culling the ribbon models (for better looking
//		  translucent ribbons).
//		- Added Protein Data Bank queries (via the File menu). Given a 
//		  pdb code and a server name, download and display the corresponding
//		  file.
//		- Put the rendering throttle back in OnDraw(). If we don't limit
//		  the number of frames per second the message loop can get clogged.
//		- Misc fixes to PDB parser.
//
//	Version 1.92
//		- More misc fixes for the PDB parser.
//		- Increased the hardcoded STRIDE limitations on the number of
//		  secondary structure elements.
//
//	Version 1.93
//		- Added a field to the network PDB dialog to monitor the progress
//		  PDB downloads.
//		- Added compression options for AVI (movie) creation.
//
//	Version 1.94
//		- Improved sphere rendering (now render partial spheres that are 
//		  always aligned with the viewer).
//		- Modified file open and save dialogs to always use the current
//		  directory.
//		- Added hydrogen bond detection and display
//		- Fixed torsion angle rotation bug
//
//	Version 1.95
//		- Disbled anti-aliasing for stippled lines. There seems to be
//		  a bug (?) in the NVida OpenGL implemetation that causes periodic
//		  redrawing of the entire screen when anti-aliasing stippled lines.
//		- Cleaned up pdb file import to improve speed and code clarity.
//		- Added the build option to inline ANY suitable function (as opposed
//		  to only "inline" functions).
//		- Added a new option (under "options") to control mouse sensitvity
//		  by scaling the virtual trackball size.
//		- Added support for the wheel found on many pointing devices.
//		  The wheel now zooms the view.
//		- Added a command on the menu bar to toggle the display of the status
//		  bar at the bottom of the window.
//		- Improved reading of DCD headers to compute the number of frames in the
//		  DCD file (needed for reading VMD generated DCD files).
//
//	Version 1.96
//		- Fixed pdb parser to allow the reading of files where the residue name
//		  and chain id have been concatinated.
//		- Allow the creation of AVI movies containing a single frame.
//		- Solvent can now only be selected if it is currently being displayed.
//
//	Version 1.97
//		- Improved the lighting defaults for solid rendering.
//		- Increased the range of the custom resolution slider bars.
//		- Added a DEBUG only option to toggle wire frame models on ('w') and
//		  off ('s').
//		- Added the ability to paste ASCII data from the clipboard into molecular 
//		  temperature and occupancy values.
//		- Improved dynamic memory management in STRIDE code.
//		- Fixed a bug that prevented the reading of multiple model files that failed
//		  to specify model numbers (but gave chain ID's instead). This enables 
//		  Qmol to read multi-layer SwissPDB files.
//		- Added user defined clipping plane (complete with "cut" sphere rendering
//		  to provide the illusion of volume and stereo corrected clipping).
//		- Added color-by-residue number.
//		- Improved normalization scheme for pdb temperature and occupancy values.
//
//	Version 1.98
//		- Added PostScript export option (via Mark J. Kilgard's OpenGL->EPS code).
//		  Adaptive grid needs some adjustment and polygon splitting (to handle
//		  overlapping polygons with the painters algorithm) must still be 
//		  implemented.
//		- Fixed hydrogen bond line stipple bug.
//		- Fixed torsion rotation bug (Qmol crashed after selecting a torsion!).
//		- Added material property sliders to the lighting dialog.
//
//	Version 1.99
//		- Improved the lighting dialog by removing the gratuitous axes bitmap
//		  that only caused text alignment problems.
//		- Added protein sequence tool.
//		- Fixed the ribbon lighting (broken as of 1.98).
//		- Fixed bug saving multiple model connect records in a pdb file.
//		- Fixed printing bug (that produced postage stamp images).
//		- Images are now copied to the clipboard as bitmaps instead of metafiles.
//
//	Version 2.00
//		- Improved the rendering of Hydrogen Bonds by stippling "by hand"
//		  (i.e. rendering as a set of paralle line segments).
//		- Changed the clipping plane representation to be a circle instead of a
//		  square.
//		- Modified the save coordinate function to save the current orientation of
//		  the molecule(s).
//		- Reduced the sensitivity of the mouse (for moving molecules).
//		- Added the ability to read gzip compressed pdb files (which is now the
//		  default format when downloading data from the Protein Databank). The zlib
//		  library provides the decompression engine.
//		- Added an "Export Image ..." option. Postscript, BMP and PNG are the 
//		  supported formats. libpng is the PNG engine.
//
//	Version 2.01
//		- Added molecular surfaces based on Marching Cube implementation by
//		  Cory Bloyd (corysama@yahoo.com; 
//		  http://astronomy.swin.edu.au/pbourke/modelling/polygonise/).
//
//	Version 2.02
//		- Fixed bug in surface generation (boundary artifacts in small
//		  molecules).
//
//	Version 2.03
//		- Added the ability to import multiple structures in a pdb file
//		  as a molecular trajectory.
//		- Fixed a bug that prevented setting surface lighting options correctly.
//		- Added persistant to Qmol settings. Most settings are now stored in the
//		  registry to provide consistent behaviour from one session to the next.
//		  Options are provided to (a) return to the default configuration and
//		  (b) purge the registry of all Qmol values (i.e. an uninstall).
//		- Added new elements to the status bar, including window size while
//		  resizing the window, atom number, atom id and amino acid.
//		- Added a color-by-amino-acid-index to the color dialog. Useful for
//		  distinguishing between the amino and carboxy terminals.
//
//	Version 3.0
//		- That's right, version 3.0 -- a major upgrade for Qmol!
//		- Allow the reading of multiple structures/trajectories into Qmol.
//		  Each structure can be moved and adjusted independantly. Each
//		  structure can be rendered with a different representation (i.e.
//		  lines, ribbons, space-filling, etc.).
//		- Structures that have been read into Qmol can be duplicated (i.e.
//		  "cloned").
//		- Multiple files can be selected for reading.
//		- Allow the calculation of RMSD between two structures (with
//		  user determined atom selection). The structures do not have to 
//		  be of the same protein or chemical.
//		- Align two molecules by rigid-body translation and rotation
//		  (using a user specified sub-set of atoms).
//		- Measurement, hydrogen bond determination and RMSD calculation
//		  can all be performed in "real" time (i.e. move a molecule and
//		  watch dependant values change).
//		- Added a "scale-to-fit" option to rescale the coordinate system
//		  while moving molecules to keep all molecules visible.
//		- The measurement and torsion angle adjustment mechanism has been
//		  (hopefully) improved. Selecting atoms and torsion angles now
//		  only occurs when the measurement dialog is displayed (for atoms)
//		  or the adjust torsion angle dialog is displayed (for torsion 
//		  angles). 
//		- Selecting an atom with the pointer displays atom and residue
//		  info in the status bar.
//		- The code for computing the optimal rigid body rotation and 
//		  translation to minimize the RMSD between two molecules has been
//		  fixed to allow alignment of co-planar structures (i.e. ring groups).
//		  Co-linear molecules are still not allowed.

#define	QMOL_VERSION	3.00

#if !defined(AFX_QMOL_H__979F19C6_DCC9_11D3_9099_00A0C9A3671F__INCLUDED_)
#define AFX_QMOL_H__979F19C6_DCC9_11D3_9099_00A0C9A3671F__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#ifndef __AFXWIN_H__
	#error include 'stdafx.h' before including this file for PCH
#endif

#include "resource.h"       // main symbols

/////////////////////////////////////////////////////////////////
// All defined values and enuerated types are included here
#include "QmolDefine.h"
/////////////////////////////////////////////////////////////////

// Include the OpenGL headers
#include <GL\gl.h>
#include <GL\glu.h>


/////////////////////////////////////////////////////////////////////////////
// CQmolApp:
// See Qmol.cpp for the implementation of this class
//

class CQmolApp : public CWinApp
{
public:
	CQmolApp();
	~CQmolApp();

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CQmolApp)
	public:
	virtual BOOL InitInstance();
	virtual CDocument* OpenDocumentFile(LPCTSTR lpszFileName);
	//}}AFX_VIRTUAL

// Implementation

	//{{AFX_MSG(CQmolApp)
	afx_msg void OnAppAbout();
	afx_msg void OnAppHelp();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};


/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Developer Studio will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_QMOL_H__979F19C6_DCC9_11D3_9099_00A0C9A3671F__INCLUDED_)
