//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/
// 
#include "stdafx.h"
#include <cderr.h>
#include "Qmol.h"

#include "QmolDoc.h"
#include "SaveCoordDialog.h"
#include "OpenCoordTrajec.h"
#include "NetworkPBDDialog.h"
#include "pdb.h"
#include "DCD.h"
#include "kabsch.h"
#include "ReadProgressDialog.h"
#include "MainFrm.h"
#include "QmolUtil.h"
#include "LinearAlgebra3D.h"
#include "ReadOptionsDialog.h"

// The Stride secondary structure assignment program header
#include "stride/stride.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CQmolDoc

IMPLEMENT_DYNCREATE(CQmolDoc, CDocument)

BEGIN_MESSAGE_MAP(CQmolDoc, CDocument)
	//{{AFX_MSG_MAP(CQmolDoc)
	ON_COMMAND(ID_OPTIONS_READ, OnOptionsRead)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CQmolDoc construction/destruction

struct AtomHash{
	int i;
	int value;
	int model;
};

// Helper functions
int CompareAtomHash( const void *arg1, const void *arg2 );

int CompareAtomHash( const void *arg1, const void *arg2 )
{
	// First compare by model number
	if(((AtomHash*)arg1)->model < ((AtomHash*)arg2)->model)
		return -1;

	if(((AtomHash*)arg1)->model == ((AtomHash*)arg2)->model){
		// If model numbers are equal, compare on the value
		if(((AtomHash*)arg1)->value < ((AtomHash*)arg2)->value)
			return -1;

		if(((AtomHash*)arg1)->value == ((AtomHash*)arg2)->value)
			return 0;

		return 1;
	}

	return 1;
}


CQmolDoc::CQmolDoc()
{
	// Allocate a little memory to get the ball rolling!
	//mol = new molecule;
	activeMol = NULL;

	// Initialize the various color schemes
	SetColorScheme(QM_COLOR_DEFAULT);
	SetColorScheme(QM_COLOR_STRUC_DEFAULT);
	SetColorScheme(QM_COLOR_AMINO);
	
	// Set the elemental masses to their default values
	SetAtomicMassScheme();

	// Initialize the server and query lists to sensible defaults.
	// These defaults will change as PDB servers come and go!
	server_list.Push(DEFAULT_PDB_SERVER);

	// Set the initial query to a PDB file know to exist
	pdb_query_list.Push("1ckw");

	// Set import and data file processing options
	connect_on_read = TRUE;
	secondary_struct_on_read = TRUE;
	multi_model_traj = FALSE;
	multi_model_traj_process = QM_VERBATIM;
}

CQmolDoc::~CQmolDoc()
{
	//if(mol != NULL)
	//	delete mol;

	// Delete the list of molecules
	for(int i = 0;i < molList.Length();i++){
		delete molList[i];
	}
}

/////////////////////////////////////////////////////////////////////////////
// CQmolDoc diagnostics

#ifdef _DEBUG
void CQmolDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CQmolDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CQmolDoc commands
void atom::SetAtomColor(float clr_index)
{
	// clr_index must be in the range [0, 1]
	int r, g, b;

	// Parameterize the 3D color space by a single parameter, clr_index.
	// clr_index = 0.0 --> Blue
	// clr_index = 0.5 --> Yellow
	// clr_index = 1.0 --> Red

	r = (int)(255.0f*clr_index);

	if(clr_index < 0.5)
		g = (int)(510.0f*clr_index);
	else // clr_index >= 0.5
		g = (int)(510.0f*(1.0 - clr_index));

	b = (int)(255.0f*(1.0 - clr_index));

	r = (r > 255) ? 255 : r;
	r = (r < 0) ? 0 : r;

	g = (g > 255) ? 255 : g;
	g = (g < 0) ? 0 : g;

	b = (b > 255) ? 255 : b;
	b = (b < 0) ? 0 : b;

	rgb_color[0] = r; 
	rgb_color[1] = g; 
	rgb_color[2] = b;

	// Make the atom opaque
	rgb_color[3] = 255;

}

void atom::SetAtomColor(GLubyte color_scheme[ELEM_LAST][4])
{
	if(elem >= 0 && elem < ELEM_LAST){
		rgb_color[0] = color_scheme[elem][0]; 
		rgb_color[1] = color_scheme[elem][1]; 
		rgb_color[2] = color_scheme[elem][2];
		rgb_color[3] = color_scheme[elem][3];
	}
	else{
		rgb_color[0] = color_scheme[ELEM_DUMMY][0]; 
		rgb_color[1] = color_scheme[ELEM_DUMMY][1]; 
		rgb_color[2] = color_scheme[ELEM_DUMMY][2];
		rgb_color[3] = color_scheme[ELEM_DUMMY][3];
	}
}

void atom::SetAtomRadius(float radius_scheme[])
{
	if(elem >= 0 && elem < ELEM_LAST){
		radius = radius_scheme[elem];
	}
	else{
		radius = radius_scheme[ELEM_DUMMY];
	}
}

void atom::SetAtomMass(float mass_scheme[])
{
	if(elem >= 0 && elem < ELEM_LAST){
		mass = mass_scheme[elem];
	}
	else{
		mass = mass_scheme[ELEM_DUMMY];
	}
}

void atom::GetElement(char *elem_name)
{
	unsigned int i;
	
	// Clean up previously allocated memory if necessary.
	if(name)
		delete [] name;
		
	name = new char [strlen(elem_name) + 1];
	strcpy(name, elem_name);
	
	// First, make the element name all upper case
	for(i = 0;i < strlen(elem_name);i++)
		elem_name[i] = toupper(elem_name[i]);
		
	// For now, use only the first letter of the element name
	i = 0;
	while(!isalpha(elem_name[i]) && (i < strlen(elem_name)))
		i++;
		
	if(i >= strlen(elem_name)){
		// Can't find any letters in this element name assuming Carbon.
		elem = ELEM_C;

		//SetAtomColor(color_scheme);
		//SetAtomRadius(radius_scheme);
		//SetAtomMass(mass_scheme);

		return;
	}
	
	switch(elem_name[i]){
		case 'C': // Carbon, Calcium, Chlorine
			switch(elem_name[i+1]){
				case 'L':
					elem = ELEM_CL;
					break;
				// We can't use CA for calcium because of ambiguity
				// with the alpha carbon.
				//case 'A':
				//	elem = ELEM_CA;
				//	break;
				default:
					elem = ELEM_C;
					break;
			};
			
			break;
		case 'N': // Nitrogen, Sodium
			// NH is nitrogen (?)
			// Catch the "NH" notation for hydrogens
			//if(elem_name[i+1] == 'H')
			//	elem = ELEM_H;
			//else
			//	elem = ELEM_N;
			switch(elem_name[i+1]){
				case 'A':
					elem = ELEM_NA;
					break;
				default:
					elem = ELEM_N;
					break;
			};
			break;
		case 'O': // Oxygen
			elem = ELEM_O;
			break;
		case 'S': // Sulfur
			// SH is sulfur (?)
			// Catch the "SH" notation for hydrogens
			//if(elem_name[i+1] == 'H')	
			//	elem = ELEM_H;
			//else
			//	elem = ELEM_S;
			elem = ELEM_S;
			break;
		case 'H': // Hydrogen
			elem = ELEM_H;
			break;
		case 'P': // Phosphorous
			elem = ELEM_P;
			break;
		case 'Z': // Zinc
			elem = ELEM_Z;
			break;
		case 'F': // Iron
			elem = ELEM_FE;
			break;
		case 'M': // Magnesium
			elem = ELEM_MG;
			break;
		case 'K': // Potassium
			elem = ELEM_K;
			break;
		case 'I': // Iodine
			elem = ELEM_I;
			break;
		case 'D': // Dummy atom
			elem = ELEM_DUMMY;
			break;
		default: // Carbon
			// Unknown atom type, assuming carbon
			elem = ELEM_C;
			break;
	};
}

// Compute the dimensions of the molecule
void molecule::GetMolDim(int trajProcess)
{
 	float xmax, ymax, zmax;
	int i, j;
	// Need at least one atom!
	if(NumAtoms() <= 0)
		return;

	// Set the geometric center to zero
	//ZeroGCM(atoms);
	ZeroCM(atoms);
	
  	xmin = atoms[0].coord[X];
  	xmax = atoms[0].coord[X];
  	ymin = atoms[0].coord[Y];
  	ymax = atoms[0].coord[Y];
  	zmin = atoms[0].coord[Z];
  	zmax = atoms[0].coord[Z];
	
	for(i = 0;i < NumAtoms();i++){
    	xmin = MIN(atoms[i].coord[X], xmin);
    	xmax = MAX(atoms[i].coord[X], xmax);
    	ymin = MIN(atoms[i].coord[Y], ymin);
    	ymax = MAX(atoms[i].coord[Y], ymax);
    	zmin = MIN(atoms[i].coord[Z], zmin);
    	zmax = MAX(atoms[i].coord[Z], zmax);
  	}
 	
	// Remove the geometric center from any trajectory
	// associated with the molcule.
	
	// Allocate (or re-allocate) the center of mass vector
	cm.Alloc(3*steps);

	for(i = 0;i < 3*steps;i++)
		cm[i] = 0.0;
	
	if(trajProcess != QM_VERBATIM)
		//ZeroGCM(traj, steps);
		ZeroCM(traj, atoms, steps);
	
	if(trajProcess == QM_KABSCH){
		// Perform a Kabsch rotation
		array<float> w(NumAtoms());

		for(i = 0;i < NumAtoms();i++)
			w[i] = atoms[i].mass;

		for(i = 0;i < steps;i++){
			try{
				kabsch(atoms, traj[i], w);
			}
			catch(QmolError &e){
				CString linenum;
				linenum.Format(" @ step %d", i);

				AfxMessageBox(e.message() + linenum);
			}
		}
	}

 	for(i = 0;i < steps;i++)
		for(j = 0;j < NumAtoms();j++){
    		xmin = MIN(traj[i][3*j], xmin);
    		xmax = MAX(traj[i][3*j], xmax);
    		ymin = MIN(traj[i][3*j + 1], ymin);
    		ymax = MAX(traj[i][3*j + 1], ymax);
    		zmin = MIN(traj[i][3*j + 2], zmin);
    		zmax = MAX(traj[i][3*j + 2], zmax);
		}
	
  	xsize = xmax - xmin;
  	ysize = ymax - ymin;
  	zsize = zmax - zmin;
	
	maxdim = (float)(sqrt(xsize*xsize + ysize*ysize + zsize*zsize));
}

void molecule::GetCurrentMolDim(float &m_maxX, float &m_minX,
								float &m_maxY, float &m_minY,
								float &m_maxZ, float &m_minZ,
								BOOL use_mol_frame /* = FALSE */)
{
	int i;
	float r[3], tmp[3];
	int numAtoms = NumAtoms();

	// Need at least one atom!
	if(numAtoms <= 0){
		m_maxX = m_minX = m_maxY = m_minY = m_maxZ = m_minZ = 0.0f;
		return;
	}
	
	// Apply the current geometric transformation to the molecular
	// coordinates to express them in the world frame.
	float m[4][4];

	if(!use_mol_frame)
		BuildRotationMatrix(m);

	// Only compute the dimensions of the current trajecotry step
	if(current_step == 0){
		if(!use_mol_frame){
			tmp[X] = atoms[0].coord[X] + R[X];
			tmp[Y] = atoms[0].coord[Y] - R[Y];
			tmp[Z] = atoms[0].coord[Z] + R[Z];
			

			r[X] = m[X][X]*tmp[X] + m[Y][X]*tmp[Y] + m[Z][X]*tmp[Z];
			r[Y] = m[X][Y]*tmp[X] + m[Y][Y]*tmp[Y] + m[Z][Y]*tmp[Z];
			r[Z] = m[X][Z]*tmp[X] + m[Y][Z]*tmp[Y] + m[Z][Z]*tmp[Z];

			m_maxX = m_minX = r[X];
  			m_maxY = m_minY = r[Y];
  			m_maxZ = m_minZ = r[Z];
		}
		else{
			m_maxX = m_minX = atoms[0].coord[X];
  			m_maxY = m_minY = atoms[0].coord[Y];
  			m_maxZ = m_minZ = atoms[0].coord[Z];
		}


		
		for(i = 1;i < numAtoms;i++){
			float *ptr = atoms[i].coord;

			if(!use_mol_frame){
				tmp[X] = ptr[X] + R[X];
				tmp[Y] = ptr[Y] - R[Y];
				tmp[Z] = ptr[Z] + R[Z];
				

				r[X] = m[X][X]*tmp[X] + m[Y][X]*tmp[Y] + m[Z][X]*tmp[Z];
				r[Y] = m[X][Y]*tmp[X] + m[Y][Y]*tmp[Y] + m[Z][Y]*tmp[Z];
				r[Z] = m[X][Z]*tmp[X] + m[Y][Z]*tmp[Y] + m[Z][Z]*tmp[Z];

				m_maxX = MAX(r[X], m_maxX);
    			m_minX = MIN(r[X], m_minX);
				m_maxY = MAX(r[Y], m_maxY);
    			m_minY = MIN(r[Y], m_minY);
				m_maxZ = MAX(r[Z], m_maxZ);
    			m_minZ = MIN(r[Z], m_minZ);
			}
			else{
				m_maxX = MAX(ptr[X], m_maxX);
    			m_minX = MIN(ptr[X], m_minX);
				m_maxY = MAX(ptr[Y], m_maxY);
    			m_minY = MIN(ptr[Y], m_minY);
				m_maxZ = MAX(ptr[Z], m_maxZ);
    			m_minZ = MIN(ptr[Z], m_minZ);
			}
  		}
	}
	else{
		int trajIndex = current_step - 1;
		
		array<float> &trajRef = traj[trajIndex];

		if(!use_mol_frame){
			tmp[X] = trajRef[X] + R[X];
			tmp[Y] = trajRef[Y] - R[Y];
			tmp[Z] = trajRef[Z] + R[Z];
			

			r[X] = m[X][X]*tmp[X] + m[Y][X]*tmp[Y] + m[Z][X]*tmp[Z];
			r[Y] = m[X][Y]*tmp[X] + m[Y][Y]*tmp[Y] + m[Z][Y]*tmp[Z];
			r[Z] = m[X][Z]*tmp[X] + m[Y][Z]*tmp[Y] + m[Z][Z]*tmp[Z];

			m_maxX = m_minX = r[X];
  			m_maxY = m_minY = r[Y];
  			m_maxZ = m_minZ = r[Z];
		}
		else{
			m_maxX = m_minX = trajRef[X];
  			m_maxY = m_minY = trajRef[Y];
  			m_maxZ = m_minZ = trajRef[Z];
		}

		for(i = 1;i < numAtoms;i++){
			float *ptr = &(trajRef[3*i]);
			if(!use_mol_frame){
				tmp[X] = ptr[X] + R[X];
				tmp[Y] = ptr[Y] - R[Y];
				tmp[Z] = ptr[Z] + R[Z];
				

				r[X] = m[X][X]*tmp[X] + m[Y][X]*tmp[Y] + m[Z][X]*tmp[Z];
				r[Y] = m[X][Y]*tmp[X] + m[Y][Y]*tmp[Y] + m[Z][Y]*tmp[Z];
				r[Z] = m[X][Z]*tmp[X] + m[Y][Z]*tmp[Y] + m[Z][Z]*tmp[Z];

				m_maxX = MAX(r[X], m_maxX);
    			m_minX = MIN(r[X], m_minX);
				m_maxY = MAX(r[Y], m_maxY);
    			m_minY = MIN(r[Y], m_minY);
				m_maxZ = MAX(r[Z], m_maxZ);
    			m_minZ = MIN(r[Z], m_minZ);
			}
			else{
				m_maxX = MAX(ptr[X], m_maxX);
    			m_minX = MIN(ptr[X], m_minX);
				m_maxY = MAX(ptr[Y], m_maxY);
    			m_minY = MIN(ptr[Y], m_minY);
				m_maxZ = MAX(ptr[Z], m_maxZ);
    			m_minZ = MIN(ptr[Z], m_minZ);
			}
  		}
	}
}

void molecule::ReadCoord(const char *coordName, BOOL m_multi_model_traj, 
						 BOOL m_connect_on_read, int file_type)
{

	if(file_type == QM_AUTO)
		file_type = AutoDetectFileType(coordName);

	switch (file_type) {
		case QM_PDB:
			{
				// Grab a pointer to the status bar on the window frame -- we'll
				// use it to keep the user informed for file reading progress.
				CMainFrame *pFrame = (CMainFrame*)AfxGetApp()->m_pMainWnd;
				CStatusBar* pStatus = &pFrame->m_wndStatusBar;

				// Only reads PDB files for now!
				pdb_file pdb;
	
				if(pStatus)
					pStatus->SetPaneText(0, "Reading pdb file ...");

				if(!pdb.Read(coordName)){
					CString buffer;
					buffer.Format("Can't open ``%s'' for reading.", coordName);
					throw QMOL_ERROR(buffer);
				}
				else{

					if(pStatus)
						pStatus->SetPaneText(0, "Processing data ...");

					ReadPdb(pdb, m_multi_model_traj, m_connect_on_read);
			
					SetAtomicRadiiScheme(QM_RADIUS_DEFAULT, 
						0,		// Value not used
						0,		// Value not used
						0.0f);	// Value not used

					// First, clear any monitors
					while(!mon.IsEmpty())
						mon.Pop();

					// Set atom monitors (if any)
					for(int i = 0;i < pdb.GetNumMonitors();i++){
						monitor tmp;
						
						tmp.pairId = pdb.GetMonitor(i).pairId;
						tmp.atom_id[0] = pdb.GetMonitor(i).atom_index[0];
						tmp.atom_id[1] = pdb.GetMonitor(i).atom_index[1];

						tmp.color[0] = pdb.GetMonitor(i).color[0];
						tmp.color[1] = pdb.GetMonitor(i).color[1];
						tmp.color[2] = pdb.GetMonitor(i).color[2];

						mon.Push(tmp);
					}

					// Set the maximum temperature and occupancy
					max_temp = (float)(fabs(pdb.GetMaxTemp()));
					min_temp = (float)(fabs(pdb.GetMinTemp()));

					max_occ = (float)(fabs(pdb.GetMaxOcc()));
					min_occ = (float)(fabs(pdb.GetMinOcc()));
				}
			}
			break;
		case QM_XYZ:
			// Not supported yet
			throw QMOL_ERROR("XYZ files aren't supported yet!");
			break;
		case QM_CRD:
			// Not supported yet
			throw QMOL_ERROR("CRD files aren't supported yet!");
			return;
			break;
		default:
			throw QMOL_ERROR("Unknown file type!");
			return;
	};
}

void molecule::SaveCoord(FILE* fout, int file_type, int precision,
						 BOOL write_conect, BOOL normalize_T, int offset /* = 0*/)
{	
	switch (file_type) {
		case QM_PDB:
			// Only reads PDB files for now!
			SavePdb(fout, precision, write_conect, normalize_T, offset);
			break;
		case QM_XYZ:
			// Not supported yet
			throw QMOL_ERROR("XYZ files aren't supported yet!");
			break;
		case QM_CRD:
			// Not supported yet
			throw QMOL_ERROR("XYZ files aren't supported yet!");
			break;
		default:
			throw QMOL_ERROR("Unknown file type!");
	};
}

// Read in pdb files.
void molecule::ReadPdb(pdb_file &pdb, BOOL m_multi_model_traj, BOOL m_connect_on_read)
{
	atom a;
	int i, j;	
	bond temp;
	
	int chain_model_count = 1;

	int curr_model;
	char curr_chain;
	int num_pdb_atoms;
	int num_model;

	// How many atoms are in the pdb file? If we are interpreting multiple
	// models as a trajectory, only count the atoms in the primary sturcture.
	if(m_multi_model_traj && pdb.GetNumRecords()){

		// Count the number of models in the pdb file
		curr_model = pdb.GetModel(0);
		num_model = 1;
		for(i = 1;i < pdb.GetNumRecords();i++){
			if(curr_model != pdb.GetModel(i)){
				num_model ++;
				curr_model = pdb.GetModel(i);
			}
		}
		
		// Make sure that all models contain the same number of atoms
		if(pdb.GetNumRecords() % num_model != 0){
			AfxMessageBox("All models in PDB file must contain the same "
				"number of atoms");
			num_pdb_atoms = pdb.GetNumRecords();
		}
		else{
			num_pdb_atoms = pdb.GetNumRecords() / num_model;
		}
	}
	else{
		num_pdb_atoms = pdb.GetNumRecords();
	}

	// ``Alloc'' deletes any previously allocated memoery 
	atoms.Alloc(num_pdb_atoms);

	if(num_pdb_atoms){
		curr_model = pdb.GetModel(0);
		curr_chain = pdb.GetChainId(0);
	}

	for(i = 0;i < num_pdb_atoms;i++){
		// Read the atomic coordinates
		pdb.GetCoord(i, atoms[i].coord[X], 
			atoms[i].coord[Y], atoms[i].coord[Z]);
		atoms[i].id = pdb.GetSerial(i);
		atoms[i].temperature = pdb.GetTemp(i);
		atoms[i].occupancy = pdb.GetOcc(i);
		//atoms[i].GetElement(pdb.GetAtomName(i), color_scheme,
		//	elem_radii_scheme, elem_mass_scheme);
		atoms[i].GetElement(pdb.GetAtomName(i));

		atoms[i].resSeq = pdb.GetResNum(i);
		
		atoms[i].isSolvent = pdb.IsWater(i);
		
		atoms[i].resName = new char [strlen(pdb.GetResName(i)) + 1];
		strcpy(atoms[i].resName, pdb.GetResName(i));

		// Save the model number (needed to save a pdb file)
		atoms[i].model = pdb.GetModel(i);

		// Save the chain id
		atoms[i].chain = pdb.GetChainId(i);

		if(!atoms[i].isSolvent && 
		 ((curr_model != atoms[i].model) || (curr_chain != atoms[i].chain))){
			curr_model = atoms[i].model;
			curr_chain = atoms[i].chain;
			chain_model_count ++;
		}
	}

	if(m_multi_model_traj && (num_model > 1)){
		// Set the trajectory data
		array<float> traj_buffer(num_pdb_atoms*3);

		// Sequentially pack each model (except for the first)
		// into the trajectory list
		for(i = 1;i < num_model;i++){
			for(j = 0;j < num_pdb_atoms;j++){
				float tmpX, tmpY, tmpZ;

				pdb.GetCoord(i*num_pdb_atoms + j, tmpX, tmpY, tmpZ);

				traj_buffer[3*j + X] = tmpX;
				traj_buffer[3*j + Y] = tmpY;
				traj_buffer[3*j + Z] = tmpZ;
			}
			
			traj.Push(traj_buffer);
		}

		// The number of frames in the trajectory (total number of models
		// minus the primary structure).
		steps = num_model - 1;
	}

	// Prepare the secondary structure storage
	secondaryStruc.Alloc(chain_model_count);

	int res_count = 1;
	int curr_res = 0;

	chain_model_count = 0;

	if(atoms.Length()){
		curr_model = atoms[0].model;
		curr_chain = atoms[0].chain;
		curr_res = atoms[0].resSeq;
	}

	for(i = 1;i < atoms.Length();i++){
		if(!atoms[i].isSolvent){
			if((curr_model != atoms[i].model) || (curr_chain != atoms[i].chain)){
				curr_model = atoms[i].model;
				curr_chain = atoms[i].chain;
				secondaryStruc[chain_model_count].sequence.Alloc(res_count);
				res_count = 1;
				curr_res = atoms[i].resSeq;
				chain_model_count ++;
			}
			else{
				if(curr_res != atoms[i].resSeq){
					res_count ++;
					curr_res = atoms[i].resSeq;
				}
			}
		}
	}

	if(chain_model_count != secondaryStruc.Length())
		secondaryStruc[chain_model_count].sequence.Alloc(res_count);

	// Start at the head of the secondaryStruc array
	chain_model_count = 0;

	if(atoms.Length()){
		curr_model = atoms[0].model;
		curr_chain = atoms[0].chain;

		secondaryStruc[chain_model_count].model = curr_model;
		secondaryStruc[chain_model_count].chain = curr_chain;

		curr_res = atoms[0].resSeq;
		res_count = 0;
	}

	for(i = 0;i < atoms.Length();i++){
		if(!atoms[i].isSolvent){
			if((curr_model != atoms[i].model) || (curr_chain != atoms[i].chain)){
				chain_model_count ++;

				curr_model = atoms[i].model;
				curr_chain = atoms[i].chain;
				curr_res = atoms[i].resSeq;
				res_count = 0;

				secondaryStruc[chain_model_count].model = curr_model;
				secondaryStruc[chain_model_count].chain = curr_chain;
			}
			else{
				if(curr_res != atoms[i].resSeq){
					res_count ++;
					curr_res = atoms[i].resSeq;
				}
				
				secondaryStruc[chain_model_count].sequence[res_count].atom_id.Push(i);
				secondaryStruc[chain_model_count].sequence[res_count].
					SetAminoAcidType(atoms[i].resName);
				secondaryStruc[chain_model_count].sequence[res_count].
					SetCAlpha(atoms[i].name, i);
				secondaryStruc[chain_model_count].sequence[res_count].
					SetO(atoms[i].name, i);
				secondaryStruc[chain_model_count].sequence[res_count].
					resSeq = atoms[i].resSeq;
			}
		}
	}

	// Verify that all sequences are valid (i.e. we have found an alpha carbon
	// for each residue).
	for(i = 0;i < secondaryStruc.Length();i++){
		array<amino> &sequence_ref = secondaryStruc[i].sequence;
		int bad_res = 0;

		for(j = 0;j < sequence_ref.Length();j++){
			if((sequence_ref[j].c_alpha_index == -1) ||
			   (sequence_ref[j].o_index == -1)){
				// Invalid residue!
				bad_res ++;
			}
		}

		if(bad_res != 0){
			// We have a bad residue, so repack the sequence array
			array<amino> tmp_sequence(sequence_ref.Length() - bad_res);
			int curr_res = 0;

			for(j = 0;j < sequence_ref.Length();j++){
				if((sequence_ref[j].c_alpha_index != -1) &&
				   (sequence_ref[j].o_index != -1)){
					tmp_sequence[curr_res] = sequence_ref[j];
					curr_res ++;
				}
			}

			sequence_ref = tmp_sequence;
		}
	}

	////////////////////////////////////////////////////////////////////
	// The use_serial_for_connect tells the pdb file to either use the 
	// atom serial number when constructing the connectivity tables (i.e. TRUE)
	// or to use the atom records (instead of serial numbers, i.e. FALSE).
	// The default value is FALSE to allow multiple model structures to re-use
	// atom numbers for each model.
	int use_serial_for_connect = FALSE;
	if(m_connect_on_read)
		// If the PDB file already contains connectivity info, we'll
		// use it, otherwise, we need to compute the molecular connectivity
		// ourselves. The connect_on_read variable allows the user
		// to disable the connectivity calculation.
		use_serial_for_connect = pdb.UpdateConnect(use_serial_for_connect);

	// Store the molecular connectivity (in terms of record
	// numbers, not serial numbers!).
	mol_connect.Alloc(pdb.GetNumConnections());
	for(i = 0;i < pdb.GetNumConnections();i++){
		mol_connect[i].Alloc(pdb.GetConnect(i).GetNumConnect());
		
		if(use_serial_for_connect){
			for(j = 0;j < pdb.GetConnect(i).GetNumConnect();j++)
				mol_connect[i][j] = pdb.SerialToRecord(pdb.GetConnect(i)[j]);
		}
		else{ // use_serial_for_connect == FALSE
			for(j = 0;j < pdb.GetConnect(i).GetNumConnect();j++)
				mol_connect[i][j] = pdb.GetConnect(i)[j];
		}
	}	
	
	int temp_serial, temp_record;

	// Compute the number of pdb bonds
	int num_pdb_bonds;

	if(m_multi_model_traj){
		// Count the number of bonds in the primary structure

		num_pdb_bonds = 0;

		int temp_curr_model = pdb.GetModel(0);

		for(i = 0;i < pdb.GetNumConnections();i++){
			if(use_serial_for_connect){
				temp_serial = pdb.GetConnect(i).GetNode();		
				temp_record = pdb.SerialToRecord(temp_serial);
			}
			else{
				temp_record = pdb.GetConnect(i).GetNode();
			}
			
			for(j = 0;j < pdb.GetConnect(i).GetNumConnect();j++){	
				temp.b = pdb.GetConnect(i)[j];
				
				if(use_serial_for_connect){
					if(temp_serial < temp.b){
						// Save record numbers in the bond stack
						temp.a = temp_record;
						temp.b = pdb.SerialToRecord(temp.b);
						
						if((pdb.GetModel(temp.a) == temp_curr_model) &&
						   (pdb.GetModel(temp.b) == temp_curr_model)){

							// Count this bond
							num_pdb_bonds ++;
						}
					}
				}
				else{ // use_serial_for_connect == FALSE
					if(temp_record < temp.b){
						// Save record numbers in the bond stack
						temp.a = temp_record;
						
						if((pdb.GetModel(temp.a) == temp_curr_model) &&
						   (pdb.GetModel(temp.b) == temp_curr_model)){

							// Count this bond
							num_pdb_bonds ++;
						}
					}
				}
			}
		}
	}
	else{
		// The number of bonds is determined by the PDB file
		num_pdb_bonds = pdb.GetNumBonds();
	}

	// Now that we know the number of bonds, allocate space for them
	bonds.Allocate(num_pdb_bonds);

	for(i = 0;i < pdb.GetNumConnections();i++){
		if(use_serial_for_connect){
			temp_serial = pdb.GetConnect(i).GetNode();		
			temp_record = pdb.SerialToRecord(temp_serial);
		}
		else{
			temp_record = pdb.GetConnect(i).GetNode();
		}
		
		for(j = 0;j < pdb.GetConnect(i).GetNumConnect();j++){
			// Make sure that we don't overrun the bond stack when
			// multi_model_traj is enabled
			if(m_multi_model_traj && (bonds.GetLength() == num_pdb_bonds)){
				break;
			}

			temp.b = pdb.GetConnect(i)[j];
			
			if(use_serial_for_connect){
				if(temp_serial < temp.b){
					// Save record numbers in the bond stack
					temp.a = temp_record;
					temp.b = pdb.SerialToRecord(temp.b);
					
					if(pdb.IsWater(temp.a) || pdb.IsWater(temp.b))
						temp.solvent = TRUE;
					else
						temp.solvent = FALSE;
					
					bonds.Push(temp);
				}
			}
			else{ // use_serial_for_connect == FALSE
				if(temp_record < temp.b){
					// Save record numbers in the bond stack
					temp.a = temp_record;
					
					if(pdb.IsWater(temp.a) || pdb.IsWater(temp.b))
						temp.solvent = TRUE;
					else
						temp.solvent = FALSE;
						
					bonds.Push(temp);
				}
			}
		}

		if(j == 0){
			// An unconnected atom - test to see if it is a solvent atom
			// (i.e. a lone oxygen atom).
			if((pdb.GetAtomName(temp_record)[0] == 'O') ||
				(strncmp(pdb.GetAtomName(temp_record), "HOH", 3) == 'O')){
				atoms[temp_record].isSolvent = TRUE;
			}
		}

		// Make sure that we don't overrun the bond stack when
		// multi_model_traj is enabled
		if(m_multi_model_traj && (bonds.GetLength() == num_pdb_bonds)){
			break;
		}
	}
}

void molecule::UpdateBondColoring()
{
	// Is a given bond between atoms of the same or different colors?
	for(int i = 0;i < bonds.GetLength();i++)
		if((atoms[bonds[i].a].rgb_color[0] == atoms[bonds[i].b].rgb_color[0]) &&
		   (atoms[bonds[i].a].rgb_color[1] == atoms[bonds[i].b].rgb_color[1]) &&
		   (atoms[bonds[i].a].rgb_color[2] == atoms[bonds[i].b].rgb_color[2]) &&
		   (atoms[bonds[i].a].rgb_color[3] == atoms[bonds[i].b].rgb_color[3]))
			bonds[i].colorBond = FALSE;
		else
			bonds[i].colorBond = TRUE;
}

void molecule::SavePdb(FILE *fout, int precision, BOOL write_conect, 
					   BOOL normalize_T, int offset)
{
	int i;
	
	pdb_file pdb(atoms.Length());

	float x, y, z;
	float tmp_x, tmp_y, tmp_z;
	
	float m[4][4];

	BuildRotationMatrix(m);

	for(i = 0;i < atoms.Length();i++){

		// Save the molecule in the orientation specified by
		// the rotation matrix "m" and the translation vector
		// "R".
		x = atoms[i].coord[X] + R[X];
		y = atoms[i].coord[Y] - R[Y];
		z = atoms[i].coord[Z] + R[Z];

		tmp_x = m[X][X]*x + m[Y][X]*y + m[Z][X]*z;
		tmp_y = m[X][Y]*x + m[Y][Y]*y + m[Z][Y]*z;
		tmp_z = m[X][Z]*x + m[Y][Z]*y + m[Z][Z]*z;

		//pdb.SetCoord(i, atoms[i].coord[X], atoms[i].coord[Y], 
		//	atoms[i].coord[Z]);
		pdb.SetCoord(i, tmp_x, tmp_y, tmp_z);

		// Maintain the same numbering system used in the pdb file
		// (with the addition of an offset to handle multiple molecules).
		pdb.SetSerial(i, atoms[i].id + offset);
		
		if(normalize_T){
			// Write out normalized temperatures (to prevent overflow of the
			// temperature field -- which has limited precision). Normalize temperature
			// values to be between 0.01 and 99.9.
			pdb.SetTemp(i, 0.01f + 99.8f*
				NormalizeByRange(atoms[i].temperature, min_temp, max_temp));
		}
		else{
			// Clamp the maximum T at 99.9 to prevent precision overflow
			pdb.SetTemp(i, min(99.9f, atoms[i].temperature));
		}

		pdb.SetOcc(i, atoms[i].occupancy);
		pdb.SetResNum(i, atoms[i].resSeq);

		pdb.SetAtomName(i, atoms[i].name);
		pdb.SetResName(i, atoms[i].resName);

		// Set the model number to correctly save multiple structures in
		// the same pdb file.
		pdb.SetModel(i, atoms[i].model);

		pdb.SetChainId(i, atoms[i].chain);
	}
	
	// Make sure that we only write the connection data for 
	// the first model (if multiple models are present).
	int first_model;

	if(atoms.Length() > 0){
		// Save the model number of the first bond.
		first_model = atoms[0].model;
	}
	else{
		first_model = -1;
	}

	// Count the number of unique connection entries
	int num_connect = 0;

	for(i = 0;i < atoms.Length();i++){
		if((first_model == -1) || (first_model == atoms[i].model)){
			num_connect ++;
		}
	}

	// Construct a set of connect entries from the set
	// of atomic bonds
	pdb.ResetConnect(num_connect);

	for(i = 0;i < bonds.GetLength();i++){
		if((first_model == -1) || (first_model == atoms[bonds[i].a].model)){
			// Maintain the same numbering system used in the pdb file
			pdb.PushConnect(bonds[i].a, pdb.GetSerial(bonds[i].b));
			pdb.PushConnect(bonds[i].b, pdb.GetSerial(bonds[i].a));
		}
	}
	
	if(!pdb.Write(fout, precision, write_conect))
		throw QMOL_ERROR("Error saving pdb file, " + CString(coordName));
}

void molecule::ReadTrajHeader(const char *trajName, int file_type)
{
	
	// Initialize the trajectory vector by setting its size
	// to zero.
	while(!traj.IsEmpty())
		traj.Pop();
	
	switch(file_type){
		case QM_DCD:
			ReadDCDHeader(trajName);
			break;
		case QM_PTF:
			throw QMOL_ERROR("PTF files are not currently supported!");
			return;
		default:
			throw QMOL_ERROR("Unknown file type!");
			return;
	}
	
	if(steps == 0)
		throw QMOL_ERROR("The specified trajectory has zero steps!");

}

void molecule::ReadDCDHeader(const char *trajName)
{
	dcd dcdData;
	int npt = 0;
	
	// Open the DCD file for reading
	dcdData.Open(trajName, DCD_READ);
			
	if(dcdData.Error()){
		throw QMOL_ERROR(dcdData.Error());
		return;
	}
	
	dcdData.read_dcd_header(npt, steps, size);
			
	if(dcdData.Error()){
		throw QMOL_ERROR(dcdData.Error());
		return;
	}
	
	if(npt != atoms.Length()){
		throw QMOL_ERROR("Incorrect Number of atoms in DCD file!");
		return;
	}
	
	// The allowed data sizes
	if((size != sizeof(float)) && 
	   (size != sizeof(double)) &&
	   (size != sizeof(long double))){
		throw QMOL_ERROR("Incompatible data size in DCD file!");
		return;
	}
}

void molecule::ReadTrajStep(const char* trajName, int file_type)
{
	switch(file_type){
		case QM_DCD:
			ReadDCDStep(trajName);
			break;
		case QM_PTF:
			throw QMOL_ERROR("PTF files are not currently supported!");
			return;
		default:
			throw QMOL_ERROR("Unknown file type!");
	}
}

void molecule::ReadDCDStep(const char *trajName)
{
	static dcd dcdData;
	int dummy;
	void* temp = NULL;
	static int i;
	int j;
	int npt = atoms.Length();
	
	// Thing to do once, before the first step
	// in read.
	if(traj.Length() == 0){
		// Read the header to bring the file pointer to 
		// the start of the data.
		dcdData.Open(trajName, DCD_READ);
		dcdData.read_dcd_header(dummy, dummy, dummy);
		
		// Allocate memory for the trajectory
		// traj.Alloc(steps*npt*3);
		// This is now dead code - memory will be dynamically allocated!
		
		// Reset the step counter
		i = 0;
	}	
	
	// Will we need to convert the data from a non-standard data format?
	if(size != sizeof(float))
		// Allocate some memory to hold the conversion buffer.
		temp = new char [3*npt*size];

	array<float> traj_buffer(npt*3);

	// Read a step ...
	switch(size){
		case sizeof(float):
			dcdData.read_dcd(traj_buffer.Pointer(), npt, size);
			break;
		case sizeof(double):
			dcdData.read_dcd(temp, npt, size);
					
			// Convert the data
			for(j = 0;j < 3*npt;j++)
					traj_buffer[j] = (float)(((double*)temp)[j]);
				break;
#ifdef PORT_WIN32
		case sizeof(long double):
			dcdData.read_dcd(temp, npt, size);
					
			// Convert the data
			for(j = 0;j < 3*npt;j++)
				traj_buffer[j] = ((long double*)temp)[j];
			break;
#endif // PORT_WIN32
	};
	
	if(temp)
		// Remove the temporary storage space
		// if we have to.
		delete [] temp;
	
	// Push the trajectory data on to the trajectory list
	traj.Push(traj_buffer);

	// Increment the step counter
	i++;
	
	if(dcdData.Error())
		throw QMOL_ERROR(dcdData.Error());	
};

void molecule::TrajExportPdb(const char *file_name, int start, int stop)
{
	int i, j;

	if(traj.Length() == 0){
		AfxMessageBox("Cannot export a zero length trajectory");
		return;
	}

	// Perform some sanity checking
	if((start > stop) || (start < 1)){
		AfxMessageBox("Invalid starting step");
		return;
	}

	if(stop > traj.Length()){
		AfxMessageBox("Invalid stopping step");
		return;
	}

	// Check to see if multiple models are present
	for(i = 1;i < NumAtoms();i++){
		if(atoms[i].model != atoms[i-1].model){
			AfxMessageBox("Unable to export a trajectory for a multi-model data set.");
			return;
		}
	}

	// Convert start and stop to "C" style array indicies
	start -= 1;
	stop -= 1;

	int num_steps = stop - start + 1;
	int num_atoms = atoms.Length();

	// Create the pdb file
	pdb_file pdb(num_atoms*num_steps);

	for(j = 0;j < num_steps;j++){
		for(i = 0;i < atoms.Length();i++){
			pdb.SetCoord(i + num_atoms*j, 
				traj[j + start][3*i + X], 
				traj[j + start][3*i + Y], 
				traj[j + start][3*i + Z]);

			// Maintain the same numbering system used in the pdb file
			pdb.SetSerial(i + num_atoms*j, atoms[i].id);

			pdb.SetTemp(i + num_atoms*j, atoms[i].temperature);

			// Set the occupancy to the model number to allow visual 
			// discrimination betweent the different models
			//pdb.SetOcc(i + num_atoms*j, atoms[i].occupancy);
			pdb.SetOcc(i + num_atoms*j, (double)(j)/num_steps);

			pdb.SetResNum(i + num_atoms*j, atoms[i].resSeq);

			pdb.SetAtomName(i + num_atoms*j, atoms[i].name);
			pdb.SetResName(i + num_atoms*j, atoms[i].resName);

			// Set the model number to correctly save multiple structures in
			// the same pdb file.
			pdb.SetModel(i + num_atoms*j, j + 1);

			pdb.SetChainId(i + num_atoms*j, atoms[i].chain);
		}
	}

	if(!pdb.Write(file_name, 3, FALSE))
		AfxMessageBox("Unable to save " + CString(file_name));

}
void molecule::ZeroGCM(array<atom> &r)
{
	float xcm, ycm, zcm;

	if(r.Length() != 0){
		// Set the geometric center to zero
		xcm = ycm = zcm = 0.0;
		
		int i;
	
		for(i = 0;i < r.Length();i++){
			xcm += r[i].coord[X];
			ycm += r[i].coord[Y];
			zcm += r[i].coord[Z];
		}
	
		xcm /= r.Length();
		ycm /= r.Length();
		zcm /= r.Length();
		
		for(i = 0;i < r.Length();i++){
			r[i].coord[X] -= xcm;
			r[i].coord[Y] -= ycm;
			r[i].coord[Z] -= zcm;
		}
	}
}

void molecule::ZeroCM(array<atom> &r)
{
	float xcm, ycm, zcm;

	if(r.Length() != 0){
		float total_mass = 0.0;

		// Set the center to zero
		xcm = ycm = zcm = 0.0;
		
		int i;
	
		for(i = 0;i < r.Length();i++){
			total_mass += r[i].mass;

			xcm += r[i].coord[X]*r[i].mass;
			ycm += r[i].coord[Y]*r[i].mass;
			zcm += r[i].coord[Z]*r[i].mass;
		}

		xcm /= total_mass;
		ycm /= total_mass;
		zcm /= total_mass;
		
		for(i = 0;i < r.Length();i++){
			r[i].coord[X] -= xcm;
			r[i].coord[Y] -= ycm;
			r[i].coord[Z] -= zcm;
		}
		
#ifdef PORT_WIN32
		// Adjust the bounding box (if present)
		if(molState->bbox){
			molState->bbox_R[0] -= xcm;
			molState->bbox_R[1] -= ycm;
			molState->bbox_R[2] -= zcm;
		}
#endif // PORT_WIN32
	}
}

void molecule::ZeroGCM(list< array<float> > &r, int steps)
{	
	if(r.Length() > 0){
		if(!steps)
			throw QMOL_ERROR("steps == 0");
		
		int i, j;
		
		// Look at the first structure on the list
		// and compute the number of atoms.
		int npt = r[0].Length()/3;
		
		// Set the geometric center of each structure to zero
		for(j = 0;j < steps;j++){
			for(i = 0;i < npt;i++){
				cm[3*j] += r[j][3*i];
				cm[3*j + 1] += r[j][3*i + 1];
				cm[3*j + 2] += r[j][3*i + 2];
			}
			
			cm[3*j] /= npt;
			cm[3*j + 1] /= npt;
			cm[3*j + 2] /= npt;
			
			for(i = 0;i < npt;i++){
				r[j][3*i] -= cm[3*j];
				r[j][3*i + 1] -= cm[3*j + 1];
				r[j][3*i + 2] -= cm[3*j + 2];
			}
		}
	}
}

void molecule::ZeroCM(list< array<float> > &r, array<atom> &ref, int steps)
{	
	if(r.Length() > 0){
		if(!steps)
			throw QMOL_ERROR("steps == 0");
		
		int i, j;
		
		float total_mass = 0.0;

		// Look at the first structure on the list
		// and compute the number of atoms.
		int npt = r[0].Length()/3;
		
		
		// Set the geometric center of each structure to zero
		for(j = 0;j < steps;j++){
			total_mass = 0.0;
			
			for(i = 0;i < npt;i++){
				total_mass += ref[i].mass;
				
				cm[3*j] += r[j][3*i]*ref[i].mass;
				cm[3*j + 1] += r[j][3*i + 1]*ref[i].mass;
				cm[3*j + 2] += r[j][3*i + 2]*ref[i].mass;
			}
			
			cm[3*j] /= total_mass;
			cm[3*j + 1] /= total_mass;
			cm[3*j + 2] /= total_mass;
			
			for(i = 0;i < npt;i++){
				r[j][3*i] -= cm[3*j];
				r[j][3*i + 1] -= cm[3*j + 1];
				r[j][3*i + 2] -= cm[3*j + 2];
			}
		}
	}
}

// Use the file extension to detect the type of file
int molecule::AutoDetectFileType(const char *name)
{
	char *ptr;
	
	// Is there a file extension?
	if((ptr = strrchr(name,'.')) == NULL)
		return QM_UNK;
	
	unsigned int i;
	char *buffer = new char [strlen(ptr)];
	
	// Drop the '.'
	ptr ++;
	
	strcpy(buffer, ptr);
	
	for(i = 0;i < strlen(buffer);i++)
		buffer[i] = tolower(buffer[i]);
	
	// Coordinate files
	if(strcmp("pdb", buffer) == 0){
		delete [] buffer;
		return QM_PDB;
	}
	
	if(strcmp("xyz", buffer) == 0){
		delete [] buffer;
		return QM_XYZ;
	}
	
	if(strcmp("crd", buffer) == 0){
		delete [] buffer;
		return QM_CRD;
	}
	
	// Trajectory files
	if(strcmp("dcd", buffer) == 0){
		delete [] buffer;
		return QM_DCD;
	}
	
	if(strcmp("ptf", buffer) == 0){
		delete [] buffer;
		return QM_PTF;
	}
	
	delete [] buffer;
	
	return QM_UNK;
}

int molecule::IdToIndex(int id)
{
	for(int i = 0;i < NumAtoms();i++)
		if(atoms[i].id == id)
			return i;
	
	// Couldn't find the requested id!
	return -1;
}

int molecule::ResIdToIndex(int id, int &chain_index)
{
	// Do we have a protein?
	if(secondaryStruc.Length() == 0)
		return -1;

	for(int i = 0;i < secondaryStruc.Length();i++){
		array<amino> &sequence_ref = secondaryStruc[i].sequence;

		for(int j = 0;j < sequence_ref.Length();j++){
			if(sequence_ref[j].resSeq == id){
				chain_index = i;
				return j;
			}
		}
	}

	// Couldn't find the requested id!
	return -1;
}

// Compute the number of amino acids in the molecule and the first valid
// residue.
int molecule::NumResidues(int &m_first)
{
	if(NumAtoms() == 0)
		return 0;

	AtomHash *resArray = new AtomHash [NumAtoms()];
	int i;

	for(i = 0;i < NumAtoms();i++){
		resArray[i].value = atoms[i].resSeq;
		resArray[i].i = i;
		//resArray[i].model = atoms[i].model;
		resArray[i].model = 1; // Can't handle multiple models ... yet.
	}

	// Sort the array of residue numbers
	qsort(resArray, NumAtoms(), sizeof(AtomHash), CompareAtomHash);

	BOOL found_first = FALSE;

	int cur = -1;
	int count = 0;

	for(i = 0;i < NumAtoms();i++){
		if((cur != resArray[i].value) && (strcmpi(atoms[resArray[i].i].name, "CA") == 0)){
			count ++;
			cur = resArray[i].value;

			if(!found_first){
				found_first = TRUE;
				m_first = cur;
			}
		}
	}

	delete [] resArray;

	return count;
}

void molecule::SetResidueColor(int index, int chain_index, 
					char R, char G, char B, char A)
{
	if(index < 0)
		throw QmolError("Invalid index", __FILE__, __LINE__);

	if((chain_index < 0) || (chain_index >= secondaryStruc.Length()))
		throw QmolError("Invalid chain index", __FILE__, __LINE__);

	// For now, set all residues that match the given residue number
	int res_id = secondaryStruc[chain_index].sequence[index].resSeq;

	for(int i = 0;i < secondaryStruc.Length();i++){
		array<amino> &amino_ref = secondaryStruc[i].sequence;

		for(int j = 0;j < amino_ref.Length();j++){
			if(amino_ref[j].resSeq == res_id){
				list<int> &atom_id_ref = amino_ref[j].atom_id;

				for(int k = 0;k < atom_id_ref.Length();k++){
					atoms[atom_id_ref[k]].rgb_color[0] = R;
					atoms[atom_id_ref[k]].rgb_color[1] = G;
					atoms[atom_id_ref[k]].rgb_color[2] = B;
					atoms[atom_id_ref[k]].rgb_color[3] = A;
				}
			}
		}
	}
}

BOOL molecule::ComputeSecondaryStructure()
{
	int i;
	int cur_model, first_chain, num_chain;
	int num_model_chain = secondaryStruc.Length();

	if(num_model_chain == 0)
		return TRUE;

	cur_model = secondaryStruc[0].model;
	first_chain = 0;
	num_chain = 1;

	try{
		// Compute the secondary structure for each model and chain
		for(i = 1;i < num_model_chain;i++){
			// Different models do not interact with each other, but
			// different chains within the same model do interact with
			// each other.
			if(cur_model != secondaryStruc[i].model){
				StrideWrapper(first_chain, num_chain);
				first_chain = i;
				num_chain = 1;
				cur_model = secondaryStruc[i].model;
			}
			else{
				num_chain ++;
			}
		}

		StrideWrapper(first_chain, num_chain);
	}
	catch(...){
		// The STRIDE code blew its cork! There are so many fixed memory 
		// data structures in the STRIDE code that it is fairly easy to 
		// run past the end of an array. The pdb file 1sva (an entire virus
		// coat) provides an example of this problem.
		AfxMessageBox("STRIDE is unable to assign the secondary structure!");
		return FALSE;
	}

	return TRUE;
}

BOOL molecule::StrideWrapper(int first_chain, int NChain)
{
	BOOL return_value = TRUE;

	CHAIN **Chain = NULL;
	HBOND **HBond = NULL;
	DONOR **Dnr = NULL;
	ACCEPTOR **Acc = NULL;
	RESIDUE *r;
	COMMAND *Cmd;
	int Cn, NHBond = 0, ValidChain = 0;
	float **PhiPsiMapHelix = NULL, 
		  **PhiPsiMapSheet = NULL;
	int i, j, k;

	// The total number of residues
	int num_residues = 0;

	// Count the total number of residues
	for(i = 0;i < NChain;i++){
		array<amino> &sequence_ref = secondaryStruc[i + first_chain].sequence;
		
		num_residues += sequence_ref.Length();
	}

	// Dynamically allocate memory for the required number of
	// chains.
	Chain = (CHAIN  **)ckalloc(NChain*sizeof(CHAIN *));
	//Chain = (CHAIN  **)ckalloc(MAX_CHAIN*sizeof(CHAIN *));

	// Dynamically allocate memory for the required number of
	// chains.
	//HBond = (HBOND  **)ckalloc(MAXHYDRBOND*sizeof(HBOND *));
	HBond = (HBOND  **)ckalloc(MAXHYDRBOND(num_residues)*sizeof(HBOND *));
	Cmd   = (COMMAND *)ckalloc(sizeof(COMMAND));

	// Allocate memory within the COMMAND structure
	/* char Active[MAX_CHAIN+1]; */
	/* char Processed[MAX_CHAIN+1]; */
	Cmd->Active = (char *)ckalloc((NChain+1)*sizeof(char));
	Cmd->Processed = (char *)ckalloc((NChain+1)*sizeof(char));

	// Initialize arrays to make memory clean up easier.
	//for(i = 0;i < MAX_CHAIN;i++)
	for(i = 0;i < NChain;i++)
		Chain[i] = NULL;

	for(i = 0;i < MAXHYDRBOND(num_residues);i++)
		HBond[i] = NULL;

	try{
		////////////// Set the COMMAND structure //////////////
		DefaultCmd(Cmd, num_residues);

		strcpy(Cmd->InputFile, ""); // <-- There is not default input file!

		Cmd->NActive = (int)strlen(Cmd->Active);
		Cmd->NProcessed = (int)strlen(Cmd->Processed);

		if( Cmd->Measure ) {
			Cmd->BrookhavenAsn = YES;
			Cmd->DsspAsn = YES;
		}
		///////////////////////////////////////////////////////

		for(i = 0;i < NChain;i++){
			array<amino> &sequence_ref = secondaryStruc[i + first_chain].sequence;

			InitChain(&(Chain[i]), Cmd);
			
			// The number of animoacids in the chain
			Chain[i]->NRes = sequence_ref.Length();	
			Chain[i]->NHetRes = 0;
			Chain[i]->NonStandRes = 0;
			Chain[i]->Ter = 0;

			Chain[i]->Id = secondaryStruc[i + first_chain].chain;
			Chain[i]->Published = TRUE; 
			Chain[i]->DsspAssigned = FALSE;

			// Translate the atomic level detail of the atoms array into
			// the CHAIN structure used by Stride.

			for(j = 0;j < Chain[i]->NRes;j++){
				// Only assign the residue memory, name and number once per residue
				Chain[i]->Rsd[j] = (RESIDUE *)ckalloc(sizeof(RESIDUE));
				
				r = Chain[i]->Rsd[j];
				
				// Initialize these pointers to NULL so we can clean up in the event of 
				// an error.
				r->Inv =  NULL;
				r->Prop = NULL;

				sprintf(r->PDB_ResNumb,"%d", sequence_ref[j].resSeq);
				strcpy(r->ResType, sequence_ref[j].GetResName());

				r->NAtom = sequence_ref[j].atom_id.Length();
				
				if(r->NAtom > MAX_AT_IN_RES)
					throw "Too many atoms in residue!";

				for(k = 0;k < r->NAtom;k++){
					int cur_atom = sequence_ref[j].atom_id[k];
					
					int str_len = strlen(atoms[cur_atom].name);

					if(str_len >= AT_FIELD){
						strncpy(r->AtomType[k], atoms[cur_atom].name, AT_FIELD - 1);
						r->AtomType[k][AT_FIELD - 1] = '\0';
					}
					else{
						strcpy(r->AtomType[k], atoms[cur_atom].name);
					}
					
					// Copy the atomic coordianates
					atoms[cur_atom].GetCoord(r->Coord[k]);

					r->Occupancy[k] = atoms[cur_atom].occupancy;
					r->TempFactor[k] = atoms[cur_atom].temperature;
				}

				r->Inv =  (INVOLVED *)ckalloc(sizeof(INVOLVED));
				r->Prop = (PROPERTY *)ckalloc(sizeof(PROPERTY));
				r->Inv->NBondDnr = 0;
				r->Inv->NBondAcc = 0;
				r->Inv->InterchainHBonds = NO;
				r->Prop->Asn     = 'C';
				r->Prop->PdbAsn  = 'C';
				r->Prop->DsspAsn = 'C';
				r->Prop->Solv    = 0.0;
				r->Prop->Phi     = 360.0;
				r->Prop->Psi     = 360.0;
			}
		}
		///////////////////////////////////////////////////////

		for( Cn=0; Cn<NChain; Cn++ )
			ValidChain += CheckChain(Chain[Cn],Cmd);

		if( !ValidChain ) 
			die("No valid chain\n");

		if( Cmd->BrookhavenAsn )
			GetPdbAsn(Chain,NChain);

		if( Cmd->DsspAsn )
			GetDsspAsn(Chain,NChain,Cmd);

		BackboneAngles(Chain,NChain);

		if( Cmd->OutSeq )
			OutSeq(Chain,NChain,Cmd);

		if( !strlen(Cmd->MapFileHelix) )
			PhiPsiMapHelix = DefaultHelixMap(Cmd);
		else
			ReadPhiPsiMap(Cmd->MapFileHelix,&PhiPsiMapHelix,Cmd);

		if( !strlen(Cmd->MapFileSheet) )
			PhiPsiMapSheet = DefaultSheetMap(Cmd);
		else
			ReadPhiPsiMap(Cmd->MapFileSheet,&PhiPsiMapSheet,Cmd);

		for( Cn=0; Cn<NChain; Cn++ )
			PlaceHydrogens(Chain[Cn]);

		if( (NHBond = FindHydrogenBonds(Chain,Cn,HBond,Cmd,&Dnr,&Acc)) == 0 ) 
			die("No hydrogen bonds found in %s\n",Cmd->InputFile);

		NoDoubleHBond(HBond,NHBond);

		DiscrPhiPsi(Chain,NChain,Cmd);

		if(Cmd->ExposedArea)
			Area(Chain,NChain,Cmd);

		for(Cn = 0;Cn < NChain;Cn++){

			if(Chain[Cn]->Valid){
    
				Helix(Chain,Cn,HBond,Cmd,PhiPsiMapHelix);
      
				for(i = 0;i < NChain;i++) 
					if(Chain[i]->Valid)
						Sheet(Chain,Cn,i,HBond,Cmd,PhiPsiMapSheet);    
      
				BetaTurn(Chain,Cn);
				GammaTurn(Chain,Cn,HBond);
			}
		}
		
		// Parse the Stride data to obtain secondary structure information
		for(i = 0;i < NChain;i++){
			for(j = 0;j < Chain[i]->NRes;j++){
				RESIDUE *ptr = Chain[i]->Rsd[j];

				secondaryStruc[i + first_chain].sequence[j].SetSSType(ptr->Prop->Asn);
			}
		}
	}
	catch(char *){
		// Catch a thrown "char *" error (Stride has been modified to
		// throw such an error). Note that the ellipsis version of catch
		// (i.e. catch (...)) will not catch a "throw;" -- must throw a 
		// variable (i.e. throw "Stride Error").
		return_value = FALSE;
	}
		
	// Clean up memory allocated by Stride
	for(i = 0;i < NChain;i++){
		if(Chain[i]){
			if(Chain[i]->File)
				free(Chain[i]->File);

			if(Chain[i]->Rsd){
				for(j = 0;j < Chain[i]->NRes;j++)
					if(Chain[i]->Rsd[j]){
						free(Chain[i]->Rsd[j]->Inv);

						free(Chain[i]->Rsd[j]->Prop);

						free(Chain[i]->Rsd[j]);
					}

				free(Chain[i]->Rsd);
			}

			if(Chain[i]->HetRsd){
				for(j = 0;j < MAX_HETRES;j++)
					if(Chain[i]->HetRsd[j])
						free(Chain[i]->HetRsd[j]);

				free(Chain[i]->HetRsd);
			}

			if(Chain[i]->Het){
				for(j = 0;j < MAX_HET;j++)
					if(Chain[i]->Het[j])
						free(Chain[i]->Het[j]);

				free(Chain[i]->Het);
			}

			if(Chain[i]->Helix){
				for(j = 0;j < MAX_HELIX;j++)
					if(Chain[i]->Helix[j])
						free(Chain[i]->Helix[j]);

				free(Chain[i]->Helix);
			}

			if(Chain[i]->Sheet){
				for(j = 0;j < MAX_SHEET;j++)
					if(Chain[i]->Sheet[j])
						free(Chain[i]->Sheet[j]);

				free(Chain[i]->Sheet);
			}

			if(Chain[i]->Turn){
				for(j = 0;j < MAX_TURN;j++)
					if(Chain[i]->Turn[j])
						free(Chain[i]->Turn[j]);

				free(Chain[i]->Turn);
			}

			if(Chain[i]->AssignedTurn){
				for(j = 0;j < MAX_TURN;j++){
					if(Chain[i]->AssignedTurn[j])
						free(Chain[i]->AssignedTurn[j]);
				}

				free(Chain[i]->AssignedTurn);
			}

			if(Chain[i]->SSbond){
				for(j = 0;j < MAX_BOND;j++)
					if(Chain[i]->SSbond[j])
						free(Chain[i]->SSbond[j]);

				free(Chain[i]->SSbond);
			}

			if(Chain[i]->Info){
				for(j = 0;j < MAX_INFO;j++)
					if(Chain[i]->Info[j])
						free(Chain[i]->Info[j]);

				free(Chain[i]->Info);
			}

			free(Chain[i]);
		}
	}

	for(i = 0;i < NHBond;i++)
		if(HBond[i]){
			free(HBond[i]);
		}

	free(Chain);
	free(HBond);

	if(Dnr){
		for(i = 0;i < MAXDONOR(num_residues);i++)
			if(Dnr[i])
				free(Dnr[i]);

		free(Dnr);
	}

	if(Acc){
		for(i = 0;i < MAXACCEPTOR(num_residues);i++)
			if(Acc[i])
				free(Acc[i]);

		free(Acc);
	}

	if(PhiPsiMapHelix)
		free(PhiPsiMapHelix);

	if(PhiPsiMapSheet)
		free(PhiPsiMapSheet);

	if(Cmd){
		if(Cmd->Active){
			free(Cmd->Active);
		}

		if(Cmd->Processed){
			free(Cmd->Processed);
		}

		free(Cmd);
	}

	return return_value;
}

void molecule::SetRibbonSize(int m_type)
{
	int i, j;

	switch(m_type){
		case QM_RENDER_RIBBON:
		case QM_RENDER_RIBBON_FLAT:
			for(i = 0;i < secondaryStruc.Length();i++){
				array<amino> &seq_ref = secondaryStruc[i].sequence;

				for(j = 0;j < seq_ref.Length();j++){
					// Assign the ribbon size based on the secondary
					// structure type
					switch(seq_ref[j].ss){
						case SS_ALPHA_HELIX:
						case SS_310_HELIX:
						case SS_PI_HELIX:
							// Helix
							seq_ref[j].major = 1.0f;
							seq_ref[j].minor = 0.5f;	
							break;
						case SS_STRAND:
							// Sheet
							seq_ref[j].major = 2.0f;
							seq_ref[j].minor = 0.2f;
							break;
						default:
							// Coil
							seq_ref[j].major = 0.2f;
							seq_ref[j].minor = 0.2f;
							break;
					};
				}
			}
			break;
		case QM_RENDER_TUBE:
			for(i = 0;i < secondaryStruc.Length();i++){
				array<amino> &seq_ref = secondaryStruc[i].sequence;

				for(j = 0;j < seq_ref.Length();j++){
					// Assign the ribbon a constant size
					seq_ref[j].major = 0.5f;
					seq_ref[j].minor = 0.5f;
				}
			}
			break;
		case QM_RENDER_TUBE_T:
			for(i = 0;i < secondaryStruc.Length();i++){
				array<amino> &seq_ref = secondaryStruc[i].sequence;

				for(j = 0;j < seq_ref.Length();j++){
					// Assign the ribbon a size based on the temperature
					// parameter of the alpha carbon					
					float t = 2.0f*(
						NormalizeByRange(atoms[seq_ref[j].c_alpha_index].temperature,
						min_temp, max_temp) + 0.1f);

					seq_ref[j].major = t;
					seq_ref[j].minor = t;
				}
			}
			break;
		case QM_RENDER_TUBE_OCC:
			for(i = 0;i < secondaryStruc.Length();i++){
				array<amino> &seq_ref = secondaryStruc[i].sequence;

				for(j = 0;j < seq_ref.Length();j++){
					// Assign the ribbon a size based on the occupancy
					// parameter of the alpha carbon
					float o = 2.0f*(
						NormalizeByRange(atoms[seq_ref[j].c_alpha_index].occupancy,
						min_occ, max_occ) + 0.1f);

					seq_ref[j].major = o;
					seq_ref[j].minor = o;
				}
			}
			break;
		default:
			break;
	};
}

// Update the temperature factors
void molecule::UpdateTemperature()
{
	if(atoms.Length() == 0)
		return;

	max_temp = atoms[0].temperature;
	min_temp = atoms[0].temperature;

	for(int i = 1;i < atoms.Length();i++){
		if(atoms[i].temperature > max_temp)
			max_temp = atoms[i].temperature;

		if(atoms[i].temperature < min_temp)
			min_temp = atoms[i].temperature;
	}
}

// Update the occupancy
void molecule::UpdateOccupancy()
{
	if(atoms.Length() == 0)
		return;

	max_occ = atoms[0].occupancy;
	min_occ = atoms[0].occupancy;

	for(int i = 1;i < atoms.Length();i++){
		if(atoms[i].occupancy > max_occ)
			max_occ = atoms[i].occupancy;

		if(atoms[i].occupancy < min_occ)
			min_occ = atoms[i].occupancy;
	}
}

void molecule::SetAtomicRadiiScheme(RADIUS_FORMAT format, int elem, int index,
			float radius)
{
	int i;

	switch(format){
		case QM_RADIUS_DEFAULT:
			// vanDerWaals radii from Rasmol abstree.h source file
			elem_radii_scheme[ELEM_C] = 1.689f;
			elem_radii_scheme[ELEM_N] = 1.527f;
			elem_radii_scheme[ELEM_O] = 1.471f;
			elem_radii_scheme[ELEM_S] = 1.972f;
			elem_radii_scheme[ELEM_H] = 1.20f;
			elem_radii_scheme[ELEM_P] = 2.051f;
			elem_radii_scheme[ELEM_CL] = 1.907f;
			elem_radii_scheme[ELEM_Z] = 1.2524f;
			elem_radii_scheme[ELEM_NA] = 2.400f;
			elem_radii_scheme[ELEM_FE] = 2.125f;
			elem_radii_scheme[ELEM_MG] = 1.636f;
			elem_radii_scheme[ELEM_K] = 2.605f;
			elem_radii_scheme[ELEM_CA] = 2.125f;
			elem_radii_scheme[ELEM_I] = 1.907f;
			elem_radii_scheme[ELEM_DUMMY] = 1.689f;

			for(i = 0;i < NumAtoms();i++)
				atoms[i].radius = elem_radii_scheme[atoms[i].elem];

			break;
		case QM_RADIUS_ELEMENT:
			elem_radii_scheme[elem] = radius;

			for(i = 0;i < NumAtoms();i++)
				atoms[i].radius = elem_radii_scheme[atoms[i].elem];

			break;
		case QM_RADIUS_CUSTOM:
			index = IdToIndex(index);

			if(index != -1)
				atoms[index].radius = radius;

			break;
		case QM_RADIUS_T:
			for(i = 0;i < NumAtoms();i++)
				// Scale the atomic radius so that the maximum value is
				// 3 angstroms
				atoms[i].radius = 3.0f*ScaleByRange(atoms[i].temperature,
					min_temp, max_temp);
			break;
		case QM_RADIUS_OCC:
			for(i = 0;i < NumAtoms();i++)
				// Scale the atomic radius so that the maximum value is
				// 3 angstroms
				atoms[i].radius = 3.0f*ScaleByRange(atoms[i].occupancy, 
					min_occ, max_occ);
			break;
	};
}

molecule& molecule::operator=(molecule &copy)
{
	moleculeState::operator=(copy);
	DDD::operator=(copy);

	// Don't try to directly copy molecular surfaces
	// (lots of complicated memory accessing involved).
	// The user will have to rebuild the surface (for now).
	// surf = copy.surf;

	atoms = copy.atoms;
	bonds = copy.bonds;

	secondaryStruc = copy.secondaryStruc;
	mon = copy.mon;

	xmin = copy.xmin;
	ymin = copy.ymin;
	zmin = copy.zmin;
  	xsize = copy.xsize;
	ysize = copy.ysize;
	zsize = copy.zsize;
	maxdim = copy.maxdim;
	cm = copy.cm;
	steps = copy.steps;
	size = copy.size;
	max_temp = copy.max_temp;
	min_temp = copy.min_temp;
	max_occ = copy.max_occ;
	min_occ = copy.min_occ;
	
	memcpy(elem_radii_scheme, copy.elem_radii_scheme,
		ELEM_LAST*sizeof(float));
	
	traj = copy.traj;

	mol_connect = copy.mol_connect;

	return (*this);
}

void amino::SetSSType(char m_ss)
{
	switch(m_ss) {
		case 'H': 
			ss = SS_ALPHA_HELIX;
			return;
		case 'G': 
			ss = SS_310_HELIX;
			return;
		case 'I': 
			ss = SS_PI_HELIX;
			return;
		case 'E': 
			ss = SS_STRAND;
			return;
		case 'B': 
		case 'b': 
            ss = SS_BRIDGE;
			return;
		case 'C': 
			ss = SS_COIL;
			return;
		case '1': 
			ss = SS_TURN_1;
			return;
		case '2': 
			ss = SS_TURN_1P;
			return;
		case '3': 
			ss = SS_TURN_2;
			return;
		case '4':
			ss = SS_TURN_2P;
			return;
		case '5': 
			ss = SS_TURN_6A;
			return;
		case '6': 
			ss = SS_TURN_6B;
			return;
		case '7': 
			ss = SS_TURN_8;
			return;
		case '8': 
			ss = SS_TURN_4;
			return;
		case '@': 
			ss = SS_TURN_GAMMA;
			return;
		case '&': 
			ss = SS_TURN_GAMMA_INV;
			return;
		case 'T': 
			ss = SS_TURN;
			return;
		default:  
			ss = SS_DUMMY;
			return;
	};
}

BOOL amino::SetCAlpha(char *m_name, int m_i)
{
	if(strcmpi(m_name, "CA") == 0){
		// We found the alpha carbon, return true
		c_alpha_index = m_i;

		return TRUE;
	}

	// This is NOT the alpha carbon
	return FALSE;
}

BOOL amino::SetO(char *m_name, int m_i)
{
	if(strcmpi(m_name, "O") == 0){
		// We found the carbonyl oxygen, return true
		o_index = m_i;

		return TRUE;
	}

	// This is NOT the carbonyl oxygen
	return FALSE;
}

// Turn a three letter amino acid code into an enumerated type
void amino::SetAminoAcidType(char *m_type)
{
	#ifdef DEBUG
	if(strlen(m_type) != 3)
		throw Error("Invalid amino acid type!", 
			__FILE__, __LINE__);
	#endif // DEBUG

	switch(tolower(m_type[0])){
		case 'a': //ALA, ASP, ARG, ASN
			if(tolower(m_type[1]) == 'l'){
				aa = AA_ALA;
				return;
			}
			
			if(tolower(m_type[1]) == 'r'){
				aa = AA_ARG;
				return;
			}

			if(tolower(m_type[2]) == 'p'){
				aa = AA_ASP;
				return;
			}

			aa = AA_ASN;
			return;

			break;
		case 'v': // VAL
			aa = AA_VAL;
			return;
			break;
		case 'p': // PHE, PRO
			if(tolower(m_type[1]) == 'h'){
				aa = AA_PHE;
				return;
			}

			aa = AA_PRO;
			return;
			break;
		case 'm': // MET
			aa = AA_MET;
			return;
			break;
		case 'i': // ILE
			aa = AA_ILE;
			return;
			break;
		case 'l': // LEU, LYS
			if(tolower(m_type[1]) == 'e'){
				aa = AA_LEU;
				return;
			}

			aa = AA_LYS;
			return;
			break;
		case 'g': // GLU, GLN, GLY
			if(tolower(m_type[2]) == 'u'){
				aa = AA_GLU;
				return;
			}

			if(tolower(m_type[2]) == 'y'){
				aa = AA_GLY;
				return;
			}

			aa = AA_GLN;
			return;
			break;
		case 's': // SER
			aa = AA_SER;
			return;
			break;
		case 't': // THR, TYR, TRP
			if(tolower(m_type[1]) == 'h'){
				aa = AA_THR;
				return;
			}

			if(tolower(m_type[1]) == 'y'){
				aa = AA_TYR;
				return;
			}

			aa = AA_TRP;
			return;
			break;
		case 'h': // HIS
			aa = AA_HIS;
			return;
			break;
		case 'c': // CYS
			aa = AA_CYS;
			return;
			break;
		default:
			aa = AA_DUMMY;
			break;
	};
}

char *amino::GetResName()
{
	static char *res_name[] = {"ALA", "VAL", "PHE", "PRO", "MET", "ILE", "LEU",
		"ASP", "GLU", "LYS", "ARG", "SER", "THR", "TYR", "HIS", "CYS", "ASN",
		"GLN", "TRP", "GLY", "UNK"};

	return res_name[aa];
}


// Read the file specified by "filename" into the current 
// document. Set the window title with "title"
BOOL CQmolDoc::ReadFile(LPCSTR filename)
{
	// Grab a pointer to the status bar on the window frame -- we'll
	// use it to keep the user informed for file reading progress.
	CMainFrame *pFrame = (CMainFrame*)AfxGetApp()->m_pMainWnd;
	CStatusBar* pStatus = &pFrame->m_wndStatusBar;

	// Add a molecule to the molecule list
	activeMol = new molecule;

	if(activeMol == NULL){
		throw QmolError("Unable to allocate memory for new molecule");
	}

	molList.Push(activeMol);

	// Make this new molecule the only selected molecule
	SelectMol(activeMol);

	// If the user has previously read in a trajectory file,
	// clear it from memory and set all trajectory parameters to
	// the default values.
	activeMol->steps = 0;
	activeMol->size = 0;
	while(!activeMol->traj.IsEmpty())
		activeMol->traj.Pop();

	activeMol->cm.Alloc(0);

	// Open the File
	activeMol->ReadCoord(filename, multi_model_traj, connect_on_read, QM_PDB);

	// Did we get a valid file?
	if(activeMol->atoms.Length() == 0){
		// No atoms found!

		return FALSE;
	}
	
	if(secondary_struct_on_read){
		if(pStatus)
			pStatus->SetPaneText(0, "Assigning secondary structure ...");

		// Assign the secondary structure
		activeMol->ComputeSecondaryStructure();
	}

	// Set the atomic properties
	for(int i = 0;i < activeMol->NumAtoms();i++){
		activeMol->atoms[i].SetAtomColor(color_scheme);
		activeMol->atoms[i].SetAtomMass(elem_mass_scheme);
	}

	activeMol->UpdateBondColoring();

	// Compute the dimensions of the molecule(s)
	if(multi_model_traj){
		// Scale the trajectory
		activeMol->GetMolDim(multi_model_traj_process);
	}
	else{
		activeMol->GetMolDim();
	}

	// Save the file name
	activeMol->SetCoordName(filename);

	if(pStatus)
		pStatus->SetPaneText(0, "Ready");

	return TRUE;
}


BOOL CQmolDoc::OnFileCoordOpen()
{
	// Open a modal dialog to get the file name
	CFileDialog fDlg(TRUE);

	fDlg.m_ofn.lpstrFilter = "PDB files\0*.pdb\0All Files\0*.*\0\0";

	// Allow the user to open multiple files
	// Allow the user to select multiple files [so we must construct a 
	// string buffer to hold the returned name(s)].
	fDlg.m_ofn.nMaxFile = 2562; // <-- Windows NT 4.0 max. Dooh!

	fDlg.m_ofn.Flags = fDlg.m_ofn.Flags | OFN_ALLOWMULTISELECT;

	// Save a pointer to the buffer for easy memory clean up!
	char *buf;

	fDlg.m_ofn.lpstrFile = buf = new char [fDlg.m_ofn.nMaxFile];

	if(buf == NULL){
		AfxMessageBox("Memory allocation error!");
		return FALSE;
	}

	// Set the first character to null to prevent a default filename from 
	// appearing.
	fDlg.m_ofn.lpstrFile[0] = '\0';

	// Set the current directory
	char dirBuffer[1024];

	// If the buffer is not big enought, DON'T set the current 
	// directory
	if(GetCurrentDirectory(1023, dirBuffer) != 0)
		fDlg.m_ofn.lpstrInitialDir = dirBuffer;

	if(fDlg.DoModal() == IDOK){
		if(CommDlgExtendedError() == FNERR_BUFFERTOOSMALL){
			AfxMessageBox("Too many files selected! "
				"Please select fewer files and try again.");

			delete [] buf;

			return FALSE;
		}

		POSITION pos = fDlg.GetStartPosition();
		CString file;
		int numFilesRead = 0;

		while(pos){
			numFilesRead++;
			file = fDlg.GetNextPathName(pos);

			// Open the File
			try{
				if(ReadFile(file) == FALSE)
					throw QmolError("Unable to read file",
					__FILE__, __LINE__);
				
				// Set the Window Title
				SetTitle(file);
				
			}
			catch(QmolError &e){
				// Was there an error opening the file?
				AfxMessageBox(e.message());
				
				SetTitle("No Molecule Loaded");
				
				delete [] buf;

				return FALSE;
			}
			catch(...){
				// Yiikes! An unhandled error has occured!
				// Was there an error opening the file?
				AfxMessageBox("Unable to read input file!");
				
				SetTitle("No Molecule Loaded");
				
				delete [] buf;

				return FALSE;
			}
		}

		if(numFilesRead == 0){
			delete [] buf;

			AfxMessageBox("Please enter a file name");
			return FALSE;
		}

		if(numFilesRead == 1){
			// Before returning, add the file name to the MRU file list
			AfxGetApp()->AddToRecentFileList(file);
		}

		delete [] buf;

		return TRUE;
	}

	return FALSE;
}

BOOL CQmolDoc::OnNetFileCoordOpen()
{

	CNetworkPBDDialog netDlg(server_list, pdb_query_list);

	if(netDlg.DoModal() == IDOK){

		// Perform new document work (just for show)
		//CDocument::OnNewDocument();
		
		// Open the File
		try{
			if(ReadFile(netDlg.tmp_pdb_file) == FALSE)
				throw QmolError("Can't read pdb file -- please check server and pdb id", 
					__FILE__, __LINE__);

			// Remove the temporary file
			try
			{
				CFile::Remove(netDlg.tmp_pdb_file);
			}
			catch(...)
			{
				throw QmolError("Can't remove temporary pdb file", 
					__FILE__, __LINE__);
			}
		}
		catch(QmolError &e){
			// Was there an error opening the file?
			AfxMessageBox(e.message());

			SetTitle("No Molecule Loaded");

			return FALSE;
		}

		// Set the title accordning to the PDB id -- not the temporary file
		// name
		SetTitle(netDlg.tmp_PDB_Id);

		return TRUE;
	}

	return FALSE;
}

BOOL CQmolDoc::OnFileCoordSave() 
{
	// Open a modal dialog to get the file name
	CSaveCoordDialog scfDlg;
	
	if(scfDlg.DoModal() == IDOK){
		// Make sure the user entered a file name
		if(scfDlg.GetFileName() == _T("")){
			AfxMessageBox("Please enter a file name");

			return OnFileCoordSave();
		}

		// Save all of the currently selected molecules
		FILE *fout = fopen(scfDlg.GetFileName(), "w");

		try{
			if(!fout)
				throw QmolError("Unable to open file for saving",
					__FILE__, __LINE__);
			
			// An atom offset to allow mulitple molecules to be
			// saved to the same file
			int atomOffset = 0;

			ListIterator<molecule*> next(molList);
			molecule **tmpMol;

			while(tmpMol = next()){
				(*tmpMol)->SaveCoord(fout, 
									 scfDlg.GetFileType(), 
									 scfDlg.GetPrecision(),
									 scfDlg.m_write_conect,
									 scfDlg.m_normalize_t,
									 atomOffset);

				// The "+1" is for the TER record after each
				// molecule
				atomOffset += (*tmpMol)->NumAtoms() + 1;
			}
		}
		catch(QmolError &e){
			// Was there an error opening the file?
			AfxMessageBox(e.message());

			if(fout)
				fclose(fout);

			return FALSE;
		}

		if(fout)
			fclose(fout);

		return TRUE;
	}

	return FALSE;
}

BOOL CQmolDoc::OnFileOpenCoordTraj() 
{
	// Open a modal dialog to get the file name
	COpenCoordTrajec octfDlg;
	
	CString trajec_file_name;
	CString coord_file_name;
	
	int trajec_file_type;
	int coord_file_type;

	if(octfDlg.DoModal() == IDOK){
		// Grab a pointer to the status bar on the window frame -- we'll
		// use it to keep the user informed for file reading progress.
		CMainFrame *pFrame = (CMainFrame*)AfxGetApp()->m_pMainWnd;
		CStatusBar* pStatus = &pFrame->m_wndStatusBar;


		trajec_file_name = octfDlg.GetTrajecFileName();
		coord_file_name = octfDlg.GetCoordFileName();
		trajec_file_type = octfDlg.GetTrajecFileType();
		coord_file_type = octfDlg.GetCoordFileType();

		// Make sure the user entered a file name
		if(coord_file_name == _T("")){
			AfxMessageBox("Please enter a coordinate file name");
			return FALSE;
		}
		
		if(trajec_file_name == _T("")){
			AfxMessageBox("Please enter a trajectory file name");
			return FALSE;
		}

		// Perform new document work (just for show)
		//CDocument::OnNewDocument();
		
		try{
			// Add a molecule to the molecule list
			activeMol = new molecule;

			if(activeMol == NULL){
				throw QmolError("Unable to allocate memory for new molecule");
			}

			molList.Push(activeMol);

			// Make this new molecule the only selected molecule
			SelectMol(activeMol);

			// Open the File
			activeMol->ReadCoord(LPCTSTR(coord_file_name), multi_model_traj,
					   connect_on_read, coord_file_type);

			// Save the file name
			activeMol->SetCoordName(coord_file_name);

			// Did we get a valid file?
			if(activeMol->atoms.Length() == 0){
				// No atoms found!

				return FALSE;
			}
			
			if(secondary_struct_on_read){
				if(pStatus)
					pStatus->SetPaneText(0, "Assigning secondary structure ...");
				
				// Assign the secondary structure
				activeMol->ComputeSecondaryStructure();
			}

			// Set the atomic properties
			for(int i = 0;i < activeMol->NumAtoms();i++){
				activeMol->atoms[i].SetAtomColor(color_scheme);
				activeMol->atoms[i].SetAtomMass(elem_mass_scheme);
			}

			activeMol->UpdateBondColoring();

			// If the user has previously read in a trajectory file,
			// clear it from memory and set all trajectory parameters to
			// the default values.
			activeMol->steps = 0;
			activeMol->size = 0;

			while(!activeMol->traj.IsEmpty())
				activeMol->traj.Pop();

			activeMol->ReadTrajHeader(LPCTSTR(trajec_file_name), trajec_file_type);
		
			// Bring up a dialog to keep track of the reading progress
			CReadProgressDialog read_progress(0, activeMol->steps, activeMol,
				trajec_file_name);

			// Read the trajectory
			read_progress.DoModal();
			
			// Did the user hit cancel?
			if(!read_progress.Status())
				throw QMOL_ERROR("Trajectory read has been cancelled");

			// Compute the bounds of the structure + trajectory
			activeMol->GetMolDim(octfDlg.GetTrajecProcess());

			// Set the Window Title
			SetTitle(LPCTSTR(coord_file_name + "+" + trajec_file_name));

			if(pStatus)
				pStatus->SetPaneText(0, "Ready");
		}
		catch(QmolError &e){
			// Set any values that may have been modified by reading the
			// trajectory header back to their default vaules.
			activeMol->steps = 0;
			activeMol->size = 0;
			while(!activeMol->traj.IsEmpty())
				activeMol->traj.Pop();
			activeMol->cm.Alloc(0);

			AfxMessageBox(e.message());

			// At this point, we need to reset the view no matter what!
			return TRUE;
		}
		catch(...){
			// Yiikes! An unhandled error!
			// Set any values that may have been modified by reading the
			// trajectory header back to their default vaules.
			activeMol->steps = 0;
			activeMol->size = 0;
			while(!activeMol->traj.IsEmpty())
				activeMol->traj.Pop();
			activeMol->cm.Alloc(0);

			AfxMessageBox("Unable to read input files!");

			// At this point, we need to reset the view no matter what!
			return TRUE;
		}

		return TRUE;
	}

	return FALSE;
}

void CQmolDoc::OnFileTrajExport(const char *file_name, int file_type, int start, int stop)
{
	switch(file_type){
		case QM_PDB:
			activeMol->TrajExportPdb(file_name, start, stop);
			break;
	};
}

void CQmolDoc::OnOptionsRead() 
{
	// Open the read options dialog to allow the user to set various
	// options.
	CReadOptionsDialog roDlg(connect_on_read, 
							 multi_model_traj,
							 secondary_struct_on_read,
							 multi_model_traj_process);

	if(roDlg.DoModal() == IDOK){
		// Apply Changes
		connect_on_read = roDlg.m_connect;
		multi_model_traj = roDlg.m_multi_struct;
		secondary_struct_on_read = roDlg.m_sec_struc;
		multi_model_traj_process = roDlg.m_traj_process;
	}
}

void CQmolDoc::SetColorScheme(int scheme)
{
	switch(scheme){
		case QM_COLOR_DEFAULT:
			// CPK is the deafult coloring scheme
		case QM_COLOR_CPK:
			// Colors from the Rasmol source file
			// abstree.h
			color_scheme[ELEM_C][0] = 200; 
			color_scheme[ELEM_C][1] = 200; 
			color_scheme[ELEM_C][2] = 200;
			color_scheme[ELEM_C][3] = 255;

			color_scheme[ELEM_N][0] = 143; 
			color_scheme[ELEM_N][1] = 143; 
			color_scheme[ELEM_N][2] = 255;
			color_scheme[ELEM_N][3] = 255;

			color_scheme[ELEM_O][0] = 240; 
			color_scheme[ELEM_O][1] = 0; 
			color_scheme[ELEM_O][2] = 0;
			color_scheme[ELEM_O][3] = 255;

			color_scheme[ELEM_S][0] = 255; 
			color_scheme[ELEM_S][1] = 200; 
			color_scheme[ELEM_S][2] = 50;
			color_scheme[ELEM_S][3] = 255;

			color_scheme[ELEM_H][0] = 225; 
			color_scheme[ELEM_H][1] = 225; 
			color_scheme[ELEM_H][2] = 225;
			color_scheme[ELEM_H][3] = 255;

			color_scheme[ELEM_P][0] = 255; 
			color_scheme[ELEM_P][1] = 165; 
			color_scheme[ELEM_P][2] = 0;
			color_scheme[ELEM_P][3] = 255;

			color_scheme[ELEM_CL][0] = 0; 
			color_scheme[ELEM_CL][1] = 255; 
			color_scheme[ELEM_CL][2] = 0;
			color_scheme[ELEM_CL][3] = 255;

			color_scheme[ELEM_Z][0] = 165; 
			color_scheme[ELEM_Z][1] = 42; 
			color_scheme[ELEM_Z][2] = 42;
			color_scheme[ELEM_Z][3] = 255;

			color_scheme[ELEM_NA][0] = 0; 
			color_scheme[ELEM_NA][1] = 0; 
			color_scheme[ELEM_NA][2] = 255;
			color_scheme[ELEM_NA][3] = 255;

			color_scheme[ELEM_FE][0] = 255; 
			color_scheme[ELEM_FE][1] = 165; 
			color_scheme[ELEM_FE][2] = 0;
			color_scheme[ELEM_FE][3] = 255;

			color_scheme[ELEM_MG][0] = 42; 
			color_scheme[ELEM_MG][1] = 128; 
			color_scheme[ELEM_MG][2] = 42;
			color_scheme[ELEM_MG][3] = 255;

			color_scheme[ELEM_K][0] = 128; 
			color_scheme[ELEM_K][1] = 128; 
			color_scheme[ELEM_K][2] = 128;
			color_scheme[ELEM_K][3] = 255;

			color_scheme[ELEM_CA][0] = 128; 
			color_scheme[ELEM_CA][1] = 128; 
			color_scheme[ELEM_CA][2] = 128;
			color_scheme[ELEM_CA][3] = 255;

			color_scheme[ELEM_I][0] = 160; 
			color_scheme[ELEM_I][1] = 32; 
			color_scheme[ELEM_I][2] = 240;
			color_scheme[ELEM_I][3] = 255;

			color_scheme[ELEM_DUMMY][0] = 255; 
			color_scheme[ELEM_DUMMY][1] = 20; 
			color_scheme[ELEM_DUMMY][2] = 147;
			color_scheme[ELEM_DUMMY][3] = 255;
			break;
		case QM_COLOR_STRUC_DEFAULT:
			// Set the default structure based color scheme
			// (by default, use the Rasmol structure color
			// scheme).
			color_scheme_struc[SS_ALPHA_HELIX][0] = 240;
			color_scheme_struc[SS_ALPHA_HELIX][1] = 0;
			color_scheme_struc[SS_ALPHA_HELIX][2] = 128;
			color_scheme_struc[SS_ALPHA_HELIX][3] = 255;

			color_scheme_struc[SS_310_HELIX][0] = 240;
			color_scheme_struc[SS_310_HELIX][1] = 0;
			color_scheme_struc[SS_310_HELIX][2] = 128;
			color_scheme_struc[SS_310_HELIX][3] = 255;

			color_scheme_struc[SS_PI_HELIX][0] = 240;
			color_scheme_struc[SS_PI_HELIX][1] = 0;
			color_scheme_struc[SS_PI_HELIX][2] = 128;
			color_scheme_struc[SS_PI_HELIX][3] = 255;

			color_scheme_struc[SS_STRAND][0] = 255;
			color_scheme_struc[SS_STRAND][1] = 255;
			color_scheme_struc[SS_STRAND][2] = 0;
			color_scheme_struc[SS_STRAND][3] = 255;

			color_scheme_struc[SS_BRIDGE][0] = 255;
			color_scheme_struc[SS_BRIDGE][1] = 255;
			color_scheme_struc[SS_BRIDGE][2] = 255;
			color_scheme_struc[SS_BRIDGE][3] = 255;

			color_scheme_struc[SS_COIL][0] = 255;
			color_scheme_struc[SS_COIL][1] = 255;
			color_scheme_struc[SS_COIL][2] = 255;
			color_scheme_struc[SS_COIL][3] = 255;

			color_scheme_struc[SS_TURN][0] = 96;
			color_scheme_struc[SS_TURN][1] = 128;
			color_scheme_struc[SS_TURN][2] = 255;
			color_scheme_struc[SS_TURN][3] = 255;

			color_scheme_struc[SS_TURN_1][0] = 96;
			color_scheme_struc[SS_TURN_1][1] = 128;
			color_scheme_struc[SS_TURN_1][2] = 255;
			color_scheme_struc[SS_TURN_1][3] = 255;

			color_scheme_struc[SS_TURN_1P][0] = 96;
			color_scheme_struc[SS_TURN_1P][1] = 128;
			color_scheme_struc[SS_TURN_1P][2] = 255;
			color_scheme_struc[SS_TURN_1P][3] = 255;

			color_scheme_struc[SS_TURN_2][0] = 96;
			color_scheme_struc[SS_TURN_2][1] = 128;
			color_scheme_struc[SS_TURN_2][2] = 255;
			color_scheme_struc[SS_TURN_2][3] = 255;

			color_scheme_struc[SS_TURN_2P][0] = 96;
			color_scheme_struc[SS_TURN_2P][1] = 128;
			color_scheme_struc[SS_TURN_2P][2] = 255;
			color_scheme_struc[SS_TURN_2P][3] = 255;

			color_scheme_struc[SS_TURN_4][0] = 96;
			color_scheme_struc[SS_TURN_4][1] = 128;
			color_scheme_struc[SS_TURN_4][2] = 255;
			color_scheme_struc[SS_TURN_4][3] = 255;

			color_scheme_struc[SS_TURN_6A][0] = 96;
			color_scheme_struc[SS_TURN_6A][1] = 128;
			color_scheme_struc[SS_TURN_6A][2] = 255;
			color_scheme_struc[SS_TURN_6A][3] = 255;

			color_scheme_struc[SS_TURN_6B][0] = 96;
			color_scheme_struc[SS_TURN_6B][1] = 128;
			color_scheme_struc[SS_TURN_6B][2] = 255;
			color_scheme_struc[SS_TURN_6B][3] = 255;

			color_scheme_struc[SS_TURN_8][0] = 96;
			color_scheme_struc[SS_TURN_8][1] = 128;
			color_scheme_struc[SS_TURN_8][2] = 255;
			color_scheme_struc[SS_TURN_8][3] = 255;

			color_scheme_struc[SS_TURN_GAMMA][0] = 96;
			color_scheme_struc[SS_TURN_GAMMA][1] = 128;
			color_scheme_struc[SS_TURN_GAMMA][2] = 255;
			color_scheme_struc[SS_TURN_GAMMA][3] = 255;

			color_scheme_struc[SS_TURN_GAMMA_INV][0] = 96;
			color_scheme_struc[SS_TURN_GAMMA_INV][1] = 128;
			color_scheme_struc[SS_TURN_GAMMA_INV][2] = 255;
			color_scheme_struc[SS_TURN_GAMMA_INV][3] = 255;

			color_scheme_struc[SS_DUMMY][0] = 255;
			color_scheme_struc[SS_DUMMY][1] = 255;
			color_scheme_struc[SS_DUMMY][2] = 255;
			color_scheme_struc[SS_DUMMY][3] = 255;
			break;
		case QM_COLOR_AMINO:
			// Set the default amino acid property based color scheme
			// (by default, use the Rasmol amino color
			// scheme).
			color_scheme_amino[AA_ALA][0] = 200;
			color_scheme_amino[AA_ALA][1] = 200;
			color_scheme_amino[AA_ALA][2] = 200;
			color_scheme_amino[AA_ALA][3] = 255;

			color_scheme_amino[AA_VAL][0] = 15;
			color_scheme_amino[AA_VAL][1] = 130;
			color_scheme_amino[AA_VAL][2] = 15;
			color_scheme_amino[AA_VAL][3] = 255;
			
			color_scheme_amino[AA_PHE][0] = 50;
			color_scheme_amino[AA_PHE][1] = 50;
			color_scheme_amino[AA_PHE][2] = 170;
			color_scheme_amino[AA_PHE][3] = 255;

			color_scheme_amino[AA_PRO][0] = 220;
			color_scheme_amino[AA_PRO][1] = 150;
			color_scheme_amino[AA_PRO][2] = 130;
			color_scheme_amino[AA_PRO][3] = 255;

			color_scheme_amino[AA_MET][0] = 230;
			color_scheme_amino[AA_MET][1] = 230;
			color_scheme_amino[AA_MET][2] = 0;
			color_scheme_amino[AA_MET][3] = 255;

			color_scheme_amino[AA_ILE][0] = 15;
			color_scheme_amino[AA_ILE][1] = 130;
			color_scheme_amino[AA_ILE][2] = 15;
			color_scheme_amino[AA_ILE][3] = 255;

			color_scheme_amino[AA_LEU][0] = 15;
			color_scheme_amino[AA_LEU][1] = 130;
			color_scheme_amino[AA_LEU][2] = 15;
			color_scheme_amino[AA_LEU][3] = 255;

			color_scheme_amino[AA_ASP][0] = 230;
			color_scheme_amino[AA_ASP][1] = 10;
			color_scheme_amino[AA_ASP][2] = 10;
			color_scheme_amino[AA_ASP][3] = 255;

			color_scheme_amino[AA_GLU][0] = 230;
			color_scheme_amino[AA_GLU][1] = 10;
			color_scheme_amino[AA_GLU][2] = 10;
			color_scheme_amino[AA_GLU][3] = 255;

			color_scheme_amino[AA_LYS][0] = 20;
			color_scheme_amino[AA_LYS][1] = 90;
			color_scheme_amino[AA_LYS][2] = 255;
			color_scheme_amino[AA_LYS][3] = 255;

			color_scheme_amino[AA_ARG][0] = 20;
			color_scheme_amino[AA_ARG][1] = 90;
			color_scheme_amino[AA_ARG][2] = 255;
			color_scheme_amino[AA_ARG][3] = 255;

			color_scheme_amino[AA_SER][0] = 250;
			color_scheme_amino[AA_SER][1] = 150;
			color_scheme_amino[AA_SER][2] = 0;
			color_scheme_amino[AA_SER][3] = 255;

			color_scheme_amino[AA_THR][0] = 250;
			color_scheme_amino[AA_THR][1] = 150;
			color_scheme_amino[AA_THR][2] = 0;
			color_scheme_amino[AA_THR][3] = 255;

			color_scheme_amino[AA_TYR][0] = 50;
			color_scheme_amino[AA_TYR][1] = 50;
			color_scheme_amino[AA_TYR][2] = 170;
			color_scheme_amino[AA_TYR][3] = 255;

			color_scheme_amino[AA_HIS][0] = 130;
			color_scheme_amino[AA_HIS][1] = 130;
			color_scheme_amino[AA_HIS][2] = 210;
			color_scheme_amino[AA_HIS][3] = 255;

			color_scheme_amino[AA_CYS][0] = 230;
			color_scheme_amino[AA_CYS][1] = 230;
			color_scheme_amino[AA_CYS][2] = 0;
			color_scheme_amino[AA_CYS][3] = 255;

			color_scheme_amino[AA_ASN][0] = 0;
			color_scheme_amino[AA_ASN][1] = 220;
			color_scheme_amino[AA_ASN][2] = 220;
			color_scheme_amino[AA_ASN][3] = 255;

			color_scheme_amino[AA_GLN][0] = 0;
			color_scheme_amino[AA_GLN][1] = 220;
			color_scheme_amino[AA_GLN][2] = 220;
			color_scheme_amino[AA_GLN][3] = 255;

			color_scheme_amino[AA_TRP][0] = 180;
			color_scheme_amino[AA_TRP][1] = 90;
			color_scheme_amino[AA_TRP][2] = 180;
			color_scheme_amino[AA_TRP][3] = 255;

			color_scheme_amino[AA_GLY][0] = 235;
			color_scheme_amino[AA_GLY][1] = 235;
			color_scheme_amino[AA_GLY][2] = 235;
			color_scheme_amino[AA_GLY][3] = 255;

			// Color AA_DUMMY like AA_ALA
			color_scheme_amino[AA_DUMMY][0] = 200;
			color_scheme_amino[AA_DUMMY][1] = 200;
			color_scheme_amino[AA_DUMMY][2] = 200;
			color_scheme_amino[AA_DUMMY][3] = 255;

			break;
	};
}

void CQmolDoc::SetAtomicMassScheme()
{
	elem_mass_scheme[ELEM_C] = 12.011f;
	elem_mass_scheme[ELEM_N] = 14.0067f;
	elem_mass_scheme[ELEM_O] = 15.9994f;
	elem_mass_scheme[ELEM_S] = 32.06f;
	elem_mass_scheme[ELEM_H] = 1.0079f;
	elem_mass_scheme[ELEM_P] = 30.97376f;
	elem_mass_scheme[ELEM_CL] = 35.453f;
	elem_mass_scheme[ELEM_Z] = 65.38f;
	elem_mass_scheme[ELEM_NA] = 22.98977f;
	elem_mass_scheme[ELEM_FE] = 55.847f;
	elem_mass_scheme[ELEM_MG] = 24.305f;
	elem_mass_scheme[ELEM_K] = 39.0983f;
	elem_mass_scheme[ELEM_CA] = 40.08f;
	elem_mass_scheme[ELEM_I] = 125.9045f;
	elem_mass_scheme[ELEM_DUMMY] = 12.011f;
}

void CQmolDoc::ColorByStruc()
{
	for(int i = 0;i < activeMol->secondaryStruc.Length();i++){
		array<amino> &sequence_ref = activeMol->secondaryStruc[i].sequence;

		for(int j = 0;j < sequence_ref.Length();j++){
			list<int> &atom_id_ref = sequence_ref[j].atom_id;
			int ss = sequence_ref[j].ss;

			for(int k = 0;k < atom_id_ref.Length();k++){
				activeMol->atoms[atom_id_ref[k]].rgb_color[0] = 
					color_scheme_struc[ss][0];
				activeMol->atoms[atom_id_ref[k]].rgb_color[1] = 
					color_scheme_struc[ss][1];
				activeMol->atoms[atom_id_ref[k]].rgb_color[2] = 
					color_scheme_struc[ss][2];
				activeMol->atoms[atom_id_ref[k]].rgb_color[3] = 
					color_scheme_struc[ss][3];
			}
		}
	}
}

void CQmolDoc::ColorByAmino()
{
	for(int i = 0;i < activeMol->secondaryStruc.Length();i++){
		array<amino> &sequence_ref = activeMol->secondaryStruc[i].sequence;

		for(int j = 0;j < sequence_ref.Length();j++){
			list<int> &atom_id_ref = sequence_ref[j].atom_id;
			int aa = sequence_ref[j].aa;

			for(int k = 0;k < atom_id_ref.Length();k++){
				activeMol->atoms[atom_id_ref[k]].rgb_color[0] = 
					color_scheme_amino[aa][0];
				activeMol->atoms[atom_id_ref[k]].rgb_color[1] = 
					color_scheme_amino[aa][1];
				activeMol->atoms[atom_id_ref[k]].rgb_color[2] = 
					color_scheme_amino[aa][2];
				activeMol->atoms[atom_id_ref[k]].rgb_color[3] = 
					color_scheme_amino[aa][3];
			}
		}
	}
}

BOOL CQmolDoc::IsValidObj(SelectObj &obj)
{
	// Make sure that the molecule pointer contained within
	// obj is in the current list.
	ListIterator<molecule *> next(molList);
	molecule **tmpMol;
	
	BOOL isValidPtr = FALSE;

	while(tmpMol = next()){
		if(*tmpMol == obj.GetMol()){
			isValidPtr = TRUE;
		}
	}

	if(!isValidPtr)
		return FALSE;

	if(obj.isAtom()){
		if(obj.GetIndex() < 0)
			return FALSE;

		if(obj.GetIndex() >= obj.GetMol()->NumAtoms())
			return FALSE;
	}

	if(obj.isBond()){
		if(obj.GetIndex() < 0)
			return FALSE;

		if(obj.GetIndex() >= obj.GetMol()->NumBonds())
			return FALSE;
	}

	return TRUE;
}

BOOL CQmolDoc::IsValidMol(molecule *m_mol)
{
	// Make sure that the molecule pointer contained within
	// obj is in the current list.
	ListIterator<molecule *> next(molList);
	molecule **tmpMol;

	while(tmpMol = next()){
		if(*tmpMol == m_mol){
			return TRUE;
		}
	}

	return FALSE;
}

// Compute the maximum dimensions of all of the molecules currently
// loaded. The translation of the center of mass must also be accounted
// for.
float CQmolDoc::GetMaxDim(BOOL m_scaleToFit)
{
	if(!activeMol){
		// No molecules are loaded, return the default maxDim
		return 1.0f;
	}

	// Get the size of the largest dimension
	ListIterator<molecule*> next(molList);
	molecule **tmpMol;

	if(m_scaleToFit){

		float maxDimX = 0.0f;
		float maxDimY = 0.0f;
		float maxDimZ = 0.0f;

		while(tmpMol = next()){
			maxDimX = max(maxDimX, (*tmpMol)->xsize + 
				(float)(fabs((*tmpMol)->R[X])));
			maxDimY = max(maxDimY, (*tmpMol)->ysize + 
				(float)(fabs((*tmpMol)->R[Y])));
			maxDimZ = max(maxDimZ, (*tmpMol)->zsize + 
				(float)(fabs((*tmpMol)->R[Z])));
		}

		return (float)sqrt(maxDimX*maxDimX + maxDimY*maxDimY + maxDimZ*maxDimZ);
	}
	

	float maxDim = 0.0f;

	while(tmpMol = next()){
		maxDim = max(maxDim, (*tmpMol)->maxdim);
	}

	return maxDim;
}

void CQmolDoc::CurrentTrajStep(int m_step)
{
	currentStep = m_step%maxStep;

	// Increment the current step for all of the molecules
	ListIterator<molecule*> next(molList);
	molecule **tmpMol;

	while(tmpMol = next()){
		int trajLength = (*tmpMol)->traj.Length();

		if(trajLength > 0){
			(*tmpMol)->current_step = currentStep % trajLength;
		}
	}
}

void CQmolDoc::IncrementTrajStep(int m_direction)
{
	// Don't do anything if there are no trajecotries loaded
	if(maxStep == 0)
		return;

	if(m_direction == QM_FORWARD)
		currentStep = (currentStep + 1)%maxStep;
	else // QM_REVERSE
		currentStep = (currentStep - 1)%maxStep;

	// Increment the current step for all of the molecules
	ListIterator<molecule*> next(molList);
	molecule **tmpMol;

	while(tmpMol = next()){
		int trajLength = (*tmpMol)->traj.Length();

		if(trajLength > 0){
			(*tmpMol)->current_step = currentStep % trajLength;
		}
	}
}

// (re-)Compute the maximum number of trajectory steps for all
// molecules.
void CQmolDoc::UpdateTraj()
{
	currentStep = 0;
	maxStep = 0;

	// Check the number of steps for all of the molecules
	ListIterator<molecule*> next(molList);
	molecule **tmpMol;

	while(tmpMol = next()){
		maxStep = max(maxStep, (*tmpMol)->traj.Length());
	}

	// Increment the max step by one to account for the
	// first (typically PDB) structure.
	maxStep++;
}

// Select a single molecule
void CQmolDoc::SelectMol(molecule *m_mol)
{
	ListIterator<molecule*> next(molList);
	molecule **tmpMol;

	while(tmpMol = next()){
		if(m_mol == (*tmpMol))
			m_mol->isSelected = TRUE;
		else
			(*tmpMol)->isSelected = FALSE;
	}
}

BOOL CQmolDoc::AllSelected()
{
	ListIterator<molecule*> next(molList);
	molecule **tmpMol;

	while(tmpMol = next()){
		if((*tmpMol)->isSelected == FALSE){
			return FALSE;
		}
	}
	
	return TRUE;
}

BOOL CQmolDoc::AllDynamic()
{
	ListIterator<molecule*> next(molList);
	molecule **tmpMol;

	while(tmpMol = next()){
		if((*tmpMol)->GetDynamic() == FALSE){
			return FALSE;
		}
	}
	
	return TRUE;
}
