//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/

#include "stdafx.h"

#include "Qmol.h"
#include "QmolDoc.h"
#include "QmolView.h"
#include "Excepts.h"
#include "QmolUtil.h"
#include "GLtoPS.h"
#include "ExportImageDialog.h"
#include "png.h"

// Header ID needed for writing bitmaps to disk.
#define BFT_BITMAP 0x4d42   /* 'BM' */ 

#define	QMOL_PNG_COLOR_DEPTH	8
#define	QMOL_BMP_COLOR_DEPTH	24

void QmolPNGErrorCallBack(png_structp png_ptr, png_const_charp msg);

void QmolPNGErrorCallBack(png_structp png_ptr, png_const_charp msg)
{
	throw QmolError(msg);
}

HBITMAP CQmolView::MakeDIB(CDC* pDC, int dib_width, int dib_height, int dib_bpp) 
{
	CRect	rect;			// Page rectangle
	CRect	mRect;			// Position of this window's client area
	int		nXRes, nYRes;
	int		tmp_width, tmp_height;
	HDC		hDC, hMemDC;
	HGLRC	hRC, hMemRC;
	LPVOID pBitmapBits;

	BITMAPINFO bitmapInfo;
	HBITMAP hDib;

	// Set our device context mapping mode.
	// Each logical unit is converted to 1 device pixel. Positive x is to
	// the right; positive y is down.
	pDC->SetMapMode(MM_TEXT);

	// Set the font scaling
	printScaleMul = GetDeviceCaps(pDC->GetSafeHdc(), LOGPIXELSY);

	// Get the size of the page (in printer device points)
	rect.top = 0;
	rect.left = 0;

	// Get the Views size (client area!)
	GetClientRect(&mRect);

	// pInfo is NULL so we're writing to the clip board. Set the destination
	// image size to be the same as the screen dimensions. Thanks to 
	// either (a) general win32 brain deadness or (b) my infinite
	// ability to make mistakes, we must clamp the maximum size of a bitmap.
	rect.right = min(MAX_DIB_SIZE, dib_width);
	rect.bottom = min(MAX_DIB_SIZE, dib_height);

	// Determine whether the Views aspect ratio is bigger or smaller than the printers
	if (mRect.Width()*rect.Height() > mRect.Height()*rect.Width()){
		// View is proportionally wider (in x) than the Printer
		nXRes = rect.Width();
		nYRes = (nXRes*mRect.Height())/mRect.Width();
		
		// Reduce the Resolution if the Bitmap size is bigger than mem_size*1024
		// (unless mem_size < 0, in which case we use as much memory as we need to).
		while((mem_size > 0) && (3*nXRes*nYRes > mem_size*1024)){
			nXRes = nXRes/2;
			nYRes = nYRes/2;
		}
	}
	else{
		// Printer area is proportionally taller (in y) than View
		nYRes = rect.Height();
		nXRes = (nYRes*mRect.Width())/mRect.Height();

		// Reduce the Resolution if the Bitmap size is bigger than mem_size*1024
		// (unless mem_size < 0, in which case we use as much memory as we need to).
		while((mem_size > 0) && (3*nXRes*nYRes > mem_size*1024)){
			nXRes = nXRes/2;
			nYRes = nYRes/2;
		}
	}

	// First of all, initialize the bitmap header information...
	memset(&bitmapInfo, 0, sizeof(BITMAPINFO));
	bitmapInfo.bmiHeader.biSize	= sizeof(BITMAPINFOHEADER);

	bitmapInfo.bmiHeader.biWidth = nXRes;
	bitmapInfo.bmiHeader.biHeight = nYRes;
	bitmapInfo.bmiHeader.biPlanes = 1;
	bitmapInfo.bmiHeader.biBitCount = dib_bpp;
	bitmapInfo.bmiHeader.biCompression = BI_RGB;
	bitmapInfo.bmiHeader.biSizeImage = 0; // Can be set to zero for BI_RGB
 
	// Create DIB
	hDib = CreateDIBSection(m_pDC->GetSafeHdc(), &bitmapInfo, DIB_RGB_COLORS, 
		&pBitmapBits, NULL, (DWORD)0);
	
	if(hDib == NULL){
		AfxMessageBox("Unable to create DIB for storing image");
		return NULL;
	}

	// create memory device context
	if((hMemDC = CreateCompatibleDC(m_pDC->GetSafeHdc())) == NULL){
		DeleteObject(hDib);
		return NULL;
	}
	HGDIOBJ hOldDib = SelectObject(hMemDC, hDib);

	// Setup pixel format
	if (!SetMemDcPixelFormat(hMemDC, dib_bpp)){
		if (hOldDib != NULL)
			SelectObject(hMemDC, hOldDib);
		DeleteObject(hDib);
		DeleteDC(hMemDC);
		return NULL;
	}

	// Create memory rendering context
	if ((hMemRC = wglCreateContext(hMemDC)) == NULL){
		if (hOldDib != NULL)
			SelectObject(hMemDC, hOldDib);

		DeleteObject(hDib);
		DeleteDC(hMemDC);
		return NULL;
	}

	// Store current rendering and device contexts
	hDC = wglGetCurrentDC();
	hRC = wglGetCurrentContext();

	// Make this hMemRC the current OpenGL rendering context.
	wglMakeCurrent(hMemDC, hMemRC);
		
	// Set up OpenGL - this must be done for each new rendering context.
	InitializeOpenGL();

	SetupGLView(nXRes, nYRes);
	
	// Save the current viewport dimensions
	tmp_width = viewWidth;
	tmp_height = viewHeight;

	viewWidth = nXRes;
	viewHeight = nYRes;

	CQmolDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	RenderScene(pDoc->molList, QM_RENDER_SCREEN);
	
	// Make sure there is no rendering left to do
	glFinish();

	// Restore the original viewport dimensions
	viewWidth = tmp_width;
	viewHeight = tmp_height;

	// The rendering context will be no longer needed
	wglMakeCurrent(NULL, NULL);
	wglDeleteContext(hMemRC);

	// Restore the font scaling
	printScaleMul = printScaleDiv;

	// Restore last rendering and device contexts
	if(hDC != NULL && hRC != NULL)
		wglMakeCurrent(hDC, hRC);

	if (hOldDib != NULL)
		SelectObject(hMemDC, hOldDib);

	// Delete our brush and device context
	DeleteDC(hMemDC);

	return hDib;
}

void CQmolView::ExportImagePS(const char *filename) 
{
	CQmolDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	FILE *fout;

	try{

		fout = fopen(filename, "w");

		if(!fout){
			throw QmolError("Unable to open " + CString(filename) + " for writing");
		}

		GLtoPS render_ps(fout);

		// This operation could take a while, as a stop gap measure
		// display the hourglass.
		::SetCursor(LoadCursor(NULL, IDC_WAIT));

		do{
			render_ps.Start();

			RenderScene(pDoc->molList, QM_RENDER_SCREEN);
		}
		while(render_ps.Stop() == FALSE);

	}
	catch(QmolError &e){
		::SetCursor(LoadCursor(NULL, IDC_ARROW));
		AfxMessageBox(e.message());
	}

	::SetCursor(LoadCursor(NULL, IDC_ARROW));

	if(fout){
		fclose(fout);
	}
}

void CQmolView::ExportImageBMP(const char *filename) 
{
	// Get the requested raster resolution
	CExportRasterImage eriDlg(viewHeight, viewWidth);

	if(eriDlg.DoModal() != IDOK)
		return;

	FILE *fout;

	fout = fopen(filename, "wb");

	if(!fout){
		AfxMessageBox("Unable to open " + CString(filename) + " for writing");
		return;
	}

	// Set the hourglass cursor - this may take a while
	::SetCursor(LoadCursor(NULL, IDC_WAIT));

	// Get the current (screen) device context
	CDC *pDC = GetDC();

	HBITMAP hDib = MakeDIB(pDC, eriDlg.m_width, eriDlg.m_height, QMOL_BMP_COLOR_DEPTH);

	if(hDib == NULL){
		AfxMessageBox("Unable to write bitmap (invalid dib)");
		fclose(fout);
		return;
	}

	DIBSECTION dibInfo;
	GetObject(hDib, sizeof(DIBSECTION), &dibInfo);

	BITMAPFILEHEADER hdr;

	// Fill in the fields of the file header
	hdr.bfType = BFT_BITMAP; 
	hdr.bfSize = dibInfo.dsBmih.biSizeImage + sizeof(BITMAPINFOHEADER) +
		sizeof (BITMAPFILEHEADER);
	hdr.bfReserved1 = 0;
	hdr.bfReserved2 = 0;
	hdr.bfOffBits   = (DWORD)sizeof(BITMAPFILEHEADER);

	// Write the file header
	fwrite(&hdr, sizeof(BITMAPFILEHEADER), 1, fout);
	
	// Write the bitmap header
	fwrite(&(dibInfo.dsBmih), sizeof(BITMAPINFOHEADER), 1, fout);

	// Write the body of the bitmap data
	fwrite(dibInfo.dsBm.bmBits, dibInfo.dsBmih.biSizeImage, 1, fout);

	if(fout)
		fclose(fout);

	::SetCursor(LoadCursor(NULL, IDC_ARROW));	

	ReleaseDC(pDC);

	DeleteObject(hDib);
}


void CQmolView::ExportImagePNG(const char *filename) 
{
	// Get the requested raster resolution
	CExportRasterImage eriDlg(viewHeight, viewWidth);

	if(eriDlg.DoModal() != IDOK)
		return;

	FILE *fout;

	fout = fopen(filename, "wb");

	if(!fout){
		AfxMessageBox("Unable to open " + CString(filename) + " for writing");
		return;
	}

	// Set the hourglass cursor - this may take a while
	::SetCursor(LoadCursor(NULL, IDC_WAIT));

	// Get the current (screen) device context
	CDC *pDC = GetDC();

	CQmolDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	// *************************************************************************************
	// The QMOL_PNG_COLOR_DEPTH specifies the bits per COLOR. MakeDIB expect bits-per-pixel.
	// Hence the 3*QMOL_PNG_COLOR_DEPTH below.
	// *************************************************************************************
	HBITMAP hDib = MakeDIB(pDC, eriDlg.m_width, eriDlg.m_height, 3*QMOL_PNG_COLOR_DEPTH);

	if(hDib == NULL){
		AfxMessageBox("Unable to write bitmap (invalid dib)");
		fclose(fout);
		return;
	}

	DIBSECTION dibInfo;
	GetObject(hDib, sizeof(DIBSECTION), &dibInfo);

	// Now that we have a DIB, the real fun begins! We need to write
	// the raw data in the png format. Based on the example code
	// included with the libpng distribution.
	png_structp png_ptr;
	png_infop info_ptr;
	png_text text_info[2];
	png_byte *row_ptr = NULL;

	try{
		
		png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
			NULL, QmolPNGErrorCallBack, NULL);

		if(png_ptr == NULL){
			throw QmolError("Unable to initialize PNG data");
		}
		
		info_ptr = png_create_info_struct(png_ptr);

		if(info_ptr == NULL){
			throw QmolError("Unable to write png file (png_create_info_struct)");
		}

		// Set up the output control using standard C streams
		png_init_io(png_ptr, fout);

		png_set_IHDR(png_ptr, info_ptr, dibInfo.dsBm.bmWidth, dibInfo.dsBm.bmHeight, 
			QMOL_PNG_COLOR_DEPTH, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE, 
			PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

		// Write comments into the image
		text_info[0].key = "Title";
		text_info[0].text = pDoc->activeMol->coordName;
		text_info[0].compression = PNG_TEXT_COMPRESSION_NONE;
		text_info[1].key = "Author";
		text_info[1].text = "Qmol";
		text_info[1].compression = PNG_TEXT_COMPRESSION_NONE;

		png_set_text(png_ptr, info_ptr, text_info, 2);

		// Write the file header information.
		png_write_info(png_ptr, info_ptr);

		// Write the dib one row at a time
		int row_size = dibInfo.dsBmih.biSizeImage/dibInfo.dsBmih.biHeight;

		// The bytes per pixel for the output png file, not the input DIB file
		int bytes_per_pixel = 3;

		unsigned char *row_ptr = new unsigned char [bytes_per_pixel*dibInfo.dsBmih.biWidth];
		unsigned char *dib_buffer = (unsigned char *)(dibInfo.dsBm.bmBits);

		if(row_ptr == NULL)
			throw QmolError("Unable to write png file (insufficient memory)");

		// Flip the y dimension
		for(int i = dibInfo.dsBmih.biHeight - 1;i >= 0;i--){
			for(int j = 0;j < dibInfo.dsBmih.biWidth;j++){

				// Swap the color packing; RGB->BGR
				row_ptr[j*bytes_per_pixel]  = dib_buffer[i*row_size + j*3+2];
				row_ptr[j*bytes_per_pixel+1]= dib_buffer[i*row_size + j*3+1];
				row_ptr[j*bytes_per_pixel+2]= dib_buffer[i*row_size + j*3];
			}
			
			png_write_row(png_ptr, row_ptr);
		}

		png_write_end(png_ptr, info_ptr);
	}
	catch(QmolError &e){
		AfxMessageBox(e.message());
	}
	catch(...){
		AfxMessageBox("Unspecified error writing png file");
	}

	if(fout)
		fclose(fout);

	::SetCursor(LoadCursor(NULL, IDC_ARROW));	

	if(info_ptr){
		png_destroy_write_struct(&png_ptr, &info_ptr);
	}
	else{
		png_destroy_write_struct(&png_ptr, png_infopp_NULL);
	}
	
	if(row_ptr)
		delete [] row_ptr;

	ReleaseDC(pDC);

	DeleteObject(hDib);
}

void CQmolView::OnExportImage() 
{
	// Open a file dialog to let the user specify a filename and file format
	CFileDialog fDlg(FALSE /* Save As*/);

	fDlg.m_ofn.lpstrTitle = "Export Image";
	fDlg.m_ofn.lpstrFilter = "Postscript files (*.ps)\0*.ps\0"
							 "Bitmap files (*.bmp)\0*.bmp\0"
							 "PNG files (*.png)\0*.png\0\0";
	fDlg.m_ofn.Flags |= OFN_OVERWRITEPROMPT;

	if(fDlg.DoModal() == IDOK){
		// What kind of file are we writing?
		switch(fDlg.m_ofn.nFilterIndex){
			case 1: // Postscript file
				ExportImagePS(LPCSTR(fDlg.GetFileName()));
				break;
			case 2:
				ExportImageBMP(LPCSTR(fDlg.GetFileName()));
				break;
			case 3:
				ExportImagePNG(LPCSTR(fDlg.GetFileName()));
				break;
			default:
				AfxMessageBox("Unknown file format!");
				break;
		}
	}
}