//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/
// 
#include "stdafx.h"
#include "QmolMovie.h"
#include "Excepts.h"
#include "QmolView.h"

BOOL QmolMovie::Init(CString &m_filename, int m_width, int m_height,
		int m_fps, int m_bpp, CView *m_pView, BOOL m_update_surf)
{
	// If the user has not cleaned up from the last invocation,
	// attempt to do it now!
	if(valid)
		Close();

	filename = m_filename;
	
	// Perform some sanity checking
	if(m_width <= 0)
		throw QMOL_ERROR("Frame width out of bounds");

	width = m_width;

	if(m_height <= 0)
		throw QMOL_ERROR("Frame height out of bounds");

	height = m_height;

	if(m_fps <= 0)
		throw QMOL_ERROR("Frame rate out of bounds");

	fps = m_fps;
	
	// Only a few color depths are allowed!
	if((m_bpp != 16) && (m_bpp != 24) && (m_bpp != 32))
		throw QMOL_ERROR("Frame depth out of bounds");

	bpp = m_bpp;

	// We must have a valid pointer to a view
	if(!m_pView)
		throw QMOL_ERROR("Bad view pointer");

	pView = m_pView;

	// Everything looks good so far. Open an AVI file and stream.

	// Make sure we are running on 1.1 
	if (HIWORD(VideoForWindowsVersion()) < 0x010a)
		throw QMOL_ERROR("The current version of VideoForWindows is too old!");

	AVIFileInit();

	// The file name must end with ".avi"!
	if(AVIFileOpen(&aviFile, LPCTSTR(filename), OF_CREATE | OF_WRITE, NULL) != AVIERR_OK)
		throw QMOL_ERROR("Error opening the AVI file. File name must end with '.avi'");

	// stream type
	aviInfo.fccType = streamtypeVIDEO;
	aviInfo.fccHandler = 0;
	aviInfo.dwScale = 1;

	// The frame rate
	aviInfo.dwRate = fps;

	// size = width*height*bytes-per-pixel
	aviInfo.dwSuggestedBufferSize = width*height*(bpp/8);

	// rectangle for stream
	SetRect(&aviInfo.rcFrame, 0, 0, width, height);

	// And create the stream;
	if(AVIFileCreateStream(aviFile, &aviStream, &aviInfo) != AVIERR_OK)
		throw QMOL_ERROR("Error opening the AVI stream");

	// Initialize the bitmap header information...
	memset(&bmpInfoInput, 0, sizeof(BITMAPINFO));
	bmpInfoInput.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	
	bmpInfoInput.bmiHeader.biWidth = width;
	bmpInfoInput.bmiHeader.biHeight = height;
	bmpInfoInput.bmiHeader.biPlanes = 1;
	bmpInfoInput.bmiHeader.biBitCount = bpp;
	bmpInfoInput.bmiHeader.biCompression = BI_RGB;
	bmpInfoInput.bmiHeader.biSizeImage = bmpInfoInput.bmiHeader.biWidth * 
		bmpInfoInput.bmiHeader.biHeight * (bmpInfoInput.bmiHeader.biBitCount/8);

	memset(&compressOpts, 0, sizeof(compressOpts));

	// Set this flag to indicate that we've manually "adjusted" the
	// compressOpts variable.
	compressOpts.cbSize = sizeof(COMPVARS);

	// Pop up a dialog to let the user choose the form of the compression
	// (including the option for no compression).
	ICCompressorChoose(NULL,      
		0,   
		&bmpInfoInput,
		aviStream,  
		&compressOpts,   
		"Compression Options");

	// If the user chooses "no compression", compressOpts.hic will be NULL but
	// fccHandler will equal mmioFOURCC('D', 'I', 'B', ' ').
	if((compressOpts.fccHandler != mmioFOURCC('D', 'I', 'B', ' ')) && !compressOpts.hic){
		// Cancle the record command after cleaning up memory and files
		Close();
		return FALSE;
	}

	// Get the output format for the compressor (if needed)
	if(compressOpts.hic){
		if(ICCompressGetFormat(compressOpts.hic, &bmpInfoInput, &bmpInfoOutput) != ICERR_OK){
			throw QMOL_ERROR("Can't get compression format!");
		}
	}

	// For debugging purposes, grab the four chacter codes used to describe
	// compression formats
	//char fcc[4];
	//fcc[0] = (char)(compressOpts.fccHandler & 0x000000ff);
	//fcc[1] = (char)((compressOpts.fccHandler & 0x0000ff00) >> 8);
	//fcc[2] = (char)((compressOpts.fccHandler & 0x00ff0000) >> 16);
	//fcc[3] = (char)((compressOpts.fccHandler & 0xff000000) >> 24);

	AVICOMPRESSOPTIONS opts;
	
	// Set the AVICOMPRESSOPTIONS structure based on the contents of compressOpts
	opts.fccType = streamtypeVIDEO;
    opts.fccHandler = compressOpts.fccHandler;
    opts.dwKeyFrameEvery = 0;
    opts.dwQuality = compressOpts.lQ; 
    opts.dwBytesPerSecond = 0; 
    opts.dwFlags = AVICOMPRESSF_VALID; 
    opts.lpFormat = 0; 
    opts.cbFormat = 0; 
    opts.lpParms = compressOpts.lpState; 
    opts.cbParms = compressOpts.cbState; 
    opts.dwInterleaveEvery = 0; 

	if(AVIMakeCompressedStream(&aviStreamCompressed, aviStream, &opts, NULL) != AVIERR_OK)
		throw QMOL_ERROR("Couldn't make compressed AVI stream!");

	///////////////////// Create the frame in memeory /////////////////////
	// Create a DIB in memory that we can render our OpenGL scene to
	
	HDC hTmpDC;

	// Create DIB
	hTmpDC = ::GetDC(NULL);
	hDib = CreateDIBSection(hTmpDC, &bmpInfoInput, DIB_RGB_COLORS, 
		&pBitmapBits, NULL, (DWORD)0);
	::ReleaseDC(NULL, hTmpDC);

	// Save the image size
	bitmapSize = bmpInfoInput.bmiHeader.biSizeImage;

	// Set the stream format
	if(AVIStreamSetFormat(aviStreamCompressed, 0, &(bmpInfoInput.bmiHeader), 
		sizeof(BITMAPINFOHEADER)) != AVIERR_OK)
		throw QMOL_ERROR("Couldn't set compressed AVI format!");

	///////////////////// Initialize OpenGL /////////////////////
	CQmolView *pQmView = (CQmolView *)pView;

	// Store current rendering and device contexts
	hDC = wglGetCurrentDC();
	hRC = wglGetCurrentContext();

	// Create memory device context
	if ((hMemDC = CreateCompatibleDC(pQmView->Get_pDC() == NULL ? 
		NULL : pQmView->Get_pDC()->GetSafeHdc())) == NULL)
		throw QMOL_ERROR("Error creating device context for movie frame!");

	hOldDib = SelectObject(hMemDC, hDib);
	
	// Setup pixel format
	if(!pQmView->SetMemDcPixelFormat(hMemDC, bpp))
		throw QMOL_ERROR("Can't set pixel format for movie frame!");

	// Create memory rendering context
	if((hMemRC = wglCreateContext(hMemDC)) == NULL)
		throw QMOL_ERROR("Error creating memory device context for movie frame!");

	// Make this hMemRC the current OpenGL rendering context.
	if(wglMakeCurrent(hMemDC, hMemRC) == FALSE){
		throw QMOL_ERROR("wglMakeCurrent failed!");
	}
	
	// Set up OpenGL - this must be done for each new rendering context.
	pQmView->InitializeOpenGL();

	#ifdef _DEBUG
	pQmView->GetOpenGLError(__FILE__, __LINE__);
	#endif // _DEBUG

	// Return the system to it's previous state untill it's time to
	// render a frame
	//if(wglMakeCurrent(hDC, hRC) == FALSE){
	//	throw QMOL_ERROR("wglMakeCurrent failed!");
	//}

	// Start the compressor
	if(compressOpts.hic){
		if(!ICSeqCompressFrameStart(&compressOpts, &bmpInfoInput)){
			throw QMOL_ERROR("Can't initialize compressor!");
		}
	}

	// Everything has been set up
	valid = TRUE;

	return TRUE;
}

void QmolMovie::Render()
{
	if(!valid)
		throw QMOL_ERROR("Invalid state");
	
	int tmp_width, tmp_height;

	CQmolView *pQmView = (CQmolView *)pView;

	// Make this hMemRC the current OpenGL rendering context.
	//if(wglMakeCurrent(hMemDC, hMemRC) == FALSE)
	//	throw QMOL_ERROR("wglMakeCurrent failed");

	//////////////////////////////////////////////////////////////////////
	// Insert OpenGL rendering code here ...
	//////////////////////////////////////////////////////////////////////
	pQmView->SetupGLView(width, height);

	#ifdef _DEBUG
	pQmView->GetOpenGLError(__FILE__, __LINE__);
	#endif // _DEBUG

	// Save the current viewport dimensions
	tmp_width = pQmView->viewWidth;
	tmp_height = pQmView->viewHeight;

	pQmView->viewWidth = width;
	pQmView->viewHeight = height;

	CQmolDoc* pDoc = pQmView->GetDocument();
	ASSERT_VALID(pDoc);

	#ifdef _DEBUG
	pQmView->GetOpenGLError(__FILE__, __LINE__);
	#endif // _DEBUG

	if(update_surf){
		ListIterator<molecule *> next(pDoc->molList);
		molecule **tmpMol;

		while(tmpMol = next()){
			if((*tmpMol)->surf.ValidSurface()){
				(*tmpMol)->surf.BuildSurface(&((*tmpMol)->atoms), 
					&((*tmpMol)->traj), (*tmpMol)->current_step);
			}
		}
	}

	pQmView->RenderScene(pDoc->molList, QM_RENDER_SCREEN);
	
	// Make sure there is no rendering left to do
	glFinish();

	// Restore the original viewport dimensions
	pQmView->viewWidth = tmp_width;
	pQmView->viewHeight = tmp_height;

	//////////////////////// Compress the frame ////////////////////////
	
	if(compressOpts.hic){
		BOOL isKeyFrame;
		LONG compressedSize = 0L;
		VOID *compressBuffer;

		compressBuffer = ICSeqCompressFrame(&compressOpts, 0,  
			pBitmapBits, &isKeyFrame,  &compressedSize);
 
		if(!compressBuffer)
			throw QMOL_ERROR("Error compressing frame");

		//////////////////////// Write out the compressed frame ////////////////////////
		if(AVIStreamWrite(aviStream,				// stream pointer
		   frames_written,
		   1,										// number to write
		   compressBuffer, //pBitmapBitsCompressed,
		   compressedSize, //bitmapSizeCompressed,	// size of this frame
		   AVIIF_KEYFRAME,							// flags....
		   NULL,
		   NULL) != AVIERR_OK){
			throw QMOL_ERROR("Can't write movie frame!");
		}
	}
	else{
		//////////////////////// Write out the uncompressed frame ////////////////////////
		if(AVIStreamWrite(aviStream,	// stream pointer
		   frames_written,
		   1,							// number to write
		   pBitmapBits,
		   bitmapSize,					// size of this frame
		   AVIIF_KEYFRAME,				// flags....
		   NULL,
		   NULL) != AVIERR_OK){
			throw QMOL_ERROR("Can't write movie frame!");
		}
	}

	// Increment the frame counter
	frames_written++;

	// Return the system to it's previous state untill it's time to
	// render a frame
	//if(wglMakeCurrent(hDC, hRC) == FALSE)
	//	throw QMOL_ERROR("wglMakeCurrent failed");
}

void QmolMovie::Close()
{
	if(!valid)
		return;
	
	valid = FALSE;

	// The rendering context will be no longer needed
	wglMakeCurrent(NULL, NULL);

	if(hMemRC){
		wglDeleteContext(hMemRC);
		hMemRC = NULL;
	}

	// Clean up the compressor
	if(compressOpts.hic){
		ICSeqCompressFrameEnd(&compressOpts);
		ICCompressorFree(&compressOpts);
		memset(&compressOpts, 0, sizeof(compressOpts));
	}

	// Return the system to it's initial state
	if(hDC && hRC){
		wglMakeCurrent(hDC, hRC);
		hDC = NULL;
		hRC = NULL;
	}

	if(hOldDib && hMemDC){
		SelectObject(hMemDC, hOldDib);
		hOldDib = NULL;
	}

	// Delete our DIB and device context
	if(hDib){
		DeleteObject(hDib);
		hDib = NULL;
	}

	if(hMemDC){
		DeleteDC(hMemDC);
		hMemDC = NULL;
	}

	pBitmapBits = NULL;
	bitmapSize = 0;

	filename = _T("");

	frames_written = 0;
	width = 0;
	height = 0;
	fps = 0;
	bpp = 0;

	if(aviStreamCompressed){
		AVIStreamClose(aviStreamCompressed);
		aviStreamCompressed = NULL;
	}

	if(aviStream){
		AVIStreamClose(aviStream);
		aviStream = NULL;
	}

	if(aviFile){
		AVIFileRelease(aviFile);
		aviFile = NULL;

		AVIFileExit();
	}

	// Fill in the header for the video stream
	memset(&aviInfo, 0, sizeof(aviInfo));
}
