//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/
// 
#include "stdafx.h"
#include "Qmol.h"
#include "QmolDoc.h"
#include "LinearAlgebra3D.h"
#include <math.h>

// Energetic cut off for the Kabsch & Sander H-bond definition
#define		HBOND_CUT_OFF	(-0.5f) // kcal/mol

#define		Q1				(0.42f)
#define		Q2				(0.20f)
#define		F_HB			(332.0f)
#define		DIST_CUT_OFF	(2.1f*2.1f)

void molecule::UpdateHBonds(int cur_step)
{
	// Are there any hydrogen bonds to compute?
	if(NumAtoms() == 0)
		return;

	// Adjust the current step index
	cur_step --;

	int num_hbonds = 0;
	
	float r[3];
  	int i, j, k;
  	float boxsizex, boxsizey, boxsizez;
  	int xb, yb, zb, xytotb, totb, axb, ayb, azb, aindex, bindex;
	int xi, yi, zi;
	float tmp_xmax, tmp_xmin, tmp_ymax, tmp_ymin, tmp_zmax, tmp_zmin;
	float tmp_xsize, tmp_ysize, tmp_zsize;

	// Find min/max bounds of molecule's coordinates.
	if(cur_step == -1){
		atoms[0].GetCoord(r);
	}
	else{
		memcpy(r, &(traj[cur_step][0]), 3*sizeof(float));
	}

	tmp_xmin = tmp_xmax = r[X];
	tmp_ymin = tmp_ymax = r[Y];
	tmp_zmin = tmp_zmax = r[Z];

  	for(i = 1;i < NumAtoms(); i++){
		if(cur_step == -1){
			atoms[i].GetCoord(r);
		}
		else{
			memcpy(r, &(traj[cur_step][3*i]), 3*sizeof(float));
		}

		if(r[X] < tmp_xmin)
			tmp_xmin = r[X];
		else if(r[X] > tmp_xmax)
        	tmp_xmax = r[X];

		if(r[Y] < tmp_ymin)
        	tmp_ymin = r[Y];
		else if(r[Y] > tmp_ymax)
        	tmp_ymax = r[Y];

		if(r[Z] < tmp_zmin)
        	tmp_zmin = r[Z];
		else if(r[Z] > tmp_zmax)
			tmp_zmax = r[Z];
	}

	// From size of molecule, break up space into boxes of side length=4 A
  	// (or, if the molecule size is too large in any direction, the boxes are
  	// increased so that a maximum of 50 boxes are used in any direction)
	tmp_xsize = tmp_xmax - tmp_xmin;
	tmp_ysize = tmp_ymax - tmp_ymin;
	tmp_zsize = tmp_zmax - tmp_zmin;

  	boxsizex = tmp_xsize >= 200.0f ? tmp_xsize/50.0f : 4.0f;
  	boxsizey = tmp_ysize >= 200.0f ? tmp_ysize/50.0f : 4.0f;
  	boxsizez = tmp_zsize >= 200.0f ? tmp_zsize/50.0f : 4.0f;

  	xb = (int)(tmp_xsize / boxsizex) + 1;
  	yb = (int)(tmp_ysize / boxsizey) + 1;
  	zb = (int)(tmp_zsize / boxsizez) + 1;

  	xytotb = yb*xb;
  	totb = xytotb*zb;

	list<int> **boxatom = new list<int>* [totb];
	list<int> *tmpbox, *nbrbox;
	
	list<bond> hbond_list;

	for(i = 0; i < totb;i++)
		boxatom[i] = NULL;

  	// For all the atoms, put them in their particular box
  	for(i = 0; i < NumAtoms(); i++) {

		// Is this atom a possible hydrogen bond donor or acceptor?
		BOOL isDonor = IsHBondDonor(i);
		BOOL isAcceptor = FALSE;

		if(!isDonor){
			isAcceptor = IsHBondAcceptor(i);

			if(!isAcceptor)
				continue;
		}
		
		if(cur_step == -1){
			atoms[i].GetCoord(r);
		}
		else{
			memcpy(r, &(traj[cur_step][3*i]), 3*sizeof(float));
		}

    	axb = (int)((r[X] - tmp_xmin) / boxsizex);
    	ayb = (int)((r[Y] - tmp_ymin) / boxsizey);
    	azb = (int)((r[Z] - tmp_zmin) / boxsizez);
		
    	aindex = azb*xytotb + ayb*xb + axb;

		if(boxatom[aindex] == NULL)
			// Only allocate memory for a list if there is at least one
			// atom in the box
			boxatom[aindex] = new list<int>;

		// Add the current record to box aindex. Donor atoms have positive indicies
		// while acceptor atoms have negative indicies. The index must be stored
		// as (+/-)(i+1) to allow the i == 0 index to be specified as a donor or
		// acceptor.
		if(isDonor)
			boxatom[aindex]->Push(i+1);
		else // isAcceptor
			boxatom[aindex]->Push(-(i+1));
	}
  
  	// Now, for each box, check atoms in (positive) neighbor boxes for bonds
  	aindex = 0;
  	for(zi = 0;zi < zb; zi++){
		for(yi = 0;yi < yb; yi++){
      		for(xi = 0;xi < xb; xi++){
        		tmpbox = boxatom[aindex];
				
				// Only check boxes that contain at least one atom
				if(tmpbox != NULL){
					for(j = 0;j < 14; j++){
	  					switch(j){
	    					case 0: 
								bindex = aindex; 
								break; // same
	    					case 1: 
								if(xi < (xb - 1)) 
									bindex = aindex + 1;
								else
									continue;
								break; // X
	    					case 2: 
								if(yi < (yb - 1)) 
									bindex = aindex + xb;
								else
									continue;
								break; // Y
            				case 3:
								if(zi < (zb - 1))
									bindex = aindex + xytotb;
								else
									continue;
								break; // Z		
	    					case 4:
								if(xi < (xb - 1) && yi < (yb - 1))
	              					bindex = aindex + xb + 1;
								else
									continue;
		    					break; // XY
	    					case 5: 
								if(xi < (xb - 1) && zi < (zb - 1))
	              					bindex = aindex + xytotb + 1;
								else
									continue;
		    					break; // XZ
	    					case 6: 
								if(yi < (yb - 1) && zi < (zb - 1))
	              					bindex = aindex + xytotb + xb;
								else
									continue;
		    					break; // YZ
	    					case 7:
								if(xi < (xb - 1) && yi > 0)
	              					bindex = aindex - xb + 1;
								else
									continue;
		    					break; // X-Y
	    					case 8:
								if(xi > 0 && zi < (zb - 1))
	              					bindex = aindex + xytotb - 1;
								else
									continue;
		    					break; // -XZ
	    					case 9: 
								if(yi > 0 && zi < (zb - 1))
	              					bindex = aindex + xytotb - xb;
								else
									continue;
		    					break; // -YZ
	    					case 10: 
								if(xi < (xb - 1) && yi < (yb - 1) && zi < (zb - 1))
	              					bindex = aindex + xytotb + xb + 1;
								else
									continue;
		    					break; // XYZ
	    					case 11:
								if(xi > 0 && yi < (yb - 1) && zi < (zb - 1))
	              					bindex = aindex + xytotb + xb - 1;
								else
									continue;
		    					break; // -XYZ
	    					case 12:
								if(xi < (xb - 1) && yi > 0 && zi < (zb - 1))
	              					bindex = aindex + xytotb - xb + 1;
								else
									continue;
		    					break; // X-YZ
	    					case 13: 
								if(xi > 0 && yi > 0 && zi < (zb - 1))
	              					bindex = aindex + xytotb - xb - 1;
								else
									continue;
		    					break; //-X-YZ
	  					};
	  					
						if((nbrbox = boxatom[bindex]) == NULL)
							// Skip to the next box if the current box is
							// empty
							continue;
							
						// The number of atoms in box a and box b.
						int num_inbox_a = tmpbox->Length();
						int num_inbox_b = nbrbox->Length();
							
						// The current atom in box a and box b.
						int curr_a, curr_b;

            			for(i = 0;i < num_inbox_a; i++){
							curr_a = (*tmpbox)[i];
							
							// Is the current atom a donor or acceptor?
							if(curr_a > 0){
								// This is a donor atom!
								curr_a --;

								for(k = 0;k < num_inbox_b; k++){
									curr_b = (*nbrbox)[k];
									
									if(curr_b > 0)
										// This is a donor atom -- skip it
										continue;
									
									curr_b = -1 - curr_b;

									// Test for hydrogen bonding here
									if(IsHBond(curr_a, curr_b, cur_step)){
										bond tmp_bond;
										
										tmp_bond.a = curr_a;
										tmp_bond.b = curr_b;

										hbond_list.Push(tmp_bond);
									}
	      						}
							}
							else{
								// Don't double count the bonds
								if(bindex == aindex)
									continue;

								// This is an acceptor atom
								curr_a = -1 - curr_a;

								for(k = 0;k < num_inbox_b; k++){
									curr_b = (*nbrbox)[k];
									
									if(curr_b < 0)
										// This is an acceptor atom -- skip it
										continue;
									
									curr_b --;

									// Test for hydrogen bonding here
									if(IsHBond(curr_b, curr_a, cur_step)){
										bond tmp_bond;
										
										tmp_bond.a = curr_a;
										tmp_bond.b = curr_b;

										hbond_list.Push(tmp_bond);
									}
	      						}	
							}
	  					}
					}
				}
				aindex++;
			}
		}
  	}
  
	// free up the storage space allocted for the grid search
	for(i = 0;i < totb; i++)
    	if(boxatom[i])  
			delete boxatom[i];
  
	delete [] boxatom;

	// Update the hbond array
	hbonds.Alloc(hbond_list.Length());

	for(i = 0;i < hbonds.Length();i++)
		hbonds[i] = hbond_list[i];
}

BOOL molecule::IsHBondDonor(int i)
{
	// Hydrogen atoms that are bonded to oxygen or nitrogen 
	// can be donors
	if(atoms[i].elem == ELEM_H){
		if(mol_connect[i].Length() != 1)
			return FALSE;
		
		int tmp_elem = atoms[mol_connect[i][0]].elem;

		if((tmp_elem == ELEM_O) || 
		   (tmp_elem == ELEM_N))
			return TRUE;

		return FALSE;
	}

	// For structures where the hydrogen atoms have not been provided,
	// the structure of the peptide plane is used to add the hydrogen
	// that is attached to the nitrogen.
	if(atoms[i].elem == ELEM_N)
		return TRUE;

	return FALSE;
}

BOOL molecule::IsHBondAcceptor(int i)
{
	// Double bonded Oxygen atoms (O == C) may be hydrogen bond acceptors
	if(atoms[i].elem == ELEM_O && (mol_connect[i].Length() == 1))
		return TRUE;

	return FALSE;
}

BOOL molecule::IsHBond(int dnr, int acc, int cur_step)
{
	// Hydrogen bonds are not allowed between different models
	if(atoms[dnr].model != atoms[acc].model)
		return FALSE;

	// Only singly bonded oxygen atoms are valid acceptors.
	// Find the atom number if the carbon atom attached to the
	// oxygen atom
	int acc_c;
	
	if(mol_connect[acc].Length() != 1)
		return FALSE;

	acc_c = mol_connect[acc][0];

	// The atom number of the nitrogen or oxygen attached to the
	// hydrogen
	int dnr_n_or_o;

	if(atoms[dnr].elem == ELEM_H){
			
		dnr_n_or_o = mol_connect[dnr][0];

		// Compute the hbond energy
		if(HBondEnergy(dnr_n_or_o, dnr, acc, acc_c, cur_step) < HBOND_CUT_OFF)
			return TRUE;

		return FALSE;
	}

	// If the donor is a nitrogen, we need to
	// (a) Make sure a hydrogen is not already present and
	// (b) compute the location of the "virtual" hydrogen from the
	// structure if the peptide plane.
	if(atoms[dnr].elem == ELEM_N){
		// Look for attached hydrogen
		for(int i = 0;i < mol_connect[dnr].Length();i++)
			if(atoms[mol_connect[dnr][i]].elem == ELEM_H)
				return FALSE;
		
		array<float> r_h(3);

		// Given the atom number of the non-hydrogen donor atom, place the
		// "virtual" hydrogen in the correct location.
		if(!PlaceHydrogen(r_h, dnr, acc, cur_step))
			// Unable to place the hydrogen -- bad or unknown geometery
			return FALSE;

		// Compute the hbond energy
		if(HBondEnergy(dnr, r_h, acc, acc_c, cur_step) < HBOND_CUT_OFF)
			return TRUE;

		return FALSE;
	}

	return FALSE;
}

float molecule::HBondEnergy(int dnr_1, int dnr_2, int acc_1, int acc_2, int cur_step)
{
	// Use the H-bond energy function and cutoff from W. Kabsch and C. Sander
	// "Dictionary of Protien Secondary Structure: Pattern Recognition of Hydrogen-
	// Bonded and Geometrical Features", (1983) Biopolymers 22, 2577-2637
	//
	// dnr_1 -- dnr_2 .... acc_1 == acc_2
	//   N   --   H   ....   O   ==  C
	//
	//	An additional pairwise distance constraint between dnr_2 and acc_1 is imposed:
	//	r_{dnr-acc} <= 2.1 angstroms
	//	The cut off value of 2.1 is an upper bound to the length of a hydrogen bond
	//	(see "Proteins: Structures and molecular properties" 2nd Ed. by T. E. Creighton).

	float x, y, z;
	float r_ON, r_CH, r_OH, r_CN;

	if(cur_step == -1){
		// r_OH == dnr_2 -> acc_1
		x = atoms[dnr_2].coord[X] - atoms[acc_1].coord[X];
		y = atoms[dnr_2].coord[Y] - atoms[acc_1].coord[Y];
		z = atoms[dnr_2].coord[Z] - atoms[acc_1].coord[Z];

		r_OH = x*x + y*y + z*z;

		if(r_OH > DIST_CUT_OFF)
			return HBOND_CUT_OFF + 1.0f;

		r_OH = (float)(sqrt(r_OH));

		// r_ON == dnr_1 -> acc_1
		x = atoms[dnr_1].coord[X] - atoms[acc_1].coord[X];
		y = atoms[dnr_1].coord[Y] - atoms[acc_1].coord[Y];
		z = atoms[dnr_1].coord[Z] - atoms[acc_1].coord[Z];

		r_ON = (float)(sqrt(x*x + y*y + z*z));

		// r_CH == dnr_2 -> acc_2
		x = atoms[dnr_2].coord[X] - atoms[acc_2].coord[X];
		y = atoms[dnr_2].coord[Y] - atoms[acc_2].coord[Y];
		z = atoms[dnr_2].coord[Z] - atoms[acc_2].coord[Z];

		r_CH = (float)(sqrt(x*x + y*y + z*z));

		// r_CN == dnr_1 -> acc_2
		x = atoms[dnr_1].coord[X] - atoms[acc_2].coord[X];
		y = atoms[dnr_1].coord[Y] - atoms[acc_2].coord[Y];
		z = atoms[dnr_1].coord[Z] - atoms[acc_2].coord[Z];

		r_CN = (float)(sqrt(x*x + y*y + z*z));
	}
	else{
		// r_OH == dnr_2 -> acc_1
		x = traj[cur_step][3*dnr_2] - traj[cur_step][3*acc_1];
		y = traj[cur_step][3*dnr_2 + Y] - traj[cur_step][3*acc_1 + Y];
		z = traj[cur_step][3*dnr_2 + Z] - traj[cur_step][3*acc_1 + Z];

		r_OH = x*x + y*y + z*z;

		if(r_OH > DIST_CUT_OFF)
			return HBOND_CUT_OFF + 1.0f;

		r_OH = (float)(sqrt(r_OH));

		// r_ON == dnr_1 -> acc_1
		x = traj[cur_step][3*dnr_1] - traj[cur_step][3*acc_1];
		y = traj[cur_step][3*dnr_1 + Y] - traj[cur_step][3*acc_1 + Y];
		z = traj[cur_step][3*dnr_1 + Z] - traj[cur_step][3*acc_1 + Z];

		r_ON = (float)(sqrt(x*x + y*y + z*z));

		// r_CH == dnr_2 -> acc_2
		x = traj[cur_step][3*dnr_2] - traj[cur_step][3*acc_2];
		y = traj[cur_step][3*dnr_2 + Y] - traj[cur_step][3*acc_2 + Y];
		z = traj[cur_step][3*dnr_2 + Z] - traj[cur_step][3*acc_2 + Z];

		r_CH = (float)(sqrt(x*x + y*y + z*z));

		// r_CN == dnr_1 -> acc_2
		x = traj[cur_step][3*dnr_1] - traj[cur_step][3*acc_2];
		y = traj[cur_step][3*dnr_1 + Y] - traj[cur_step][3*acc_2 + Y];
		z = traj[cur_step][3*dnr_1 + Z] - traj[cur_step][3*acc_2 + Z];

		r_CN = (float)(sqrt(x*x + y*y + z*z));
	}

	return Q1*Q2*F_HB*(1.0f/r_ON + 1.0f/r_CH - 1.0f/r_OH - 1.0f/r_CN);
}

float molecule::HBondEnergy(int dnr_1, array<float> &r_h, int acc_1, int acc_2, 
							int cur_step)
{
	// Use the H-bond energy function and cutoff from W. Kabsch and C. Sander
	// "Dictionary of Protien Secondary Structure: Pattern Recognition of Hydrogen-
	// Bonded and Geometrical Features", (1983) Biopolymers 22, 2577-2637
	//
	// dnr_1 -- dnr_2 .... acc_1 == acc_2
	//   N   --   H   ....   O   ==  C
	//
	//	An additional pairwise distance constraint between dnr_2 and acc_1 is imposed:
	//	r_{dnr-acc} <= 2.1 angstroms
	//	The cut off value of 2.1 is an upper bound to the length of a hydrogen bond
	//	(see "Proteins: Structures and molecular properties" 2nd Ed. by T. E. Creighton).

	float x, y, z;
	float r_ON, r_CH, r_OH, r_CN;

	if(cur_step == -1){
		// r_OH == dnr_2 -> acc_1
		x = r_h[X] - atoms[acc_1].coord[X];
		y = r_h[Y] - atoms[acc_1].coord[Y];
		z = r_h[Z] - atoms[acc_1].coord[Z];

		r_OH = x*x + y*y + z*z;

		if(r_OH > DIST_CUT_OFF)
			return HBOND_CUT_OFF + 1.0f;
		
		r_OH = (float)(sqrt(r_OH));

		// r_ON == dnr_1 -> acc_1
		x = atoms[dnr_1].coord[X] - atoms[acc_1].coord[X];
		y = atoms[dnr_1].coord[Y] - atoms[acc_1].coord[Y];
		z = atoms[dnr_1].coord[Z] - atoms[acc_1].coord[Z];

		r_ON = (float)(sqrt(x*x + y*y + z*z));

		// r_CH == dnr_2 -> acc_2
		x = r_h[X] - atoms[acc_2].coord[X];
		y = r_h[Y] - atoms[acc_2].coord[Y];
		z = r_h[Z] - atoms[acc_2].coord[Z];

		r_CH = (float)(sqrt(x*x + y*y + z*z));

		// r_CN == dnr_1 -> acc_2
		x = atoms[dnr_1].coord[X] - atoms[acc_2].coord[X];
		y = atoms[dnr_1].coord[Y] - atoms[acc_2].coord[Y];
		z = atoms[dnr_1].coord[Z] - atoms[acc_2].coord[Z];

		r_CN = (float)(sqrt(x*x + y*y + z*z));
	}
	else{
		// r_OH == dnr_2 -> acc_1
		x = r_h[X] - traj[cur_step][3*acc_1];
		y = r_h[Y] - traj[cur_step][3*acc_1 + Y];
		z = r_h[Z] - traj[cur_step][3*acc_1 + Z];

		r_OH = x*x + y*y + z*z;

		if(r_OH > DIST_CUT_OFF)
			return HBOND_CUT_OFF + 1.0f;
		
		r_OH = (float)(sqrt(r_OH));

		// r_ON == dnr_1 -> acc_1
		x = traj[cur_step][3*dnr_1] - traj[cur_step][3*acc_1];
		y = traj[cur_step][3*dnr_1 + Y] - traj[cur_step][3*acc_1 + Y];
		z = traj[cur_step][3*dnr_1 + Z] - traj[cur_step][3*acc_1 + Z];

		r_ON = (float)(sqrt(x*x + y*y + z*z));

		// r_CH == dnr_2 -> acc_2
		x = r_h[X] - traj[cur_step][3*acc_2];
		y = r_h[Y] - traj[cur_step][3*acc_2 + Y];
		z = r_h[Z] - traj[cur_step][3*acc_2 + Z];

		r_CH = (float)(sqrt(x*x + y*y + z*z));

		// r_CN == dnr_1 -> acc_2
		x = traj[cur_step][3*dnr_1] - traj[cur_step][3*acc_2];
		y = traj[cur_step][3*dnr_1 + Y] - traj[cur_step][3*acc_2 + Y];
		z = traj[cur_step][3*dnr_1 + Z] - traj[cur_step][3*acc_2 + Z];

		r_CN = (float)(sqrt(x*x + y*y + z*z));
	}

	return Q1*Q2*F_HB*(1.0f/r_ON + 1.0f/r_CH - 1.0f/r_OH - 1.0f/r_CN);
}


BOOL molecule::PlaceHydrogen(array<float> &r_h, int dnr, int acc, int cur_step)
{
	if(atoms[dnr].elem == ELEM_N){
		// What is the connectivity of the donor atom?
		switch(mol_connect[dnr].Length()){
			case 1:
				{
					// This atom belongs to a side chain. Assume the hydrogen atoms are
					// free to align (subject to a bond angle constraint)
					// with the acceptor atom when computing the hydrogen location.
					//
					//			 A
					//		    .
					//		   .
					//		  .
					//	     H
					//      /
					// C-- D
					float r_DA[3], r_CD[3], r_perp[3];
					float mag;
					//const float h_angle = 1.95f;
					const float h_angle = 1.2f;

					if(cur_step == -1){
						r_DA[X] = atoms[acc].coord[X] - atoms[dnr].coord[X];
						r_DA[Y] = atoms[acc].coord[Y] - atoms[dnr].coord[Y];
						r_DA[Z] = atoms[acc].coord[Z] - atoms[dnr].coord[Z];
					}
					else{
						r_DA[X] = traj[cur_step][3*acc] - traj[cur_step][3*dnr];
						r_DA[Y] = traj[cur_step][3*acc + Y] - traj[cur_step][3*dnr + Y];
						r_DA[Z] = traj[cur_step][3*acc + Z] - traj[cur_step][3*dnr + Z];
					}
					
					mag = (float)(sqrt(DOT(r_DA, r_DA)));

					if(mag < PRECISION)
						return FALSE;

					mag = 1.0f/mag;

					r_DA[X] *= mag;
					r_DA[Y] *= mag;
					r_DA[Z] *= mag;

					if(cur_step == -1){
						r_CD[X] = atoms[dnr].coord[X] - atoms[mol_connect[dnr][0]].coord[X];
						r_CD[Y] = atoms[dnr].coord[Y] - atoms[mol_connect[dnr][0]].coord[Y];
						r_CD[Z] = atoms[dnr].coord[Z] - atoms[mol_connect[dnr][0]].coord[Z];
					}
					else{
						r_CD[X] = traj[cur_step][3*dnr] - traj[cur_step][3*mol_connect[dnr][0]];
						r_CD[Y] = traj[cur_step][3*dnr + Y] - traj[cur_step][3*mol_connect[dnr][0] + Y];
						r_CD[Z] = traj[cur_step][3*dnr + Z] - traj[cur_step][3*mol_connect[dnr][0] + Z];
					}

					mag = (float)(sqrt(DOT(r_CD,r_CD)));

					if(mag < PRECISION)
						return FALSE;
					
					mag = 1.0f/mag;

					r_CD[X] *= mag;
					r_CD[Y] *= mag;
					r_CD[Z] *= mag;
					
					mag = DOT(r_CD, r_DA);

					r_perp[X] = r_DA[X] - r_CD[X]*mag;
					r_perp[Y] = r_DA[Y] - r_CD[Y]*mag;
					r_perp[Z] = r_DA[Z] - r_CD[Z]*mag;

					mag = (float)(sqrt(DOT(r_perp,r_perp)));

					if(mag < PRECISION)
						return FALSE;	
					
					mag = 1.0f/mag;

					r_perp[X] *= mag;
					r_perp[Y] *= mag;
					r_perp[Z] *= mag;

					if(cur_step == -1){
						r_h[X] = atoms[dnr].coord[X];
						r_h[Y] = atoms[dnr].coord[Y];
						r_h[Z] = atoms[dnr].coord[Z];
					}
					else{
						r_h[X] = traj[cur_step][3*dnr];
						r_h[Y] = traj[cur_step][3*dnr + Y];
						r_h[Z] = traj[cur_step][3*dnr + Z];
					}
					
					float s = (float)(sin(h_angle));
					float c = (float)(cos(h_angle));
					r_h[X] = r_h[X] + r_perp[X]*s + r_CD[X]*c;
					r_h[Y] = r_h[Y] + r_perp[Y]*s + r_CD[Y]*c;
					r_h[Z] = r_h[Z] + r_perp[Z]*s + r_CD[Z]*c;

					return TRUE;
				}
				break;
			case 2:
				{
					// This atom belongs to the peptide plane. The hydrogen position is fixed
					// by the peptide plane.
					float A[3], B[3];

					if(cur_step == -1){
						A[X] = atoms[dnr].coord[X] - atoms[mol_connect[dnr][0]].coord[X];
						A[Y] = atoms[dnr].coord[Y] - atoms[mol_connect[dnr][0]].coord[Y];
						A[Z] = atoms[dnr].coord[Z] - atoms[mol_connect[dnr][0]].coord[Z];

						B[X] = atoms[dnr].coord[X] - atoms[mol_connect[dnr][1]].coord[X];
						B[Y] = atoms[dnr].coord[Y] - atoms[mol_connect[dnr][1]].coord[Y];
						B[Z] = atoms[dnr].coord[Z] - atoms[mol_connect[dnr][1]].coord[Z];

					}
					else{
						A[X] = traj[cur_step][3*dnr] - 
							traj[cur_step][3*mol_connect[dnr][0]];
						A[Y] = traj[cur_step][3*dnr + Y] - 
							traj[cur_step][3*mol_connect[dnr][0] + Y];
						A[Z] = traj[cur_step][3*dnr + Z] - 
							traj[cur_step][3*mol_connect[dnr][0] + Z];

						B[X] = traj[cur_step][3*dnr] - 
							traj[cur_step][3*mol_connect[dnr][1]];
						B[Y] = traj[cur_step][3*dnr + Y] - 
							traj[cur_step][3*mol_connect[dnr][1] + Y];
						B[Z] = traj[cur_step][3*dnr + Z] - 
							traj[cur_step][3*mol_connect[dnr][1] + Z];
					}

					A[X] += B[X];
					A[Y] += B[Y];
					A[Z] += B[Z];
					
					float mag = (float)(sqrt(DOT(A,A)));

					// Don't divide by zero!
					if(mag < PRECISION)
						return FALSE;
					
					mag = 1.0f/mag;

					if(cur_step == -1){
						r_h[X] = A[X]*mag + atoms[dnr].coord[X];
						r_h[Y] = A[Y]*mag + atoms[dnr].coord[Y];
						r_h[Z] = A[Z]*mag + atoms[dnr].coord[Z];
					}
					else{
						r_h[X] = A[X]*mag + traj[cur_step][3*dnr];
						r_h[Y] = A[Y]*mag + traj[cur_step][3*dnr + Y];
						r_h[Z] = A[Z]*mag + traj[cur_step][3*dnr + Z];
					}

					return TRUE;
				}
				break;
		};

		return FALSE;
	}
	
	return FALSE;
}