//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/

#include "stdafx.h"

#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>
#include "pdb.h"
#include "Excepts.h"

// Utility functions
int IsStringInteger(char *string);
int NumDecimal(char *string);
void GetTwoNum(char *string, float &a, float &b);
void GetThreeNum(char *string, float &a, float &b, float &c);

// Does the given string contain only integers (or '+' and '-')?
int IsStringInteger(char *string)
{
	// Test for NULL (just in case)
	if(string == NULL)
		return 0; // False
		
	int i = 0;
	while(string[i] != '\0')
		if(((string[i] < '0') || (string[i] > '9')) && 
			(string[i] != '-') && 
			(string[i] != '+'))
			return 0; //False
		else
			i++;
			
	return 1;
}

// Count the number of decimal points in the string
int NumDecimal(char *string)
{
	char *ptr = strchr(string, '.');

	if(ptr == NULL)
		return 0;

	return 1 + NumDecimal(ptr + 1);
}

// Read two decimal numbers from a single string. Assume that
// each number has equal precision.
void GetTwoNum(char *string, float &a, float &b)
{
	// Make a copy of the input string
	char *tmp = new char[strlen(string) + 1];
	strcpy(tmp, string);

	// Find the location of the last decimal point.
	char *ptr = strrchr(tmp, '.');
	
	// Find the start of the second number (i.e. the 
	// location of the first decimal point + the precision + 1).
	// Note that strlen(ptr) = precision + 1.
	ptr = strchr(tmp, '.') + strlen(ptr);

	b = (float)(atof(ptr));

	ptr[0] = '\0';

	a = (float)(atof(tmp));

	delete [] tmp;
}

// Read three decimal numbers from a single string. Assume that
// each number has equal precision.
void GetThreeNum(char *string, float &a, float &b, float &c)
{
	// Make a copy of the input string
	char *tmp = new char[strlen(string) + 1];
	strcpy(tmp, string);

	// Find the location of the last decimal point.
	char *ptr = strrchr(tmp, '.');
	
	// Find the start of the second number (i.e. the 
	// location of the first decimal point + the precision + 1).
	// Note that strlen(ptr) = precision + 1.
	ptr = strchr(tmp, '.') + strlen(ptr);

	GetTwoNum(ptr, b, c);

	ptr[0] = '\0';

	a = (float)(atof(tmp));

	delete [] tmp;
}

/////////////////////////////////////////////////////////////////
// Member functions of the pdb_record class.
/////////////////////////////////////////////////////////////////

void pdb_record::SetRecord(char *rec)
{
	if((rec == NULL) || (strlen(rec) > PDB_RECORD_FIELD))
		throw QMOL_ERROR("SetRecord: Error in assigning PDB_RECORD_FIELD!");
	
	// Note that record is a fixed length string!
	int i = 0;
	while(rec[i] != '\0'){
		record[i] = toupper(rec[i]);
		i++;
	}
	
	record[i] = '\0';
}

void pdb_record::SetAtomName(char *atom)
{
	int len;

	if((atom == NULL) || ((len = strlen(atom)) > PDB_NAME_FIELD))
		throw QMOL_ERROR("SetAtomName: Error in assigning PDB_NAME_FIELD");

	strcpy(name, atom);
	
	// Make the atom name all caps
	for(int i = 0;i < len;i++)
		name[i] = toupper(name[i]);
}

void pdb_record::SetResName(char *resname)
{
	if((strlen(resname) > (PDB_RESIDUE_FIELD - 1)))
		throw QMOL_ERROR("SetResName: Error in assigning PDB_RESIDUE_FIELD");
	
	int i = 0;
	
	if(resname == NULL){
		is_water = FALSE;
		for(i = 0;i < PDB_RESIDUE_FIELD - 1;i++)
			// Make resname blank
			resName[i] = ' ';
			
		resName[PDB_RESIDUE_FIELD - 1] = '\0';
		
		return;
	}
	
	// Note that record is a fixed length string!
	while(resname[i] != '\0'){
		resName[i] = toupper(resname[i]);
		i++;
	}
	resName[i] = '\0';
	
	is_water = FALSE;
	
	// Is the atom part of a water molecule?
	// Test by comparing the residue name to a list
	// of all known water names.
	if(strncmp(resName, "TIP", 3) == 0)
		is_water = TRUE;
	
	if(strncmp(resName, "SPC", 3) == 0)
		is_water = TRUE;
	
	if(strncmp(resName, "HOH", 3) == 0)
		is_water = TRUE;

	if(strncmp(resName, "SOL", 3) == 0)
		is_water = TRUE;
}
		
pdb_record& pdb_record::operator =(const pdb_record& rec)
{
    strcpy(record, rec.record);
    serial = rec.serial;
    strcpy(name, rec.name);
    strncpy(resName, rec.resName, PDB_RESIDUE_FIELD);
    resSeq = rec.resSeq;
	is_water = rec.is_water;
	model = rec.model;
	chainId = rec.chainId;
    x = rec.x;
    y = rec.y;
    z = rec.z;
    occupancy = rec.occupancy;
    tempFactor = rec.tempFactor;
    
    return (*this);
}

// The ParseRecord function really needs to be rewritten! I would recommend
// parsing a line of text into a list of string tokens at the very begining
// (instead of using the strtok function throughout the body of the function).
//
// The model number is specified by "m".
int pdb_record::ParseRecord(int &m, pdb_record* ptr_last_record, gzFile f_in)
{
	char buffer[PDB_BUFFER_SIZE];
	char *temp;
    int str_len;
	char *mask = " \t\r\n";
	static BOOL read_mol_ter = FALSE;
	static BOOL substitue_chain_id = FALSE;
	int new_chain_id, old_chain_id;

#ifdef USE_ZLIB
	while(gzgets(f_in, buffer, PDB_BUFFER_SIZE) != NULL){	
#else
	while(fgets(buffer, PDB_BUFFER_SIZE, f_in) != NULL){
#endif	
		temp = strtok(buffer, mask);

		if(temp == NULL)
			// We just read a blank line -- skip the line and keep reading
			continue;

		int is_hetatm = 1;

		if((strcmp(temp, "ATOM") == 0) || 
			((is_hetatm = strcmp(temp, "HETATM")) == 0)){
			SetModel(m);
			SetRecord(temp);
			SetSerial(atoi(strtok(NULL, mask)));
			
			temp = strtok(NULL, mask);
			
			// Make sure that the atom name and the residue name have not been
			// concatinated.
			if((str_len = strlen(temp)) <= 4){
				// We've read what looks like a valid name - keep
				// reading after setting the residue name

				SetAtomName(temp);

				temp = strtok(NULL, mask);
				
				// Allow small molecule files (i.e. non-protein)
				// to skip the residue name and go directly to the coordiante data
				//if(isdigit(temp[0]))
				if(IsStringInteger(temp)){
					SetResName("UNK0");
				}
				else{

					// Check to make sure that the atom name is not two separate
					// characters (as can happen in hetatom groups)
					if(is_hetatm == 0){ // 0 == TRUE in this case
						// This is a HETATOM record
						if(strlen(temp) == 1){
							// Throw away this one character (which is most likely the
							// second part of a two part atom name).
							temp = strtok(NULL, mask);
							SetResName(temp);
						}
						else{
							// Everything is ok -- just set the residue name
							SetResName(temp);
						}
					}
					else{
						// This is an ATOM record
						SetResName(temp);
					}
					
					temp = strtok(NULL, mask);
				}
			}
			else{ // strlen(name) > 4
				// We've read a name that is too long - chances are we've 
				// concatinated the atom name and the residue name! For lack of
				// a better plan, assign the last three characters of the name field
				// to the residue field and keep reading...
				strncpy(name, temp, min(str_len - 3, PDB_NAME_FIELD));

				if(str_len - 3 < PDB_NAME_FIELD)
					name[str_len - 3] = '\0';
				else
					name[PDB_NAME_FIELD - 1] = '\0';

				// Assign the last three characters of the string to be the residue
				// name
				SetResName(temp + str_len - 3);

				// Read the next field (the chain ID)
				temp = strtok(NULL, mask);
			}

			if(is_hetatm == 0){
				// This is a HETATM record and will be considered solvent
				// (note that this overwrites that is_water assignment that
				// is made in SetResName() above).
				is_water = TRUE;
			}

			if(temp == NULL)
				return FALSE;
			
			char *tmp = NULL;

			// Need to test for a chain ID (don't assume non-numeric
			// character is used for a chain ID). By default, if no chain ID
			// is specified, every record get a ' ' for a chain ID.
			SetChainId(' ');

			if(!isdigit(temp[0])){
				// Yup - we found a chain ID
				SetChainId(temp[0]);

				if(strlen(temp) == 1){
					// Read in the next field
					temp = strtok(NULL, mask);
				}
				else{
					// The chain ID is contigous with the residue
					// number. Just increment the pointer to skip the
					// chain ID.
					temp ++;
				}
			}
			else{
				// We just read a number, check to see if the next number is
				// an integer (i.e. a residue number) or a decimal value (i.e.
				// x coordinate).
				tmp = strtok(NULL, mask);

				if(NumDecimal(tmp) == 0){
					// It's an integer, so temp DOES contain a chainId
					SetChainId(temp[0]);

					temp = tmp;
					tmp = NULL;
				}
				else{
					// We've just read the X coordinate into tmp, while temp contains
					// the residue number. Check the length of the residue name to
					// make sure that we haven't concatinated the residue name and the
					// chain id.
					int res_len = strlen(resName);
					if(res_len > 4){
						// Assign the last character of the residue name
						// to the chain id
						SetChainId(resName[res_len - 1]);
						resName[res_len - 1] = '\0';
					}
				}
			}


			if(read_mol_ter){
				// This is the first record after a pdb "TER" record has been
				// read. If the chain ID is the same as the previous record, substitue
				// a different chain ID.
				if(ptr_last_record != NULL){
					if(GetChainId() == ptr_last_record->GetChainId()){
						substitue_chain_id = TRUE;
						old_chain_id = GetChainId();
						// Upper case
						if(isupper(old_chain_id)){
							new_chain_id = (old_chain_id + 1 - 'A')%('Z' - 'A') + 'A';
						}
						else{
							// Lower case
							if(islower(old_chain_id)){
								new_chain_id = (old_chain_id + 1 - 'a')%('z' - 'a') + 'a';
							}
							else{
								// White space
								if(isspace(old_chain_id)){
									new_chain_id = 'A';
								}
								else{
									// Must be a number
									new_chain_id = (old_chain_id + 1 - '0')%('9' - '0') + '0';
								}
							}
						}
					}
				}

				read_mol_ter = FALSE;
			}

			if(substitue_chain_id){
				if(old_chain_id == GetChainId()){
					SetChainId(new_chain_id);
				}
				else{
					// No longer reading the old chain
					substitue_chain_id = FALSE;
				}
			}

			SetResNum(atoi(temp));
			
			// There is a bug with this parsing scheme! If a pdb with fixed
			// precision has x, y and/or z values that are so large that they
			// "run together" then only the first part of the string will
			// be interpreted as a number.
			//char *tmp_x, *tmp_y, *tmp_z;
			//x = (float)(atof(tmp_x = strtok(NULL, " \t")));
			//y = (float)(atof(tmp_y = strtok(NULL, " \t")));
			//z = (float)(atof(tmp_z = strtok(NULL, " \t")));

			// If we've already read in the x coordinate, tmp will point to
			// string (and will be non-NULL).
			if(tmp == NULL)
				tmp = strtok(NULL, mask);

			switch(NumDecimal(tmp)){
				case 0:
					// This is a non-standard pdb file (the coordiante
					// is missing a decimal place i.e. entered as an integer).
				case 1:
					// Everything is ok - read x.
					x = (float)(atof(tmp));
					tmp = strtok(NULL, mask);

					switch(NumDecimal(tmp)){
						case 0:
						case 1:
						// Everything is ok - read y.
						y = (float)(atof(tmp));
					
						// Read z.
						z = (float)(atof(strtok(NULL, mask)));
						break;
					case 2:
						// Woops - concatinated y and z
						GetTwoNum(tmp, y, z);
						break;
					};
					break;
				case 2:
					// Woops - concatinated x and y
					GetTwoNum(tmp, x, y);
					
					// Read z.
					z = (float)(atof(strtok(NULL, mask)));
					break;
				case 3:
					// Woops - concatinated x, y and z
					GetThreeNum(tmp, x, y, z);
					break;
			};
			
	    	if((temp = strtok(NULL, mask)) == NULL){
				SetOcc(1.00);
				SetTemp(0.00);
				return TRUE;
	    	}
	    	else
				SetOcc((float)(atof(temp)));
	    
	    	if((temp = strtok(NULL, mask)) == NULL){
				SetTemp(0.00);
				return TRUE;
	    	}
	    	else
				SetTemp((float)(atof(temp)));
			
			
			// Test to make sure that we're not reading in another model
			// that has not properly set its model number (but set the chain
			// id instead).
			if(ptr_last_record != NULL){
				if(ptr_last_record->GetSerial() > GetSerial()){
					if(ptr_last_record->GetChainId() != GetChainId()){
						if(ptr_last_record->GetModel() == GetModel()){
							// We're reading in a new model! Increment the 
							// model counter.
							m++;
							SetModel(m);
						}
					}
				}
			}

	    	return TRUE;
		}
		else{
			if(strcmp(temp, "MODEL") == 0){
				// Read the model number
				m = atoi(strtok(NULL, mask));

				// Don't return - keep reading records!
			}

			if(strcmp(temp, "TER") == 0){
				read_mol_ter = TRUE;
			}
		}
 	}
    
	return FALSE;
}

/////////////////////////////////////////////////////////////////
// Member functions of the pdb_file class.
/////////////////////////////////////////////////////////////////

int pdb_file::SerialToRecord(int i)
{
	// Find the record number that corresponds to the given serial 
	// number.
	
	// Try the obvious choice first (i.e. serial number [1...num_records]
	if((i >= 1) && (i <= num_records))
		if(line[i-1].GetSerial() == i)
			return (i-1);
	
	int j;

	static int offset = -1;

	// Do we have an estimate of the serial number offset?
	if(offset != -1){
		int min = max(i - offset, 0);
		int max = min(i + offset, num_records);

		for(j = min;j < max;j++)
			if(line[j].GetSerial() == i)
				return j;
	}

	// Now try a brute force search but save the 
	// result in the offset variable.
	for(j = 0;j < num_records;j++)
		if(line[j].GetSerial() == i){
			offset = max(offset, abs(i - j));
			return j;
		}
	
	// If we get here, the serial number was not present in the file
	// (but theres no need for a fatal error - just test the return
	// value: 9/2/99).
	return -1;
}

int pdb_file::CountConnect(gzFile f_in)
{
    char buffer[PDB_BUFFER_SIZE];
    int count = 0;
    char *tmp;
	char *mask = " \t\r\n";

#ifdef USE_ZLIB
	while(gzgets(f_in, buffer, PDB_BUFFER_SIZE) != NULL){
#else
	while(fgets(buffer, PDB_BUFFER_SIZE, f_in) != NULL){
#endif
		tmp = strtok(buffer, mask);
		if(tmp && (strcmp(tmp, "CONECT") == 0))
			count ++;
	}

#ifdef USE_ZLIB
	gzrewind(f_in); 
#else
	rewind(f_in);
#endif	
    
    return count;    
}

int pdb_file::CountMonitors(gzFile f_in)
{
    char buffer[PDB_BUFFER_SIZE];
    int count = 0;

	// Note that this counting scheme will be derailed by entries
	// like: REMARK BLAA BLAA BLAA BLAA MONITOR BLAA BLAA BLAA.
	// But I don't think this can be helped.
#ifdef USE_ZLIB
	while(gzgets(f_in, buffer, PDB_BUFFER_SIZE) != NULL){
#else
	while(fgets(buffer, PDB_BUFFER_SIZE, f_in) != NULL){
#endif
		if(strstr(buffer, "MONITOR") != NULL)
			count ++;
	}

#ifdef USE_ZLIB
	gzrewind(f_in); 
#else
	rewind(f_in);
#endif	
    
    return count;    
}

// Destructor
pdb_file::~pdb_file()
{
	if(line != NULL)
		delete [] line;
	
	if(conn != NULL)
		delete [] conn;

	if(monitor != NULL)
		delete [] monitor;
}

// Default constructor
pdb_file::pdb_file()
{
	num_records = num_connect = num_bonds = 0;
	bbox = FALSE;
	line = NULL;
	conn = NULL;
	monitor = NULL;
}

// New pdb file constructor
pdb_file::pdb_file(int records)
{
	num_records = records;
	line = new pdb_record [num_records];
   
	num_bonds = 0;
	num_connect = 0;
	conn = NULL;
	monitor = NULL;

	bbox = FALSE;

	// Start off with some sensible defaults
	for(int i = 0;i < num_records;i++){
		SetRecord(i, "ATOM");
		SetSerial(i, i);
		SetAtomName(i, "C");
		SetResName(i, "UNK0");
		SetResNum(i, 1);
		SetCoord(i, 0.0, 0.0, 0.0);
		SetOcc(i, 1.0);
		SetTemp(i, 0.0);
	}
}

// Copy constructor
pdb_file::pdb_file(const pdb_file& pdb)
{
	num_records = num_connect = num_bonds = 0;
	bbox = FALSE;
	line = NULL;
	conn = NULL;
	monitor = NULL;

	(*this) = pdb;
}

pdb_file& pdb_file::operator =(const pdb_file& pdb)
{
	int i;

	if(pdb.num_records <= 0)
		throw QMOL_ERROR("Attempting to copy empty structure!");
    
	num_records = pdb.num_records;
	line = new pdb_record [num_records];
   
	num_bonds = pdb.num_bonds;
	
	num_connect = pdb.num_connect;
	conn = new pdb_connect [num_connect];
    
	for(i = 0;i < num_records;i++)
		line[i] = pdb.line[i];
    
	for(i = 0;i < num_connect;i++)
		conn[i] = pdb.conn[i];
	
	// Copy the bounding box
	bbox = pdb.bbox;
	
	if(bbox)
		for(i = 0;i < 3;i++){
			bbox_R[i] = pdb.bbox_R[i];
			bbox_A[i] = pdb.bbox_A[i];
			bbox_B[i] = pdb.bbox_B[i];
			bbox_C[i] = pdb.bbox_C[i];
		}
	
	return (*this);
}

#define   X     0
#define   Y     1
#define   Z     2

// If the pdb file contains a bounding box, read it.
int pdb_file::GetBBox(gzFile f_in)
{
  	char *temp = NULL;
	
	int A = FALSE, B = FALSE, C = FALSE, R = FALSE;
		
	char buffer[PDB_BUFFER_SIZE];
	
#ifdef USE_ZLIB
	gzrewind(f_in); 
#else
	rewind(f_in);
#endif	

#ifdef USE_ZLIB
	while(gzgets(f_in, buffer, PDB_BUFFER_SIZE) != NULL){
#else
	while(fgets(buffer, PDB_BUFFER_SIZE, f_in) != NULL){
#endif
		if((temp = strstr(buffer, "BBOX_R")) != NULL){
			sscanf(temp, "BBOX_R %f %f %f", bbox_R + X, bbox_R + Y, bbox_R + Z);
			R = TRUE;
		}
			
		if((temp = strstr(buffer, "BBOX_A")) != NULL){
			sscanf(temp, "BBOX_A %f %f %f", bbox_A + X, bbox_A + Y, bbox_A + Z);
			A = TRUE;
		}
			
		if((temp = strstr(buffer, "BBOX_B")) != NULL){
			sscanf(temp, "BBOX_B %f %f %f", bbox_B + X, bbox_B + Y, bbox_B + Z);
			B = TRUE;
		}
			
		if((temp = strstr(buffer, "BBOX_C")) != NULL){
			sscanf(temp, "BBOX_C %f %f %f", bbox_C + X, bbox_C + Y, bbox_C + Z);
			C = TRUE;
		}
	}
	
	if(A && B && C && R)
		bbox = TRUE;
	else
		bbox = FALSE;
	
#ifdef USE_ZLIB
	gzrewind(f_in); 
#else
	rewind(f_in);
#endif	
	
	return bbox;
}

void pdb_file::GetBBoxR(float &x, float &y, float &z)
{
	if(bbox){
		x = bbox_R[X];
		y = bbox_R[Y];
		z = bbox_R[Z];
	}
	else{
		//QMWarning("No bounding box present");
		x = y = z = 0.0;
	}
}

void pdb_file::GetBBoxA(float &x, float &y, float &z)
{
	if(bbox){
		x = bbox_A[X];
		y = bbox_A[Y];
		z = bbox_A[Z];
	}
	else{
		//QMWarning("No bounding box present");
		x = y = z = 0.0;
	}
}

void pdb_file::GetBBoxB(float &x, float &y, float &z)
{
	if(bbox){
		x = bbox_B[X];
		y = bbox_B[Y];
		z = bbox_B[Z];
	}
	else{
		//QMWarning("No bounding box present");
		x = y = z = 0.0;
	}
}

void pdb_file::GetBBoxC(float &x, float &y, float &z)
{
	if(bbox){
		x = bbox_C[X];
		y = bbox_C[Y];
		z = bbox_C[Z];
	}
	else{
		//QMWarning("No bounding box present");
		x = y = z = 0.0;
	}
}

void pdb_file::GetCoordBySerial(int num, double &a, double &b, 
								double &c)
{   
	if((num > 0) && (num <= num_records) && 
       (num == line[num-1].GetSerial())){
		line[num-1].GetCoord(a, b, c);
		return;
    }
    else{
		for(int i = 0;i < num_records;i++)
			if(line[i].GetSerial() == num){
				line[i].GetCoord(a, b, c);
				return;
			}
	}

	throw QMOL_ERROR("GetCoordBySerial: Serial number not found");
}

int pdb_file::HashSerial(int &rec, int serial)
{
    for(int i = 0;i < num_records;i++)
		if(GetSerial(i) == serial){
			rec = i;
			return TRUE;
		}

    return FALSE;
}		

// Translate coordinates in a pdb
void pdb_file::Translate(float *r)
{
  int i;
  float x, y, z;

  for(i = 0;i < num_records;i++){
    line[i].GetCoord(x, y, z);
    line[i].SetCoord(x + r[X], y + r[Y], z + r[Z]);
  }
}

void pdb_file::Rotate(float **m)
{
  int i;
  float x, y, z;

  for(i = 0;i < num_records;i++){
    line[i].GetCoord(x, y, z);
    line[i].SetCoord(m[X][X]*x + m[X][Y]*y + m[X][Z]*z,
		     m[Y][X]*x + m[Y][Y]*y + m[Y][Z]*z,
		     m[Z][X]*x + m[Z][Y]*y + m[Z][Z]*z);
  }
}

// Concatenate two pdb files
void pdb_file::Concatinate(const pdb_file& a, const pdb_file& b)
{
	int i, j;
    
	// Clean up any old data.
	if(line != NULL)
		delete [] line;
    
	num_records = a.num_records + b.num_records;
	num_connect = a.num_connect + b.num_connect;
	 
	num_bonds = a.num_bonds + b.num_bonds;
	 
	line = new pdb_record [num_records];
	conn = new pdb_connect [num_connect];
    
	for(i = 0;i < a.num_records;i++)
		line[i] = a.line[i];
    
	for(i = 0;i < a.num_connect;i++)
		conn[i] = a.conn[i];
	
	for(i = 0;i < b.num_records;i++){
		line[i+a.num_records] = b.line[i];
	
	// Adjust the atom serial numbers
	line[i+a.num_records].SetSerial(b.line[i].GetSerial() +
			    a.line[a.num_records - 1].GetSerial());

	// Adjust the residue serial numbers
	line[i+a.num_records].SetResNum(b.line[i].GetResNum() +
			    a.line[a.num_records - 1].GetResNum());			    
	}
    
	for(i = 0;i < b.num_connect;i++){
		conn[i + a.num_connect] = b.conn[i];
		conn[i + a.num_connect].SetNode(conn[i + a.num_connect].GetNode()
			+ a.line[a.num_records - 1].GetSerial());
		for(j = 0;j < b.conn[i].GetNumConnect();j++)
			conn[i + a.num_connect][j] += 
				a.line[a.num_records - 1].GetSerial();
	}
    
}

// pdb file constructor
pdb_file::pdb_file(const char *filename)
{
	num_records = num_connect = num_bonds = 0;
	line = NULL;
	conn = NULL;
	bbox = FALSE;
	monitor = NULL;

	if(!Read(filename))
		//fprintf(stderr, "Error! Could not read the file %s.\n", filename);
		throw QMOL_ERROR("Could not read the file!");
}

int pdb_file::Read(const char *filename)
{
	int i;
	int model = 0; // The model number (used for multiple copies of a 
				   // molecule in the same pdb file).

#ifdef USE_ZLIB
	// Note the mode ("rb") used to open the (possibly) compressed file.
	// Opening the file as "r" WILL NOT WORK!
	gzFile f_in = gzopen(filename, "rb");
#else
	FILE* f_in = fopen(filename,"r");
#endif // USE_ZLIB
    
	if(!f_in)
		throw QMOL_ERROR("Can't open file!");
    
	if(line != NULL)
		delete [] line;
    
	if(conn != NULL)
		delete [] conn;

	if(monitor != NULL)
		delete [] monitor;
	
	num_records = CountRecords(f_in);
	num_connect = CountConnect(f_in);
	num_monitors = CountMonitors(f_in);
	
	line = new pdb_record [num_records];
	conn = new pdb_connect [num_connect];
	monitor = new pdb_monitor [num_monitors];
	
	if(num_records != 0)
		if(!line[0].ParseRecord(model, NULL, f_in))
			throw QMOL_ERROR("Error parsing file!");

	for(i = 1;i < num_records;i++)
		if(!line[i].ParseRecord(model, &(line[i-1]), f_in))
			throw QMOL_ERROR("Error parsing file!");
   	
#ifdef USE_ZLIB
	gzrewind(f_in); 
#else
	rewind(f_in);
#endif	
    
	for(i = 0;i < num_connect;i++)
		if(!conn[i].ParseConnect(f_in))
			throw QMOL_ERROR("Error parsing file!");
	
#ifdef USE_ZLIB
	gzrewind(f_in); 
#else
	rewind(f_in);
#endif	

	if(num_connect != 0)
		num_bonds = CountBonds();
	else
		num_bonds = 0;
	
	// If the monitor data can not be correctly parsed, just skip the
	// monitor in question
	for(i = 0;i < num_monitors;i++)
		if(!monitor[i].ParseMonitor(f_in)){
			i --;
			num_monitors --;
		}
	
	// Now, lookup all of the atom index numbers
	for(i = 0;i < num_monitors;i++)
		if(!HashSerial(monitor[i].atom_index[0], monitor[i].atom_serial[0]) ||
		   !HashSerial(monitor[i].atom_index[1], monitor[i].atom_serial[1])){
			// Entry is invalid! Just skip it.
			i --;
			num_monitors --;
		}

#ifdef USE_ZLIB
	gzrewind(f_in); 
#else
	rewind(f_in);
#endif	

	// Read in the bounding box (if present)
	GetBBox(f_in);
	
	// All done!
#ifdef USE_ZLIB
	gzclose(f_in); 
#else
	fclose(f_in);
#endif
	
	return TRUE;
}

int pdb_file::CountRecords(gzFile f_in)
{
	char buffer[PDB_BUFFER_SIZE];
	char *temp;
	int count = 0;
	char *mask = " \t\r\n";

#ifdef USE_ZLIB
	while(gzgets(f_in, buffer, PDB_BUFFER_SIZE) != NULL){
#else
	while(fgets(buffer, PDB_BUFFER_SIZE, f_in) != NULL){
#endif
		temp = strtok(buffer, mask);
		if((temp != NULL) && ((strcmp(temp, "ATOM") == 0) || (strcmp(temp, "HETATM") == 0)))
			count ++;
	}
	
#ifdef USE_ZLIB
	gzrewind(f_in); 
#else
	rewind(f_in);
#endif	
    
	return count;
}

int pdb_file::Write(const char* name, int precision, BOOL write_conect)
{
	FILE* f_out = fopen(name,"w");
	float x, y, z;
	int i;
	int current_model;

	// This string allows variable precision 
	// formatting of the data fields.
	char format[32];
	
	if(!f_out)
		throw QmolError("Can't open file");

	// Do we have any data to write?
	if(num_records == 0){
		// We have an empty pdb file, so close the file pointer
		// and return.
		fclose(f_out);
		return TRUE;
	}

	current_model = GetModel(0);

	// From the PDB file format definition:
	// COLUMNS       DATA TYPE      FIELD         DEFINITION
	// ----------------------------------------------------------------------
	// 1 -  6       Record name    "MODEL "
	// 
	// 11 - 14       Integer        serial        Model serial number.

	fprintf(f_out, "MODEL%5d\n", current_model);

	for(i = 0;i < num_records;i++){
		// Do we need to increment the model?
		if(current_model != GetModel(i)){
			current_model = GetModel(i);
			fprintf(f_out, "ENDMDL\nMODEL%5d\n", current_model);
		}

		fprintf(f_out,"%-6s", GetRecord(i));
		fprintf(f_out,"%5d ", GetSerial(i));
		fprintf(f_out," %-4s", GetAtomName(i));
		fprintf(f_out,"%-4s", GetResName(i));

		// Add the chainID here!
		fprintf(f_out,"%c", GetChainId(i));

		fprintf(f_out,"%4d", GetResNum(i));
		GetCoord(i, x, y, z);
		
		sprintf(format,"%%%d.%df", 9 + precision, precision);
		fprintf(f_out, format, x);
		sprintf(format,"%%%d.%df", 5 + precision, precision);
		fprintf(f_out, format, y);
		fprintf(f_out, format, z);
		
		fprintf(f_out,"%6.2f", GetOcc(i));
		fprintf(f_out,"%6.2f", GetTemp(i));
		
		fprintf(f_out,"\n");
	}

	if(num_records > 0){
		// Write the molecule terminator
		fprintf(f_out,"TER   ");
		fprintf(f_out,"%5d", GetSerial(num_records - 1) + 1);
		fprintf(f_out,"      ");
		fprintf(f_out,"%-4s", GetResName(num_records - 1));
		fprintf(f_out,"%c", GetChainId(num_records - 1));
		fprintf(f_out,"%4d\n", GetResNum(num_records - 1));
	}

    fprintf(f_out, "ENDMDL\n");

	if(write_conect)
		for(i = 0;i < num_connect;i++)
			conn[i].Write(f_out);
    
	// All done
	fclose(f_out);
	return TRUE;
}

int pdb_file::Write(FILE* fout, int precision, BOOL write_conect)
{
	float x, y, z;
	int i;
	int current_model;

	// This string allows variable precision 
	// formatting of the data fields.
	char format[32];

	// Do we have any data to write?
	if(num_records == 0){
		// We have an empty pdb file, so return without writing to the
		// file.
		return TRUE;
	}

	current_model = GetModel(0);

	// From the PDB file format definition:
	// COLUMNS       DATA TYPE      FIELD         DEFINITION
	// ----------------------------------------------------------------------
	// 1 -  6       Record name    "MODEL "
	// 
	// 11 - 14       Integer        serial        Model serial number.

	fprintf(fout, "MODEL%5d\n", current_model);

	for(i = 0;i < num_records;i++){
		// Do we need to increment the model?
		if(current_model != GetModel(i)){
			current_model = GetModel(i);
			fprintf(fout, "ENDMDL\nMODEL%5d\n", current_model);
		}

		fprintf(fout,"%-6s", GetRecord(i));
		fprintf(fout,"%5d ", GetSerial(i));
		fprintf(fout," %-4s", GetAtomName(i));
		fprintf(fout,"%-4s", GetResName(i));

		// Add the chainID here!
		fprintf(fout,"%c", GetChainId(i));

		fprintf(fout,"%4d", GetResNum(i));
		GetCoord(i, x, y, z);
		
		sprintf(format,"%%%d.%df", 9 + precision, precision);
		fprintf(fout, format, x);
		sprintf(format,"%%%d.%df", 5 + precision, precision);
		fprintf(fout, format, y);
		fprintf(fout, format, z);
				
		fprintf(fout,"%6.2f", GetOcc(i));
		fprintf(fout,"%6.2f", GetTemp(i));
		
		fprintf(fout,"\n");
	}
    
	if(num_records > 0){
		// Write the molecule terminator
		fprintf(fout,"TER   ");
		fprintf(fout,"%5d", GetSerial(num_records - 1) + 1);
		fprintf(fout,"      ");
		fprintf(fout,"%-4s", GetResName(num_records - 1));
		fprintf(fout,"%c", GetChainId(num_records - 1));
		fprintf(fout,"%4d\n", GetResNum(num_records - 1));
	}

	fprintf(fout, "ENDMDL\n");

	if(write_conect)
		for(i = 0;i < num_connect;i++)
			conn[i].Write(fout);

	return TRUE;
}

void pdb_file::Print(BOOL write_conect)
{
	float x, y, z;
	int i;
	int current_model;

	// Do we have any data to write?
	if(num_records == 0){
		// We have an empty pdb file, so return without writing any
		// data
		return;
	}

	current_model = GetModel(0);

	// From the PDB file format definition:
	// COLUMNS       DATA TYPE      FIELD         DEFINITION
	// ----------------------------------------------------------------------
	// 1 -  6       Record name    "MODEL "
	// 
	// 11 - 14       Integer        serial        Model serial number.

	printf("MODEL%5d\n", current_model);

	for(i = 0;i < num_records;i++){
		// Do we need to increment the model?
		if(current_model != GetModel(i)){
			current_model = GetModel(i);
			printf("ENDMDL\nMODEL%5d\n", current_model);
		}

		printf("%-6s", GetRecord(i));
		printf("%5d ", GetSerial(i));
		printf(" %-4s", GetAtomName(i));
		printf("%-4s", GetResName(i));
		printf("%c", GetChainId(i));
		printf("%5d", GetResNum(i));
		GetCoord(i, x, y, z);
		printf("%12.3f", x);
		printf("%8.3f", y);
		printf("%8.3f", z);
		printf("%6.2f", GetOcc(i));
		printf("%6.2f", GetTemp(i));
	
		printf("\n");
	}
    
	if(num_records > 0){
		// Write the molecule terminator
		fprintf(stdout,"TER   ");
		fprintf(stdout,"%5d", GetSerial(num_records - 1) + 1);
		fprintf(stdout,"      ");
		fprintf(stdout,"%-4s", GetResName(num_records - 1));
		fprintf(stdout,"%c", GetChainId(num_records - 1));
		fprintf(stdout,"%4d\n", GetResNum(num_records - 1));
	}

	printf("ENDMDL\n");

	if(write_conect)
		for(i = 0;i < num_connect;i++)
			conn[i].Write(stdout);    
}

// Determine the molecular connectivity based on pairwise distances. 
// Code based on the find_bonds_from_coordinates() function from VMD.
// The inner loop (comparing the computed pair-wise distances to a 
// sum of particle Van der Waals distances) is inspired by Rasmol.
// If use_serial is true, build the connectivity table out of atom 
// serial numbers, otherwise, use atom record numbers.
void pdb_file::ReConnect(int use_serial /* = TRUE */)
{
	if(conn != NULL){
		delete [] conn;
		conn = NULL;
  	}
	
	num_connect = num_records;
	
	num_bonds = 0;
	
	// Are there any atoms to connect? If not, just return.
	if(num_connect == 0)
		return;
		
	conn = new pdb_connect [num_connect];
	
	float r[3];
	float cutoff, dist, dr;
  	int i, j, k;
  	float xmin, xmax, ymin, ymax, zmin, zmax;
  	float boxsizex, boxsizey, boxsizez;
  	int xb, yb, zb, xytotb, totb, axb, ayb, azb, aindex, bindex, starti;
	int xi, yi, zi;
	
	// Set the node ID for each of the connection records
	if(use_serial){
		for(i = 0;i < num_connect;i++)
			conn[i].SetNode(line[i].GetSerial());
	}
	else{
		for(i = 0;i < num_connect;i++)
			conn[i].SetNode(i);
	}
		
  	// Find min/max bounds of molecule's coordinates.
	xmin = xmax = line[0].GetX();
	ymin = ymax = line[0].GetY();
	zmin = zmax = line[0].GetZ();

  	for(i = 1;i < num_records; i++){
		line[i].GetCoord(r[X], r[Y], r[Z]);

		if(r[X] < xmin)
			xmin = r[X];
		else if(r[X] > xmax)
        	xmax = r[X];

		if(r[Y] < ymin)
        	ymin = r[Y];
		else if(r[Y] > ymax)
        	ymax = r[Y];

		if(r[Z] < zmin)
        	zmin = r[Z];
		else if(r[Z] > zmax)
			zmax = r[Z];
	}
  
	// From size of molecule, break up space into boxes of side length=4 A
  	// (or, if the molecule size is too large in any direction, the boxes are
  	// increased so that a maximum of 50 boxes are used in any direction)
  	boxsizex = (float)((xmax - xmin >= 200.0 ? (xmax - xmin)/50.0 : 4.0));
  	boxsizey = (float)((ymax - ymin >= 200.0 ? (ymax - ymin)/50.0 : 4.0));
  	boxsizez = (float)((zmax - zmin >= 200.0 ? (zmax - zmin)/50.0 : 4.0));
  	xb = (int)((xmax - xmin) / boxsizex) + 1;
  	yb = (int)((ymax - ymin) / boxsizey) + 1;
  	zb = (int)((zmax - zmin) / boxsizez) + 1;
  	xytotb = yb*xb;
  	totb = xytotb*zb;
	list<int> **boxatom = new list<int>* [totb];
	list<int> *tmpbox, *nbrbox;

  	for(i = 0;i < totb; i++)
    	boxatom[i] = NULL;

  	// For all the atoms, put them in their particular box
  	for(i = 0; i < num_records; i++) {
    	axb = (int)((line[i].GetX() - xmin) / boxsizex);
    	ayb = (int)((line[i].GetY() - ymin) / boxsizey);
    	azb = (int)((line[i].GetZ() - zmin) / boxsizez);
		
    	aindex = azb*xytotb + ayb*xb + axb;

		if(boxatom[aindex] == NULL)
			// Only allocate memory for a list if there is at least one
			// atom in the box
			boxatom[aindex] = new list<int>;

		// Add the current record to box aindex
		boxatom[aindex]->Push(i);
	}
  
  	// Now, for each box, check atoms in (positive) neighbor boxes for bonds
  	aindex = 0;
  	for(zi = 0;zi < zb; zi++){
		for(yi = 0;yi < yb; yi++){
      		for(xi = 0;xi < xb; xi++){
        		tmpbox = boxatom[aindex];
				
				// Only check boxes that contain at least one atom
				if(tmpbox != NULL){
					for(j = 0;j < 14; j++){
	  					switch(j){
	    					case 0: 
								bindex = aindex; 
								break; // same
	    					case 1: 
								if(xi < (xb - 1)) 
									bindex = aindex + 1;
								else
									continue;
								break; // X
	    					case 2: 
								if(yi < (yb - 1)) 
									bindex = aindex + xb;
								else
									continue;
								break; // Y
            				case 3:
								if(zi < (zb - 1))
									bindex = aindex + xytotb;
								else
									continue;
								break; // Z		
	    					case 4:
								if(xi < (xb - 1) && yi < (yb - 1))
	              					bindex = aindex + xb + 1;
								else
									continue;
		    					break; // XY
	    					case 5: 
								if(xi < (xb - 1) && zi < (zb - 1))
	              					bindex = aindex + xytotb + 1;
								else
									continue;
		    					break; // XZ
	    					case 6: 
								if(yi < (yb - 1) && zi < (zb - 1))
	              					bindex = aindex + xytotb + xb;
								else
									continue;
		    					break; // YZ
	    					case 7:
								if(xi < (xb - 1) && yi > 0)
	              					bindex = aindex - xb + 1;
								else
									continue;
		    					break; // X-Y
	    					case 8:
								if(xi > 0 && zi < (zb - 1))
	              					bindex = aindex + xytotb - 1;
								else
									continue;
		    					break; // -XZ
	    					case 9: 
								if(yi > 0 && zi < (zb - 1))
	              					bindex = aindex + xytotb - xb;
								else
									continue;
		    					break; // -YZ
	    					case 10: 
								if(xi < (xb - 1) && yi < (yb - 1) && zi < (zb - 1))
	              					bindex = aindex + xytotb + xb + 1;
								else
									continue;
		    					break; // XYZ
	    					case 11:
								if(xi > 0 && yi < (yb - 1) && zi < (zb - 1))
	              					bindex = aindex + xytotb + xb - 1;
								else
									continue;
		    					break; // -XYZ
	    					case 12:
								if(xi < (xb - 1) && yi > 0 && zi < (zb - 1))
	              					bindex = aindex + xytotb - xb + 1;
								else
									continue;
		    					break; // X-YZ
	    					case 13: 
								if(xi > 0 && yi > 0 && zi < (zb - 1))
	              					bindex = aindex + xytotb - xb - 1;
								else
									continue;
		    					break; //-X-YZ
	  					};
	  					
						if((nbrbox = boxatom[bindex]) == NULL)
							// Skip to the next box if the current box is
							// empty
							continue;
							
						// The number of atoms in box a and box b.
						int num_inbox_a = tmpbox->Length();
						int num_inbox_b = nbrbox->Length();
							
						// The current atom in box a and box b.
						int curr_a, curr_b;

            			for(i = 0;i < num_inbox_a; i++){
							curr_a = (*tmpbox)[i];

							line[curr_a].GetCoord(r[X], r[Y], r[Z]);

	      					if(bindex == aindex)
	        					starti = i + 1;
	      					else
	        					starti = 0;
										
	      					for(k = starti;k < num_inbox_b; k++){
								curr_b = (*nbrbox)[k];

								// Only compare molecules that belong to the same model
								// (for NMR sturctures where a bundle of coordinates are
								// stored in a single pdb file).
								if(line[curr_a].GetModel() == line[curr_b].GetModel()){

									cutoff = GetVDW(curr_a) + GetVDW(curr_b);
									cutoff *= cutoff;
									
									dr = line[curr_b].GetX() - r[X];
									if((dist = dr*dr) < cutoff){
										dr = line[curr_b].GetY() - r[Y];
										if((dist += dr*dr) < cutoff){
											dr = line[curr_b].GetZ() - r[Z];
											if((dist += dr*dr) < cutoff){
												// add a new bond
												num_bonds ++;
												if(use_serial){
													conn[curr_a].
														Add(line[curr_b].GetSerial());
													conn[curr_b].
														Add(line[curr_a].GetSerial());
												}
												else{
													conn[curr_a].Add(curr_b);
													conn[curr_b].Add(curr_a);
												}
											}
										}
									}
								}
	      					}
	  					}
					}
				}
				aindex++;
			}
		}
  	}
  
	// free up the storage space allocted for the grid search
	for(i = 0;i < totb; i++)
    	if(boxatom[i])  
			delete boxatom[i];
  
	delete [] boxatom;
}


// Preserve any CONNECTION records that may be in the pdb file
// and supplement them with the necessary connection records to
// make a complete set (based on pairwise distance). If use_serial
// is true, build the connectivity table out of atom serial numbers,
// otherwise, use atom record numbers.
int pdb_file::UpdateConnect(int use_serial /* = TRUE */)
{
	// Is there anything for us to do?
	if((num_connect == num_records) || (num_records == 0))
		// There allready ARE connect records, so we'll
		// need to use the serial numbers.
		return TRUE;
	
	if((num_connect == 0) || (conn == NULL)){
		ReConnect(use_serial);
		return use_serial;
	}
	
	int i;

	// How many models are in the file? If there are more than 1,
	// skip the throw away the old CONECT records and DON'T use the
	// atom serial numbers. This is a potential bug! Not quite sure what
	// the best solution is at present.
	for(i = 1;i < num_records;i++)
		if(line[i].GetModel() != line[0].GetModel()){
			use_serial = FALSE;
			ReConnect(use_serial);
			return use_serial;
		}
	

	// If we get to this point, we HAVE to use atom serial
	// number to determine connect entries.
	use_serial = TRUE;

	// Save the current connect definitions
	int j, k, num_old_connect = num_connect;
	
	// Make sure that any pre-existing CONECT records have
	// valid entries
	for(i = 0;i < num_connect;i++)
		for(j = 0;j < conn[i].GetNumConnect();j++)
			if(SerialToRecord(conn[i][j]) == -1){
				ReConnect(use_serial);
				return use_serial;
			}
			
	pdb_connect *old_connect = new pdb_connect [num_old_connect];
	
	for(i = 0;i < num_old_connect;i++)
		old_connect[i] = conn[i];
	
	ReConnect(use_serial);

	// Merge the connect definitions and update the number of bonds
	// (i.e. num_bonds holds the number of bonds determined by ReConnect()).
	for(j = 0, i = 0;(i < num_connect) && (j < num_old_connect);i++)
		if(conn[i].GetNode() == old_connect[j].GetNode()){
			// Update the bound count
			num_bonds += old_connect[j].GetNumConnect() - conn[i].GetNumConnect();

			// Remove obsolete connections
			for(k = 0;k < conn[i].GetNumConnect();k++){
				int rec_num = SerialToRecord(conn[i][k]);

				conn[rec_num].Remove(conn[i].GetNode());
			}
			
			// Update the connect record
			conn[i] = old_connect[j++];

			// Add new connections
			for(k = 0;k < conn[i].GetNumConnect();k++){
				int rec_num = SerialToRecord(conn[i][k]);

				conn[rec_num].Add(conn[i].GetNode());
			}
		}

	// All done - free up the memory
	delete [] old_connect;

	return use_serial;
}

int pdb_file::CountBonds()
{	
	int i, j, k, n, num;

	list<pdb_bond> bond;
	num = 0;

	// Compute the atomic connectivity
	for(i = 0;i < num_connect;i++){
		int tmpnode = GetConnect(i).GetNode();
		n = GetConnect(i).GetNumConnect();

		for(j = 0;j < n;j++){
			int tmpbranch = GetConnect(i)[j];

			for(k = 0;k < num;k++){
				pdb_bond tmpbond = bond[k];

				if(((tmpnode == tmpbond.a) && (tmpbranch == tmpbond.b)) || 
					((tmpnode == tmpbond.b) && (tmpbranch == tmpbond.a)))
					// Break out of the inner-loop
					break;
			}
			// add the bond to the list if we have already
			// found all of the bounds.
			if(k == num){
				pdb_bond tmpBond;

				tmpBond.a = tmpnode;
				tmpBond.b = tmpbranch;
				
				bond.Push(tmpBond);

				num ++;
			}
		}
	}
	
	return num;
}

float pdb_file::GetMaxTemp()
{
	if(num_records == 0)
		return 0.0f;

	float max = (float)fabs(line[0].GetTemp());

	for(int i = 1;i < num_records;i++){
		if(fabs(max) < fabs(line[i].GetTemp()))
			max = line[i].GetTemp();
	}

	return max;
}

float pdb_file::GetMinTemp()
{
	if(num_records == 0)
		return 0.0f;

	float min = (float)fabs(line[0].GetTemp());

	for(int i = 1;i < num_records;i++){
		if(fabs(min) > fabs(line[i].GetTemp()))
			min = line[i].GetTemp();
	}

	return min;
}

float pdb_file::GetMaxOcc()
{
	float max = 0.0f;

	for(int i = 0;i < num_records;i++)
		if(fabs(max) < fabs(line[i].GetOcc()))
			max = line[i].GetOcc();

	return max;
}

float pdb_file::GetMinOcc()
{
	float min = 0.0f;

	for(int i = 0;i < num_records;i++)
		if(fabs(min) > fabs(line[i].GetOcc()))
			min = line[i].GetOcc();

	return min;
}

pdb_connect::pdb_connect()
{
    node = -1;
}
		 
pdb_connect::pdb_connect(pdb_connect& source)
{
    (*this) = source;
}

int pdb_connect::ParseConnect(gzFile f_in)
{
	char buffer[PDB_BUFFER_SIZE];
	char *num;
	char *mask = " \t\r\n";
	
#ifdef USE_ZLIB
	while(gzgets(f_in, buffer, PDB_BUFFER_SIZE) != NULL){
#else
	while(fgets(buffer, PDB_BUFFER_SIZE, f_in) != NULL){
#endif
		if(strcmp(strtok(buffer, mask), "CONECT") == 0){
    	    
			num = strtok(NULL, mask);
			if(num != NULL){
				// Make sure that the string only consists of numbers
				if(!IsStringInteger(num)){
					//fprintf(stderr, "Error parsing CONECT in pdb file!\n");
					//fprintf(stderr, "Found a non-numeric entry.\n");
					return FALSE;
				}
					
				node = atoi(num);
			}
			else{
				//fprintf(stderr, "Error parsing CONECT in pdb file!\n");
				return FALSE;
			}
	    	
			// Make sure that the string only consists of numbers.
			// If we read an non-numeric entry, then we're done with this
			// CONECT entry (and move on to the next).
			while(((num = strtok(NULL, mask)) != NULL) && IsStringInteger(num)){

				// According to the PDB Format Description Version 2.3, each number that appears
				// in a CONECT statement can be at most 5 digits long. If we have a string that
				// is greater that five digits long, we need to split it up (ah the joys of a 
				// fixed field format!).
				int con_len = strlen(num);

				if(con_len > 5){
					char tmpCon[6];

					tmpCon[5] = '\0';

					while(con_len > 5){
						tmpCon[4] = num[--con_len];
						tmpCon[3] = num[--con_len];
						tmpCon[2] = num[--con_len];
						tmpCon[1] = num[--con_len];
						tmpCon[0] = num[--con_len];
						
						branch.Push(atoi(tmpCon));
					}

					num[con_len] = '\0';
					
				}

				branch.Push(atoi(num));
			}
			
			return TRUE;
		}
	}
	
	return FALSE;
}

void pdb_connect::Write(FILE *fout)
{
	fprintf(fout, "CONECT");
	fprintf(fout, "%5d", node);
    
	for(int i = 0;i < branch.Length();i++)
		fprintf(fout, "%5d", branch[i]);
    
	fprintf(fout, "\n");
}

int pdb_monitor::ParseMonitor(gzFile fin)
{
	char buffer[PDB_BUFFER_SIZE];
	char *temp;

#ifdef USE_ZLIB
	while(gzgets(fin, buffer, PDB_BUFFER_SIZE) != NULL){
#else
	while(fgets(buffer, PDB_BUFFER_SIZE, fin) != NULL){
#endif
		if((temp = strstr(buffer, "MONITOR")) != NULL){
			if(sscanf(temp, "MONITOR %d %d %d %d %d %d", 
				&pairId, &(atom_serial[0]), &(atom_serial[1]),
				&(color[0]), &(color[1]), &(color[2])) != 6)
				return FALSE;
			
			// Make sure that the color records have the proper format
			for(int i = 0;i < 3;i++)
				if((color[i] < 0) || (color[i] > 255))
					return FALSE;

			return TRUE;
		}
	}

	return TRUE;
}


